/********************************************************************************
*                                                                               *
*                      F X   M O V I E   M A N A G E R                          *
*                                                                               *
*********************************************************************************
* Copyright(C) 2008,2009,2010 by Luigi Ziviani <ziviani_luigi@yahoo.it>.        *
* All Rights Reserved.                                                          *
*********************************************************************************
* Id: fxmoviemanager.cpp,   v 4.1                    2008/01/01                 *
*********************************************************************************
* This program is free software; you can redistribute it and/or modify          *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* This program is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     * 
********************************************************************************/
#include "config.h"

#include <fx.h>
#include <xincs.h>
#include <fxkeys.h>

#include "fxmoviemanager.h"
#include "icons.h"
#include "OperationsDialog.h"
#include "PropertiesDialog.h"
#include "MimeTypeDialog.h"
#include "ScanVirusDialog.h"

#ifdef HAVE_CLAMAV_H
#include <clamav.h>
#endif

#ifndef WIN32
#include <unistd.h>
#endif

#include <iostream>
#include <stdexcept>
#include <stdlib.h>

#ifdef HAVE_LIBFFMPEGTHUMBNAILER_VIDEOTHUMBNAILER_H
#include "libffmpegthumbnailer/videothumbnailer.h"
#endif
#ifdef HAVE_LIBFFMPEGTHUMBNAILER_FILMSTRIPFILTER_H
#include "libffmpegthumbnailer/filmstripfilter.h"
#endif

#define CLOCKTIMER      1000000000

using namespace FX;
using namespace std;
using namespace ffmpegthumbnailer;

// list of movie files 
static const FXchar moviefiles[]=
	"All Files (*)"
	"\nmpeg (*.mpeg,*.MPEG,*.mpg,*.MPG,*.mpe,*.MPE)"
	"\nwindows (*.wmv,*.WMV,*.wmx,*.WMX,*.asf,*.ASF,*.asx,*.ASX)"
	"\navi (*.avi,*.AVI)"
	"\nquicktime (*.mov,*.MOV,*.qt,*.QT,*.mp4,*.MP4)"
	"\nmobile (*.3gp,*.3GP,*.3gp2,*.3GP2)"
	"\nogg (*.ogv,*.OGV,*.ogm,*.OGM,*.ogg,*.OGG)"
	"\nvob (*.vob,*.VOB)"
	"\nflash (*.flv,*.FLV)"
	"\nfli (*.fli,*.FLI)"
	"\ndv (*.dif,*.DIF,*.dv,*.DV,)"
	"\ngl (*.gl,*.GL)"
	"\ndl (*.dl,*.DL)"
	"\nmovie (*.movie,*.MOVIE)";

static const FXchar audiofiles[]=
	"All Files (*)"
	"\nbasic (*.au,*.AU,*.snd,*.SND)"
	"\nmp (*.mp2,*.MP2,*.mp3,*.MP3,*.mpga,*.MPGA,*.mpega,*.MPEGA)"
	"\nreal (*.ra,*.RA,*.rm,*.RM,,*.ram,*.RAM)"
	"\nwav (*.wav,*.WAV)"
	"\nwma (*.wma,*.WMA,*.wax,*.WAX)"
	"\nmod (*.mod,*.MOD)"
	"\nmidi (*.mid,*.MID,*.midi,*.MIDI)";


FXDEFMAP(FXMovieManager) FXMovieManagerMap[] = {
	//________Message_Type_____________________________ID______________________Message_Handler__________
	// buttons
  	FXMAPFUNC(SEL_TIMEOUT,          	FXMovieManager::ID_CLOCKTIME,   	FXMovieManager::onClock),
	FXMAPFUNC(SEL_COMMAND,      		FXMovieManager::ID_QUIT,       		FXMovieManager::onCmdQuit),
	FXMAPFUNC(SEL_SIGNAL,        		FXMovieManager::ID_QUIT,       		FXMovieManager::onCmdQuit),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_FILE,       		FXMovieManager::onCmdFile),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_LOOP,       		FXMovieManager::onCmdLoop),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_FULL,       		FXMovieManager::onCmdFull),

	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_AUDIO,			FXMovieManager::onCmdAudio),
	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_AUDIO,			FXMovieManager::onUpdSelectedTable),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_PLAY,       		FXMovieManager::onCmdPlay),
	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_PLAY,			FXMovieManager::onUpdSelectedTable),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_PAUSE,      		FXMovieManager::onCmdPause),
	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_PAUSE,			FXMovieManager::onUpdSelectedTable),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_STOP,       		FXMovieManager::onCmdStop),
	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_STOP,			FXMovieManager::onUpdSelectedTable),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_DROP,       		FXMovieManager::onCmdDrop),
	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_DROP,			FXMovieManager::onUpdSelectedTable),

  	FXMAPFUNCS(SEL_UPDATE,          	FXMovieManager::ID_SIZE_0,     		FXMovieManager::ID_SIZE_11, 	FXMovieManager::onUpdSizeSwitch),
  	FXMAPFUNCS(SEL_COMMAND,         	FXMovieManager::ID_SIZE_0,     		FXMovieManager::ID_SIZE_11, 	FXMovieManager::onCmdSizeSwitch),
  	FXMAPFUNCS(SEL_UPDATE,          	FXMovieManager::ID_VIDEO_0,    		FXMovieManager::ID_VIDEO_9, 	FXMovieManager::onUpdVideoSwitch),
  	FXMAPFUNCS(SEL_COMMAND,         	FXMovieManager::ID_VIDEO_0,    		FXMovieManager::ID_VIDEO_9, 	FXMovieManager::onCmdVideoSwitch),

 	// table list
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_PUTBACK,    		FXMovieManager::onCmdPutBack),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_MOVEUP,     		FXMovieManager::onCmdMoveUp),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_MOVEDOWN,   		FXMovieManager::onCmdMoveDown),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_APPEND,     		FXMovieManager::onCmdAppend),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_INSERT,     		FXMovieManager::onCmdInsert),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_SELECTALL,  		FXMovieManager::onCmdSelectAll),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_DELETE,     		FXMovieManager::onCmdDelete),
	FXMAPFUNC(SEL_COMMAND,       		FXMovieManager::ID_CLEAR,      		FXMovieManager::onCmdClear),
	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_SAVELIST,		FXMovieManager::onCmdSavePlaylist),
	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_OPENLIST,		FXMovieManager::onCmdOpenPlaylist),
	FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,	FXMovieManager::ID_TABLE,   		FXMovieManager::onListRightMouse),

  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_PANEL,			FXMovieManager::onCmdPanel),
	// left panel directory list
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_DIRECTORYLIST_LP,FXMovieManager::onCmdDirectory_lp),
	FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,	FXMovieManager::ID_DIRECTORYLIST_LP,FXMovieManager::onDirListPopup_lp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_NEWDIR_LP,		FXMovieManager::onCmdNewDir_lp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_NEWDIR_LP,		FXMovieManager::onUpdNewDir_lp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_DIRDELETE_LP,	FXMovieManager::onCmdDirDelete_lp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_DIRDELETE_LP,	FXMovieManager::onUpdNewDir_lp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_DIRRENAME_LP,	FXMovieManager::onCmdDirRename_lp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_DIRRENAME_LP,	FXMovieManager::onUpdNewDir_lp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_UPDIRECTORY_LP,	FXMovieManager::onCmdUpDirectory_lp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_UPDIRECTORY_LP,	FXMovieManager::onUpdNewDir_lp),
	// right panel directory list
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_DIRECTORYLIST_RP,FXMovieManager::onCmdDirectory_rp),
	FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,	FXMovieManager::ID_DIRECTORYLIST_RP,FXMovieManager::onDirListPopup_rp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_NEWDIR_RP,		FXMovieManager::onCmdNewDir_rp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_NEWDIR_RP,		FXMovieManager::onUpdNewDir_rp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_DIRDELETE_RP,	FXMovieManager::onCmdDirDelete_rp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_DIRDELETE_RP,	FXMovieManager::onUpdNewDir_rp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_DIRRENAME_RP,	FXMovieManager::onCmdDirRename_rp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_DIRRENAME_RP,	FXMovieManager::onUpdNewDir_rp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_UPDIRECTORY_RP,	FXMovieManager::onCmdUpDirectory_rp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_UPDIRECTORY_RP,	FXMovieManager::onUpdNewDir_rp),
	// mapping commands
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_MAPHOST_LP,		FXMovieManager::onCmdMapHost_lp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_UNMAPHOST_LP,	FXMovieManager::onCmdUnmapHost_lp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_MAPHOST_RP,		FXMovieManager::onCmdMapHost_rp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_UNMAPHOST_RP,	FXMovieManager::onCmdUnmapHost_rp),

	// left file list
	FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,	FXMovieManager::ID_FILELIST_LP,		FXMovieManager::onFileListPopup_lp),
	FXMAPFUNC(SEL_DOUBLECLICKED, 		FXMovieManager::ID_FILELIST_LP,   	FXMovieManager::onCmdFileListDC_lp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILECOPY_LP,		FXMovieManager::onCmdFileCopy_lp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILECOPY_LP,		FXMovieManager::onUpdSelectedFile_lp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILESYMLINK_LP,	FXMovieManager::onCmdFileSymlink_lp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILESYMLINK_LP,	FXMovieManager::onUpdSelectedFile_lp),
   	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILERENAME_LP,	FXMovieManager::onCmdFileRename_lp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILERENAME_LP,	FXMovieManager::onUpdSelectedFile_lp),
   	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILEMOVE_LP,		FXMovieManager::onCmdFileMove_lp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILEMOVE_LP,		FXMovieManager::onUpdSelectedFile_lp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILEDELETE_LP,	FXMovieManager::onCmdFileDelete_lp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILEDELETE_LP,	FXMovieManager::onUpdSelectedFile_lp),
	// right file list
	FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,	FXMovieManager::ID_FILELIST_RP,		FXMovieManager::onFileListPopup_rp),
	FXMAPFUNC(SEL_DOUBLECLICKED, 		FXMovieManager::ID_FILELIST_RP,   	FXMovieManager::onCmdFileListDC_rp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILECOPY_RP,		FXMovieManager::onCmdFileCopy_rp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILECOPY_RP,		FXMovieManager::onUpdSelectedFile_rp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILESYMLINK_RP,	FXMovieManager::onCmdFileSymlink_rp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILESYMLINK_RP,	FXMovieManager::onUpdSelectedFile_rp),
   	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILERENAME_RP,	FXMovieManager::onCmdFileRename_rp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILERENAME_RP,	FXMovieManager::onUpdSelectedFile_rp),
   	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILEMOVE_RP,		FXMovieManager::onCmdFileMove_rp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILEMOVE_RP,		FXMovieManager::onUpdSelectedFile_rp),
  	FXMAPFUNC(SEL_COMMAND,				FXMovieManager::ID_FILEDELETE_RP,	FXMovieManager::onCmdFileDelete_rp),
  	FXMAPFUNC(SEL_UPDATE,				FXMovieManager::ID_FILEDELETE_RP,	FXMovieManager::onUpdSelectedFile_rp),
	// left filelist player
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_FILEPLAY_LP,		FXMovieManager::onCmdFilePlay_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILEPLAY_LP,		FXMovieManager::onUpdSelectedFile_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_FILEPAUSE_LP,	FXMovieManager::onCmdFilePause_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILEPAUSE_LP,	FXMovieManager::onUpdSelectedFile_lp),
	FXMAPFUNC(SEL_COMMAND,       	FXMovieManager::ID_FILESTOP_LP,   	FXMovieManager::onCmdFileStop_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILESTOP_LP,		FXMovieManager::onUpdSelectedFile_lp),
	// right filelist player
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_FILEPLAY_RP,		FXMovieManager::onCmdFilePlay_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILEPLAY_RP,		FXMovieManager::onUpdSelectedFile_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_FILEPAUSE_RP,	FXMovieManager::onCmdFilePause_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILEPAUSE_RP,	FXMovieManager::onUpdSelectedFile_rp),
	FXMAPFUNC(SEL_COMMAND,       	FXMovieManager::ID_FILESTOP_RP,   	FXMovieManager::onCmdFileStop_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILESTOP_RP,		FXMovieManager::onUpdSelectedFile_rp),

  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_FILEPREVIEW_LP,	FXMovieManager::onCmdFilePreview_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILEPREVIEW_LP, 	FXMovieManager::onUpdSelectedFile_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_FILEPREVIEW_RP,	FXMovieManager::onCmdFilePreview_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILEPREVIEW_RP, 	FXMovieManager::onUpdSelectedFile_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_FORCEFILEPREVIEW_LP,	FXMovieManager::onCmdForceFilePreview_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FORCEFILEPREVIEW_LP,	FXMovieManager::onUpdSelectedFile_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_FORCEFILEPREVIEW_RP,	FXMovieManager::onCmdForceFilePreview_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FORCEFILEPREVIEW_RP,	FXMovieManager::onUpdSelectedFile_rp),

  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_WIDESCREEN,		FXMovieManager::onCmdWideScreen),

  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_INCFILEPREVIEW_LP,	FXMovieManager::onCmdIncFilePreview_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_INCFILEPREVIEW_LP,	FXMovieManager::onUpdSelectedFile_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_INCFILEPREVIEW_RP,	FXMovieManager::onCmdIncFilePreview_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_INCFILEPREVIEW_RP,	FXMovieManager::onUpdSelectedFile_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_DECFILEPREVIEW_LP,	FXMovieManager::onCmdDecFilePreview_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_DECFILEPREVIEW_LP,	FXMovieManager::onUpdSelectedFile_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_DECFILEPREVIEW_RP,	FXMovieManager::onCmdDecFilePreview_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_DECFILEPREVIEW_RP,	FXMovieManager::onUpdSelectedFile_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_CLEARPREVIEW_LP,	FXMovieManager::onCmdClearPreview_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_CLEARPREVIEW_RP,	FXMovieManager::onCmdClearPreview_rp),

  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_TERMINAL_LP,		FXMovieManager::onCmdTerminal_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_TERMINAL_RP,		FXMovieManager::onCmdTerminal_rp),

  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMMAND_LP,		FXMovieManager::onCmdCommand_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMMAND_RP,		FXMovieManager::onCmdCommand_rp),

	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_SCANVIRUS,		FXMovieManager::onCmdScanVirus),

  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_TAR_LP,	FXMovieManager::onCmdCompressTAR_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_XZ_LP,	FXMovieManager::onCmdCompressXZ_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_GZ_LP,	FXMovieManager::onCmdCompressGZ_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_BZ2_LP,	FXMovieManager::onCmdCompressBZ2_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_ZIP_LP,	FXMovieManager::onCmdCompressZIP_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_7ZIP_LP,FXMovieManager::onCmdCompress7ZIP_lp),

  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_TAR_RP,	FXMovieManager::onCmdCompressTAR_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_XZ_RP,	FXMovieManager::onCmdCompressXZ_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_GZ_RP,	FXMovieManager::onCmdCompressGZ_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_BZ2_RP,	FXMovieManager::onCmdCompressBZ2_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_ZIP_RP,	FXMovieManager::onCmdCompressZIP_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_COMPRESS_7ZIP_RP,FXMovieManager::onCmdCompress7ZIP_rp),

  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_EXTRACT_LP,		FXMovieManager::onCmdExtract_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_EXTRACT_RP,		FXMovieManager::onCmdExtract_rp),

	// properties 
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_PROPERTIES_LP,	FXMovieManager::onCmdProperties_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_PROPERTIES_RP,	FXMovieManager::onCmdProperties_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_VIDEO_ASPECT,	FXMovieManager::onUpdMovieAspect),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_AUDIO_FORMAT,	FXMovieManager::onUpdAudioFormat),

  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_CREATED_LP,	FXMovieManager::onUpdCreatedTime_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_MODIFIED_LP,	FXMovieManager::onUpdModifiedTime_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_TYPE_LP,	FXMovieManager::onUpdFileType_lp),
	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_MIMETYPE_LP,	FXMovieManager::onUpdFileMimeType_lp),
	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_COMMAND_LP,	FXMovieManager::onUpdFileCommand_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_LOCATION_LP,	FXMovieManager::onUpdFileLocation_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_SIZE_LP,	FXMovieManager::onUpdFileSize_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_DESC_LP,	FXMovieManager::onUpdFileDesc_lp),

  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_CREATED_RP,	FXMovieManager::onUpdCreatedTime_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_MODIFIED_RP,	FXMovieManager::onUpdModifiedTime_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_TYPE_RP,	FXMovieManager::onUpdFileType_rp),
	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_MIMETYPE_RP,	FXMovieManager::onUpdFileMimeType_rp),
	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_COMMAND_RP,	FXMovieManager::onUpdFileCommand_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_LOCATION_RP,	FXMovieManager::onUpdFileLocation_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_SIZE_RP,	FXMovieManager::onUpdFileSize_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_FILE_DESC_RP,	FXMovieManager::onUpdFileDesc_rp),

  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_OWNER_LP,		FXMovieManager::onUpdOwner_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_OWNER_RP,		FXMovieManager::onUpdOwner_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_GROUP_LP,		FXMovieManager::onUpdGroup_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_GROUP_RP,		FXMovieManager::onUpdGroup_rp),

  	FXMAPFUNCS(SEL_COMMAND,			FXMovieManager::ID_READOWNER_LP,	FXMovieManager::ID_STICKY_LP,		FXMovieManager::onCmdFilemode_lp),
  	FXMAPFUNCS(SEL_UPDATE,			FXMovieManager::ID_READOWNER_LP,	FXMovieManager::ID_STICKY_LP,		FXMovieManager::onUpdFilemode_lp),
  	FXMAPFUNCS(SEL_COMMAND,			FXMovieManager::ID_READOWNER_RP,	FXMovieManager::ID_STICKY_RP,		FXMovieManager::onCmdFilemode_rp),
  	FXMAPFUNCS(SEL_UPDATE,			FXMovieManager::ID_READOWNER_RP,	FXMovieManager::ID_STICKY_RP,		FXMovieManager::onUpdFilemode_rp),

	// extensions association 
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_MIMETYPE,		FXMovieManager::onCmdMimeType),

  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_OPEN_LP,		FXMovieManager::onUpdItemFound_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_OPEN_LP,		FXMovieManager::onCmdOpen_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_OPEN_RP,		FXMovieManager::onUpdItemFound_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_OPEN_RP,		FXMovieManager::onCmdOpen_rp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_OPENWITH_LP,		FXMovieManager::onUpdItemFound_lp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_OPENWITH_LP,		FXMovieManager::onCmdOpenWith_lp),
  	FXMAPFUNC(SEL_UPDATE,			FXMovieManager::ID_OPENWITH_RP,		FXMovieManager::onUpdItemFound_rp),
  	FXMAPFUNC(SEL_COMMAND,			FXMovieManager::ID_OPENWITH_RP,		FXMovieManager::onCmdOpenWith_rp),
};

FXIMPLEMENT(FXMovieManager, FXMainWindow, FXMovieManagerMap, ARRAYNUMBER(FXMovieManagerMap))

FXMovieManager::FXMovieManager(FXApp *a)
	:FXMainWindow(a, "F X M o v i e M a n a g e r", NULL, NULL, DECOR_TITLE|DECOR_MINIMIZE|DECOR_MAXIMIZE|DECOR_CLOSE|DECOR_BORDER|DECOR_SHRINKABLE|DECOR_STRETCHABLE|DECOR_MENU, 0, 0, 800, 42) {

	// set icon on the main window
	logoicon=new FXPNGIcon(getApp(),fxmm32x32);
	setIcon(logoicon);

	if(!themeSelection()){
		FXMessageBox::error(this,MBOX_OK,"CONFIGURATION ERROR","Unable to load THEME!");
	}
	if(!languageSelection()){
		FXMessageBox::error(this,MBOX_OK,"CONFIGURATION ERROR","Unable to load LANGUAGE!");
	}

	// popup list menu icons
    m_upicon=new FXGIFIcon(getApp(),m_up);
    m_downicon=new FXGIFIcon(getApp(),m_down);
    m_appendicon=new FXGIFIcon(getApp(),m_append);
  	m_cancelicon=new FXGIFIcon(getApp(),m_cancel);
  	m_accepticon=new FXGIFIcon(getApp(),m_accept);
	// popup file menu icons
	m_propicon=new FXGIFIcon(getApp(),m_properties);
    m_copyicon=new FXGIFIcon(getApp(),m_copy);
    m_moveicon=new FXGIFIcon(getApp(),m_move);
    m_renameicon=new FXGIFIcon(getApp(),m_rename);
    m_deleteicon=new FXGIFIcon(getApp(),m_delete);
    m_consoleicon=new FXGIFIcon(getApp(),m_console);
    m_terminalicon=new FXGIFIcon(getApp(),m_terminal);
    m_refreshicon=new FXGIFIcon(getApp(),m_refresh);
    m_renamediricon=new FXGIFIcon(getApp(),m_renamedir);
    m_showhiddenicon=new FXGIFIcon(getApp(),m_showhidden);
    m_bugicon=new FXGIFIcon(getApp(),m_bug);
    m_archiveicon=new FXGIFIcon(getApp(),m_archive);
	m_mapicon=new FXBMPIcon(getApp(),m_maphost,0,IMAGE_ALPHAGUESS);
	m_unmapicon=new FXBMPIcon(getApp(),m_unmaphost,0,IMAGE_ALPHAGUESS);
	// menubar
	menubar = new FXMenuBar(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X);
	quitbutton = new FXButton(menubar,quitbuttontext,quiticon,this,ID_QUIT,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);
	openbutton = new FXButton(menubar,openbuttontext,openicon,this,ID_FILE,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);
	audiobutton = new FXToggleButton(menubar,audiobuttontext1,audiobuttontext2,audioicon,audioicon,this,ID_AUDIO,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);
	dropbutton = new FXToggleButton(menubar,dropbuttontext1,dropbuttontext2,dropicon,dropicon,this,ID_DROP,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);
	stopbutton = new FXButton(menubar,stopbuttontext,stopicon,this,ID_STOP,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);
	pausebutton = new FXButton(menubar,pausebuttontext,pauseicon,this,ID_PAUSE,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);
	playbutton = new FXButton(menubar,playbuttontext,playicon,this,ID_PLAY,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);
	loopbutton = new FXToggleButton(menubar,loopbuttontext1,loopbuttontext2,loopicon,loopicon,this,ID_LOOP,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);
	zoombutton = new FXToggleButton(menubar,zoombuttontext1,zoombuttontext2,videozoomicon,videozoomicon,this,ID_FULL,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);

	videosizemenu = new FXMenuPane(this);
   	videosizebutton = new FXMenuButton(menubar,sizebuttontext,videosizeicon,videosizemenu,BUTTON_TOOLBAR|MENUBUTTON_DOWN|MENUBUTTON_ATTACH_LEFT|FRAME_RAISED);
	videosizebutton->showMenu(FALSE);
    new FXMenuRadio(videosizemenu,"1008x752",this,ID_SIZE_0);
    new FXMenuRadio(videosizemenu,"1008x576",this,ID_SIZE_1);
   	new FXMenuRadio(videosizemenu,"720x576",this,ID_SIZE_2);
  	new FXMenuRadio(videosizemenu,"640x480",this,ID_SIZE_3);
    new FXMenuRadio(videosizemenu,"560x320",this,ID_SIZE_4);
  	new FXMenuRadio(videosizemenu,"500x284",this,ID_SIZE_5);
  	new FXMenuRadio(videosizemenu,"480x576",this,ID_SIZE_6);
  	new FXMenuRadio(videosizemenu,"480x480",this,ID_SIZE_7);
  	new FXMenuRadio(videosizemenu,"352x288",this,ID_SIZE_8);
  	new FXMenuRadio(videosizemenu,"352x240",this,ID_SIZE_9);
  	new FXMenuRadio(videosizemenu,"320x240",this,ID_SIZE_10);
  	new FXMenuRadio(videosizemenu,"FILESET",this,ID_SIZE_11);

	videooutmenu = new FXMenuPane(this);
    videooutbutton = new FXMenuButton(menubar,outputbuttontext,videoouticon,videooutmenu,BUTTON_TOOLBAR|MENUBUTTON_DOWN|MENUBUTTON_ATTACH_LEFT|FRAME_RAISED);
	videooutbutton->showMenu(FALSE);
   	new FXMenuRadio(videooutmenu,"X11 ",this,ID_VIDEO_0);
  	new FXMenuRadio(videooutmenu,"XV  ",this,ID_VIDEO_1);
  	new FXMenuRadio(videooutmenu,"GL  ",this,ID_VIDEO_2);
  	new FXMenuRadio(videooutmenu,"GL2 ",this,ID_VIDEO_3);
  	new FXMenuRadio(videooutmenu,"SDL ",this,ID_VIDEO_4);
  	new FXMenuRadio(videooutmenu,"DGA ",this,ID_VIDEO_5);
  	new FXMenuRadio(videooutmenu,"XVID",this,ID_VIDEO_6);
  	new FXMenuRadio(videooutmenu,"JPEG",this,ID_VIDEO_7);
  	new FXMenuRadio(videooutmenu,"PNG ",this,ID_VIDEO_8);
  	new FXMenuRadio(videooutmenu,"NULL",this,ID_VIDEO_9);

	FXVerticalFrame *ctrl_frame=new FXVerticalFrame(menubar,LAYOUT_FILL_X,0,0,0,0, 0,0,0,0, 0,0);
  	speedslider=new FXSlider(ctrl_frame,NULL,0,LAYOUT_TOP|LAYOUT_FILL_X|LAYOUT_FIX_HEIGHT|SLIDER_HORIZONTAL|SLIDER_INSIDE_BAR|SLIDER_TICKS_TOP,0,0,200,18);
  	speedslider->setRange(2,20);
	speedslider->setValue(10);
	speedslider->setIncrement(1);
	speedslider->setTipText("Speed");
	speedslider->setHelpText("Speed control");

  	volumeslider=new FXSlider(ctrl_frame,NULL,0,LAYOUT_TOP|LAYOUT_FILL_X|LAYOUT_FIX_HEIGHT|SLIDER_HORIZONTAL|SLIDER_INSIDE_BAR|SLIDER_TICKS_BOTTOM,0,0,200,18);
  	volumeslider->setRange(0,100);
	volumeslider->setValue(50);
	volumeslider->setIncrement(5);
	volumeslider->setTipText("Volume");
	volumeslider->setHelpText("Volume control");

	// Splitter
	splitter1=new FXSplitter(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y|SPLITTER_TRACKING|SPLITTER_VERTICAL|SPLITTER_REVERSED);

	// Make tablelist container
	listframe=new FXHorizontalFrame(splitter1,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0, 0,0);
	table=new FXTable(listframe,this,ID_TABLE,TABLE_COL_SIZABLE|TABLE_NO_COLSELECT|TABLE_NO_ROWSELECT|TABLE_ROW_RENUMBER|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 2,2,2,2);
  	table->setVisibleColumns(9);
  	table->setHelpText("Play List Table");
  	table->setRowRenumbering(true);
  	table->setTableSize(100,9);

	FXint c, r;
	for (r=0; r<100; r++){
	  table->setRowHeight(r,68);
	}
	for (c=1; c<9; c++){
	  table->setColumnWidth(c,30);
	}

	resetTable();

  	// Initialize column headers
    table->setColumnText(0,"movie name");
    table->setColumnText(1,"hour");
    table->setColumnText(2,"min");
    table->setColumnText(3,"sec");
    table->setColumnText(4,"cen");
    table->setColumnText(5,"hour");
    table->setColumnText(6,"min");
    table->setColumnText(7,"sec");
    table->setColumnText(8,"cen");

    table->setColumnTipText(0,"full path and movie name");
    table->setColumnTipText(1,"start hours");
    table->setColumnTipText(2,"start minutes");
    table->setColumnTipText(3,"start seconds");
    table->setColumnTipText(4,"start cents");
    table->setColumnTipText(5,"duration hour");
    table->setColumnTipText(6,"duration minutes");
    table->setColumnTipText(7,"duration seconds");
    table->setColumnTipText(8,"duration cents");

	// new FXButton(menubar,"FileList",listicon,listframe,FXWindow::ID_TOGGLESHOWN,BUTTON_TOOLBAR|FRAME_RAISED|ICON_AFTER_TEXT);
	hidebutton = new FXToggleButton(menubar,hidebuttontext1,hidebuttontext2,hideicon,hideicon,this,ID_PUTBACK,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);

	new FXButton(menubar,confbuttontext,logoicon,this,ID_MIMETYPE,BUTTON_TOOLBAR|FRAME_RAISED|ICON_UNDER_TEXT);

	// right table buttons
	listbuttonframe=new FXVerticalFrame(listframe,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0, 0,0);
	new FXButton(listbuttonframe,upbuttontext,upicon,this,FXMovieManager::ID_MOVEUP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(listbuttonframe,downbuttontext,downicon,this,FXMovieManager::ID_MOVEDOWN,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(listbuttonframe,allbuttontext1,showicon,this,FXMovieManager::ID_SELECTALL,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(listbuttonframe,clearbuttontext,deleteicon,this,FXMovieManager::ID_CLEAR,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(listbuttonframe,slistbuttontext,openicon,this,FXMovieManager::ID_SAVELIST,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(listbuttonframe,olistbuttontext,openicon,this,FXMovieManager::ID_OPENLIST,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);


	// Right mouse popup
	popupmenu = new FXMenuPane(this);
	playlistlabel = new FXLabel(popupmenu,"Play List");
	playlistlabel->setTextColor(FXRGB(150,0,0));
	new FXMenuCommand(popupmenu,"Move Up",m_upicon,this,FXMovieManager::ID_MOVEUP);
	new FXMenuCommand(popupmenu,"Move Down",m_downicon,this,FXMovieManager::ID_MOVEDOWN);
	new FXMenuSeparator(popupmenu);
	new FXMenuCommand(popupmenu,"Select All",m_accepticon,this,FXMovieManager::ID_SELECTALL);
	new FXMenuCommand(popupmenu,"Remove All",m_cancelicon,this,FXMovieManager::ID_CLEAR);
	new FXMenuSeparator(popupmenu);
	new FXMenuCommand(popupmenu,"Append item",m_appendicon,this,FXMovieManager::ID_APPEND);
	new FXMenuCommand(popupmenu,"Insert item",m_appendicon,this,FXMovieManager::ID_INSERT);
	new FXMenuCommand(popupmenu,"Remove item",m_cancelicon,this,FXMovieManager::ID_DELETE);


	// Make file associations object; shared between FXFileList and FXDirList
	associations=new FXFileDict(getApp());

	// File box
  	splitter2=new FXSplitter(splitter1,LAYOUT_SIDE_TOP|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_FILL_Y|SPLITTER_TRACKING|SPLITTER_HORIZONTAL);

  	filebuttonframe_lp=new FXHorizontalFrame(splitter2,LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0,0,0);
  	filebuttonframe_rp=new FXHorizontalFrame(splitter2,LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0,0,0);

	vbuttonframe_lp=new FXVerticalFrame(filebuttonframe_lp,LAYOUT_FILL_Y|LAYOUT_SIDE_LEFT,0,0,0,0, 0,0,0,0, 0,0);
  	fileframe_lp=new FXVerticalFrame(filebuttonframe_lp,LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_THICK|FRAME_SUNKEN|LAYOUT_SIDE_RIGHT, 0,0,180,0, 0,0,0,0,0,0);

  	fileframe_rp=new FXVerticalFrame(filebuttonframe_rp,LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_THICK|FRAME_SUNKEN|LAYOUT_SIDE_LEFT, 0,0,180,0, 0,0,0,0,0,0);
	vbuttonframe_rp=new FXVerticalFrame(filebuttonframe_rp,LAYOUT_FILL_Y|LAYOUT_SIDE_RIGHT,0,0,0,0, 0,0,0,0, 0,0);

  	// left Switcher
  	leftpanel_tabbook=new FXTabBook(fileframe_lp,this,ID_PANEL,PACK_UNIFORM_WIDTH|PACK_UNIFORM_HEIGHT|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_SIDE_RIGHT);

  	// First item is a directory list
  	tab_dirlist_lp=new FXTabItem(leftpanel_tabbook,"&Directory List",NULL);
  	dirframe_tab_lp=new FXHorizontalFrame(leftpanel_tabbook,FRAME_THICK|FRAME_RAISED);
  	// Directory List
  	dirlist_lp=new FXDirList(dirframe_tab_lp,this,ID_DIRECTORYLIST_LP,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_RIGHT|TREELIST_SHOWS_LINES|TREELIST_SHOWS_BOXES|TREELIST_BROWSESELECT|DIRLIST_NO_OWN_ASSOC);
  	dirlist_lp->setAssociations(associations);
  	dirlist_lp->dropEnable();

  	// Second item is a file list
  	tab_filelist_lp=new FXTabItem(leftpanel_tabbook,"&File List",NULL);
  	fileframe_tab_lp=new FXHorizontalFrame(leftpanel_tabbook,FRAME_THICK|FRAME_RAISED);
	// left panel filelist_lp
	filelist_lp=new FXFileList(fileframe_tab_lp,this,ID_FILELIST_LP,LAYOUT_FILL_X|LAYOUT_FILL_Y|ICONLIST_BIG_ICONS|ICONLIST_AUTOSIZE|FILELIST_NO_OWN_ASSOC);
	filelist_lp->setAssociations(associations);
  	filelist_lp->dropEnable();


  	// right Switcher
  	rightpanel_tabbook=new FXTabBook(fileframe_rp,this,ID_PANEL,PACK_UNIFORM_WIDTH|PACK_UNIFORM_HEIGHT|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_SIDE_LEFT);

  	// First item is a directory list
  	tab_dirlist_rp=new FXTabItem(rightpanel_tabbook,"&Directory List",NULL);
  	dirframe_tab_rp=new FXHorizontalFrame(rightpanel_tabbook,FRAME_THICK|FRAME_RAISED);
  	// Directory List
  	dirlist_rp=new FXDirList(dirframe_tab_rp,this,ID_DIRECTORYLIST_RP,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_RIGHT|TREELIST_SHOWS_LINES|TREELIST_SHOWS_BOXES|TREELIST_BROWSESELECT|DIRLIST_NO_OWN_ASSOC);
  	dirlist_rp->setAssociations(associations);
  	dirlist_rp->dropEnable();

  	// Second item is a file list
  	tab_filelist_rp=new FXTabItem(rightpanel_tabbook,"&File List",NULL);
  	fileframe_tab_rp=new FXHorizontalFrame(rightpanel_tabbook,FRAME_THICK|FRAME_RAISED);
	// right panel filelist_rp
	filelist_rp=new FXFileList(fileframe_tab_rp,this,ID_FILELIST_RP,LAYOUT_FILL_X|LAYOUT_FILL_Y|ICONLIST_BIG_ICONS|ICONLIST_AUTOSIZE|FILELIST_NO_OWN_ASSOC);
	filelist_rp->setAssociations(associations);
  	filelist_rp->dropEnable();

	// Scanvirus dialog
    scanvirus=new ScanVirusDialog(this, "Virus Scanning Window");

  	// MimeType panel
  	mimetype=new MimeTypeDialog(this);

  	// Properties panel
  	properties_lp=new PropertiesDialog_lp(this);
  	properties_rp=new PropertiesDialog_rp(this);

  	// Status bar
  	statusbar=new FXHorizontalFrame(fileframe_lp,LAYOUT_SIDE_BOTTOM|FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X);
  	statusline=new FXStatusLine(statusbar,this,LAYOUT_FILL_X|LAYOUT_FILL_Y);

  	// Clock on status bar with same colour
  	clock=new FXTextField(statusbar,8,NULL,0,FRAME_SUNKEN|JUSTIFY_RIGHT|LAYOUT_RIGHT|LAYOUT_FILL_Y|TEXTFIELD_READONLY,0,0,0,0,2,2,1,1);
  	clock->setBackColor(statusbar->getBackColor());


	// left filelist buttons *************************************************
	new FXButton(vbuttonframe_lp,upbuttontext,upicon,this,ID_UPDIRECTORY_LP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);

  	selectbutton=new FXToggleButton(vbuttonframe_lp,allbuttontext1,allbuttontext2,showicon,hiddenicon,filelist_lp,FXFileList::ID_SELECT_INVERSE,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(vbuttonframe_lp,createbuttontext,createicon,this,ID_FILEPREVIEW_LP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(vbuttonframe_lp,recreatebuttontext,recreateicon,this,ID_FORCEFILEPREVIEW_LP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(vbuttonframe_lp,deletebuttontext,deletehiddenicon,this,ID_CLEARPREVIEW_LP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);

	sizeframe=new FXHorizontalFrame(vbuttonframe_lp,LAYOUT_FILL_X,0,0,0,0, 0,0,0,0, 0,0);
	new FXButton(sizeframe,increasebuttontext,increaseicon,this,ID_INCFILEPREVIEW_LP,BUTTON_TOOLBAR|LAYOUT_LEFT);
	new FXButton(sizeframe,decreasebuttontext,decreaseicon,this,ID_DECFILEPREVIEW_LP,BUTTON_TOOLBAR|LAYOUT_RIGHT);

  	widebutton=new FXToggleButton(vbuttonframe_lp,widebuttontext1,widebuttontext2,videoicon,videoicon,this,ID_WIDESCREEN,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);

	new FXMenuSeparator(vbuttonframe_lp);

	new FXButton(vbuttonframe_lp,stopbuttontext,stopicon,this,ID_FILESTOP_LP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(vbuttonframe_lp,pausebuttontext,pauseicon,this,ID_FILEPAUSE_LP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(vbuttonframe_lp,playbuttontext,playicon,this,ID_FILEPLAY_LP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);

	new FXMenuSeparator(vbuttonframe_lp);

	switchframe=new FXHorizontalFrame(vbuttonframe_lp,LAYOUT_FILL_X|LAYOUT_BOTTOM,0,0,0,0, 0,0,0,0, 0,0);
	new FXButton(switchframe,detlisttext,detlisticon,filelist_lp,FXFileList::ID_SHOW_DETAILS,BUTTON_TOOLBAR|LAYOUT_LEFT);
	new FXButton(switchframe,icnlisttext,icnlisticon,filelist_lp,FXFileList::ID_SHOW_BIG_ICONS,BUTTON_TOOLBAR|LAYOUT_RIGHT);

  	new FXToggleButton(vbuttonframe_lp,imgbuttontext2,imgbuttontext1,imgpreview2icon,imgpreview1icon,filelist_lp,FXFileList::ID_TOGGLE_IMAGES,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_BOTTOM);

	// right filelist buttons **********************************************
	new FXButton(vbuttonframe_rp,upbuttontext,upicon,this,ID_UPDIRECTORY_RP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);

  	selectbutton=new FXToggleButton(vbuttonframe_rp,allbuttontext1,allbuttontext2,showicon,hiddenicon,filelist_rp,FXFileList::ID_SELECT_INVERSE,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(vbuttonframe_rp,createbuttontext,createicon,this,ID_FILEPREVIEW_RP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(vbuttonframe_rp,recreatebuttontext,recreateicon,this,ID_FORCEFILEPREVIEW_RP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);

	new FXButton(vbuttonframe_rp,deletebuttontext,deletehiddenicon,this,ID_CLEARPREVIEW_RP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);

	sizeframe=new FXHorizontalFrame(vbuttonframe_rp,LAYOUT_FILL_X,0,0,0,0, 0,0,0,0, 0,0);
	new FXButton(sizeframe,increasebuttontext,increaseicon,this,ID_INCFILEPREVIEW_RP,BUTTON_TOOLBAR|LAYOUT_LEFT);
	new FXButton(sizeframe,decreasebuttontext,decreaseicon,this,ID_DECFILEPREVIEW_RP,BUTTON_TOOLBAR|LAYOUT_RIGHT);

	new FXMenuSeparator(vbuttonframe_rp);

	// seek value for movie preview
    seekspinner=new FXRealSpinner(vbuttonframe_rp,2,NULL,0,REALSPIN_CYCLIC|FRAME_SUNKEN|FRAME_THICK|LAYOUT_SIDE_TOP|LAYOUT_FILL_X);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
    seekspinner->setRange(0,99);
   	seekspinner->setValue(1);
    seekspinner->setIncrement(1);
#else
    seekspinner->setRange(0.0,9.1);
   	seekspinner->setValue(0.1);
    seekspinner->setIncrement(0.1);
#endif
	new FXMenuSeparator(vbuttonframe_rp);

	new FXButton(vbuttonframe_rp,stopbuttontext,stopicon,this,ID_FILESTOP_RP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(vbuttonframe_rp,pausebuttontext,pauseicon,this,ID_FILEPAUSE_RP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);
	new FXButton(vbuttonframe_rp,playbuttontext,playicon,this,ID_FILEPLAY_RP,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X);

	new FXMenuSeparator(vbuttonframe_rp);

	switchframe=new FXHorizontalFrame(vbuttonframe_rp,LAYOUT_FILL_X|LAYOUT_BOTTOM,0,0,0,0, 0,0,0,0, 0,0);
	new FXButton(switchframe,detlisttext,detlisticon,filelist_rp,FXFileList::ID_SHOW_DETAILS,BUTTON_TOOLBAR|LAYOUT_LEFT);
	new FXButton(switchframe,icnlisttext,icnlisticon,filelist_rp,FXFileList::ID_SHOW_BIG_ICONS,BUTTON_TOOLBAR|LAYOUT_RIGHT);

  	new FXToggleButton(vbuttonframe_rp,imgbuttontext2,imgbuttontext1,imgpreview2icon,imgpreview1icon,filelist_rp,FXFileList::ID_TOGGLE_IMAGES,BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_BOTTOM);

	// ****************************************************
  	// Add accelerators into the table
  	//		 addAccel(	 FXHotKey	  target	(FXSelector seldn, FXSelector selup)

  	getAccelTable()->addAccel(MKUINT(KEY_F1,0),this,FXSEL(SEL_COMMAND,ID_MIMETYPE));

	// CONTROLMASK --> Left Panel
  	getAccelTable()->addAccel(MKUINT(KEY_BackSpace,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_UPDIRECTORY_LP));	
  	getAccelTable()->addAccel(MKUINT(KEY_Delete,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILEDELETE_LP));
  	getAccelTable()->addAccel(MKUINT(KEY_F2,CONTROLMASK),filelist_lp,FXSEL(SEL_COMMAND,FXFileList::ID_SELECT_INVERSE));
  	getAccelTable()->addAccel(MKUINT(KEY_F3,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILEPREVIEW_LP));
  	getAccelTable()->addAccel(MKUINT(KEY_F4,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FORCEFILEPREVIEW_LP));
  	getAccelTable()->addAccel(MKUINT(KEY_F5,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_INCFILEPREVIEW_LP));
  	getAccelTable()->addAccel(MKUINT(KEY_F6,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_DECFILEPREVIEW_LP));
  	getAccelTable()->addAccel(MKUINT(KEY_F7,CONTROLMASK),filelist_lp,FXSEL(SEL_COMMAND,FXFileList::ID_SHOW_BIG_ICONS));
  	getAccelTable()->addAccel(MKUINT(KEY_F8,CONTROLMASK),filelist_lp,FXSEL(SEL_COMMAND,FXFileList::ID_SHOW_DETAILS));
 	// OPERATIONS DIALOG
	getAccelTable()->addAccel(MKUINT(KEY_r,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILERENAME_LP));
	getAccelTable()->addAccel(MKUINT(KEY_R,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILERENAME_LP));

  	getAccelTable()->addAccel(MKUINT(KEY_c,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILECOPY_LP));
  	getAccelTable()->addAccel(MKUINT(KEY_C,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILECOPY_LP));
	getAccelTable()->addAccel(MKUINT(KEY_m,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILEMOVE_LP));
	getAccelTable()->addAccel(MKUINT(KEY_M,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILEMOVE_LP));
	getAccelTable()->addAccel(MKUINT(KEY_s,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILESYMLINK_LP));
	getAccelTable()->addAccel(MKUINT(KEY_S,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILESYMLINK_LP));
	// PLAYER
	getAccelTable()->addAccel(MKUINT(KEY_z,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILEPLAY_LP));
	getAccelTable()->addAccel(MKUINT(KEY_Z,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILEPLAY_LP));
  	getAccelTable()->addAccel(MKUINT(KEY_a,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILEPAUSE_LP));	
  	getAccelTable()->addAccel(MKUINT(KEY_A,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILEPAUSE_LP));
   	getAccelTable()->addAccel(MKUINT(KEY_q,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILESTOP_LP));	
	getAccelTable()->addAccel(MKUINT(KEY_Q,CONTROLMASK),this,FXSEL(SEL_COMMAND,ID_FILESTOP_LP));


	// NONE     --> Right Panel 
  	getAccelTable()->addAccel(MKUINT(KEY_BackSpace,0),this,FXSEL(SEL_COMMAND,ID_UPDIRECTORY_RP));
	getAccelTable()->addAccel(MKUINT(KEY_Delete,0),this,FXSEL(SEL_COMMAND,ID_FILEDELETE_RP));
  	getAccelTable()->addAccel(MKUINT(KEY_F2,0),filelist_rp,FXSEL(SEL_COMMAND,FXFileList::ID_SELECT_INVERSE));
  	getAccelTable()->addAccel(MKUINT(KEY_F3,0),this,FXSEL(SEL_COMMAND,ID_FILEPREVIEW_RP));
  	getAccelTable()->addAccel(MKUINT(KEY_F4,0),this,FXSEL(SEL_COMMAND,ID_FORCEFILEPREVIEW_RP));
  	getAccelTable()->addAccel(MKUINT(KEY_F5,0),this,FXSEL(SEL_COMMAND,ID_INCFILEPREVIEW_RP));
  	getAccelTable()->addAccel(MKUINT(KEY_F6,0),this,FXSEL(SEL_COMMAND,ID_DECFILEPREVIEW_RP));
  	getAccelTable()->addAccel(MKUINT(KEY_F7,0),filelist_rp,FXSEL(SEL_COMMAND,FXFileList::ID_SHOW_BIG_ICONS));
  	getAccelTable()->addAccel(MKUINT(KEY_F8,0),filelist_rp,FXSEL(SEL_COMMAND,FXFileList::ID_SHOW_DETAILS));
 	// OPERATIONS DIALOG
	getAccelTable()->addAccel(MKUINT(KEY_r,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILERENAME_RP));
	getAccelTable()->addAccel(MKUINT(KEY_R,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILERENAME_RP));

  	getAccelTable()->addAccel(MKUINT(KEY_c,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILECOPY_RP));
  	getAccelTable()->addAccel(MKUINT(KEY_C,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILECOPY_RP));
	getAccelTable()->addAccel(MKUINT(KEY_m,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILEMOVE_RP));
	getAccelTable()->addAccel(MKUINT(KEY_M,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILEMOVE_RP));
	getAccelTable()->addAccel(MKUINT(KEY_s,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILESYMLINK_RP));
	getAccelTable()->addAccel(MKUINT(KEY_S,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILESYMLINK_RP));
	// PLAYER
	getAccelTable()->addAccel(MKUINT(KEY_z,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILEPLAY_RP));
	getAccelTable()->addAccel(MKUINT(KEY_Z,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILEPLAY_RP));
  	getAccelTable()->addAccel(MKUINT(KEY_a,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILEPAUSE_RP));	
  	getAccelTable()->addAccel(MKUINT(KEY_A,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILEPAUSE_RP));
   	getAccelTable()->addAccel(MKUINT(KEY_q,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILESTOP_RP));	
	getAccelTable()->addAccel(MKUINT(KEY_Q,ALTMASK),this,FXSEL(SEL_COMMAND,ID_FILESTOP_RP));


	// Progress bar
  	progressframe=new FXHorizontalFrame(fileframe_rp,LAYOUT_FILL_X|FRAME_THICK|FRAME_SUNKEN, 0,0,0,0, 0,0,0,0,0,0);
    progressbar=new FXProgressBar(progressframe,NULL,0,PROGRESSBAR_HORIZONTAL|PROGRESSBAR_PERCENTAGE|PROGRESSBAR_NORMAL|LAYOUT_FILL_X);


  	// Make a tool tip
  	tooltip=new FXToolTip(getApp(),0);

	// inithialize some parameters
	FXString currentUser = FXSystem::currentUserName();
	printf("the Current User is: %s\n", currentUser.text());
	FXString currentTemp = FXSystem::getTempDirectory();
	printf("the Temp Directory is: %s\n", currentTemp.text());
	FXString createTempDir="test -d "+currentTemp+"/.fxmm-"+currentUser+" || mkdir "+currentTemp+"/.fxmm-"+currentUser;
	system (createTempDir.text());

    currentInfo=currentTemp+"/.fxmm-"+currentUser+"/info.txt";
	currentFifo=currentTemp+"/.fxmm-"+currentUser+"/fifo";
	removeTempFifo="test -p "+currentFifo+" && rm "+currentFifo;
	createTempFifo="test -p "+currentFifo+" || mkfifo "+currentFifo;
	system (createTempFifo.text());

	s_command = "mplayer -slave -idle -quiet -ontop -input file="+currentFifo+" ";
	language = "en";
	isrunning = FALSE;
	widescreen = FALSE;
	imgw=80; imgh=64;

  	FXString init_command="mplayer -vo help";
  	system (init_command.text());  // initialize lazy MPlayer
}

FXMovieManager::~FXMovieManager() {

	delete audioicon;
	delete createicon; 
	delete decreaseicon; 
	delete deleteicon; 
	delete deletehiddenicon; 
	delete downicon;
	delete dropicon; 
	delete hiddenicon;
	delete hideicon; 
	delete increaseicon; 
	delete loopicon;
	delete openicon; 
	delete pauseicon;
	delete playicon;
	delete quiticon;
	delete recreateicon; 
	delete showicon; 
	delete stopicon;
	delete upicon;
	delete videoicon;
	delete videozoomicon;
	delete videosizeicon;
	delete videoouticon;
	delete icnlisticon;
	delete detlisticon;

	delete m_bugicon;
	delete m_upicon;
	delete m_downicon;
	delete m_appendicon;
	delete m_accepticon;
	delete m_cancelicon;
	delete m_archiveicon;
	delete m_mapicon;
	delete m_unmapicon;
	delete m_copyicon;
	delete m_moveicon;
	delete m_renameicon;
	delete m_deleteicon;
	delete m_propicon;
	delete m_consoleicon;
	delete m_terminalicon;
    delete m_refreshicon;
    delete m_renamediricon;
    delete m_showhiddenicon;

  	getApp()->removeTimeout(this,ID_CLOCKTIME);
}



/*********************** UPDATE CLOCK **************************/

// Update clock
long FXMovieManager::onClock(FXObject*,FXSelector,void*){
#ifdef HAVE_LIBFOX_1_7
  FXTime now=FXThread::time();
  // show time
  clock->setText(FXSystem::localTime(tr("%H:%M:%S"),now));
  // tooltip date
  clock->setTipText(FXSystem::localTime(tr("%A %d %B %Y"),now));
  getApp()->addTimeout(this,ID_CLOCKTIME,CLOCKTIMER);
#endif
  return 0;
  }


/*********************** ROUTINES ******************************/

// Auto Show Thumbnails option
void FXMovieManager::showFilelistThumbnails_lp() {
    showThumbnails(filelist_lp);
    return;
}

void FXMovieManager::showFilelistThumbnails_rp() {
    showThumbnails(filelist_rp);
    return;
}

void FXMovieManager::showThumbnails(FXFileList* panel) {
    FXString file,filetoplay;
    for(FXint i=0; i<panel->getNumItems(); i++){

       	file=panel->getItemFilename(i);
        if(file=="..") continue;
		if(panel->isItemDirectory(i)) continue;
    	if(loadfilevideo(file)==FALSE) continue;
		if(autoThumbnailShow==TRUE){
			FXString rootname,imagename;
        	rootname="."+FXPath::stripExtension(file);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        	imagename=FXPath::enquote(panel->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        	imagename=FXPath::enquote(panel->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
    		FXFile truefile;
			FXbool istruefile = truefile.open(imagename);
			if(autoThumbnailCreate==TRUE){
				if(!istruefile){
        			filetoplay=FXPath::absolute(panel->getDirectory(),file);
					if(!createImageMovie(filetoplay,rootname)){
						FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Creation Failure","Error on creating thumbnail: %s",imagename.text());
  	    				if(answer==MBOX_CLICKED_CANCEL){
	      					return;
	    				}
					}
					if(!previewMovie(panel, imagename, i, imgw, imgh)){
						FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
  	    				if(answer==MBOX_CLICKED_CANCEL){
	      					return;
	    				}
					}
				}
			}
			if(istruefile){
				if(!previewMovie(panel, imagename, i, imgw, imgh)){
					FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
  	    			if(answer==MBOX_CLICKED_CANCEL){
	      				return;
	    			}
				}
			}
			else{
				continue;
			}
    	}
    }
    return;
}


// Load video files if extension match
FXbool FXMovieManager::loadfilevideo(const FXString& filename){
	FXString ext=FXPath::extension(filename);

	if(comparecase(ext,"mpeg")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MPEG")==0){
		return TRUE;
	}
	else if(comparecase(ext,"mpg")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MPG")==0){
		return TRUE;
	}
	else if(comparecase(ext,"mpe")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MPE")==0){
		return TRUE;
	}

	else if(comparecase(ext,"wmv")==0){
		return TRUE;
	}
	else if(comparecase(ext,"WMV")==0){
		return TRUE;
	}
	else if(comparecase(ext,"wmx")==0){
		return TRUE;
	}
	else if(comparecase(ext,"WMX")==0){
		return TRUE;
	}
	else if(comparecase(ext,"asf")==0){
		return TRUE;
	}
	else if(comparecase(ext,"ASF")==0){
		return TRUE;
	}
	else if(comparecase(ext,"asx")==0){
		return TRUE;
	}
	else if(comparecase(ext,"ASX")==0){
		return TRUE;
	}

	else if(comparecase(ext,"avi")==0){
		return TRUE;
	}
	else if(comparecase(ext,"AVI")==0){
		return TRUE;
	}

	else if(comparecase(ext,"mp4")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MP4")==0){
		return TRUE;
	}
	else if(comparecase(ext,"mov")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MOV")==0){
		return TRUE;
	}
	else if(comparecase(ext,"mod")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MOD")==0){
		return TRUE;
	}
	else if(comparecase(ext,"qt")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"QT")==0){
		return TRUE;
	}

	else if(comparecase(ext,"3gp")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"3GP")==0){
		return TRUE;
	}
	else if(comparecase(ext,"3gp2")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"3GP2")==0){
		return TRUE;
	}

	else if(comparecase(ext,"ogv")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"OGV")==0){
		return TRUE;
	}
	else if(comparecase(ext,"ogm")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"OGM")==0){
		return TRUE;
	}
	else if(comparecase(ext,"ogg")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"OGG")==0){
		return TRUE;
	}

	else if(comparecase(ext,"vob")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"VOB")==0){
		return TRUE;
	}

	else if(comparecase(ext,"flv")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"FLV")==0){
		return TRUE;
	}
	else if(comparecase(ext,"fli")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"FLI")==0){
		return TRUE;
	}

	else if(comparecase(ext,"gl")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"GL")==0){
		return TRUE;
	}
	else if(comparecase(ext,"dl")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"DL")==0){
		return TRUE;
	}

	else if(comparecase(ext,"dif")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"DIF")==0){
		return TRUE;
	}
	else if(comparecase(ext,"dv")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"DV")==0){
		return TRUE;
	}

	else if(comparecase(ext,"movie")==0){ 
		return TRUE;
	}
	else if(comparecase(ext,"MOVIE")==0){
		return TRUE;
	}
	else {
		//FXMessageBox::error(this,MBOX_OK,"Error on loading Movie","Unsupported Type: %s",ext.text());
		return FALSE;
	}
	return FALSE;
}


// Load audio files if extension match
FXbool FXMovieManager::loadfileaudio(const FXString& filename){
	FXString ext=FXPath::extension(filename);

	if(comparecase(ext,"au")==0){
		return TRUE;
	}
	else if(comparecase(ext,"AU")==0){
		return TRUE;
	}
	else if(comparecase(ext,"snd")==0){
		return TRUE;
	}
	else if(comparecase(ext,"SND")==0){
		return TRUE;
	}

	else if(comparecase(ext,"mp2")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MP2")==0){
		return TRUE;
	}
	else if(comparecase(ext,"mp3")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MP3")==0){
		return TRUE;
	}
	else if(comparecase(ext,"mpga")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MPGA")==0){
		return TRUE;
	}
	else if(comparecase(ext,"mpega")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MPEGA3")==0){
		return TRUE;
	}

	else if(comparecase(ext,"ra")==0){
		return TRUE;
	}
	else if(comparecase(ext,"RA")==0){
		return TRUE;
	}
	else if(comparecase(ext,"rm")==0){
		return TRUE;
	}
	else if(comparecase(ext,"RM")==0){
		return TRUE;
	}
	else if(comparecase(ext,"ram")==0){
		return TRUE;
	}
	else if(comparecase(ext,"RAM")==0){
		return TRUE;
	}

	else if(comparecase(ext,"wav")==0){
		return TRUE;
	}
	else if(comparecase(ext,"WAV")==0){
		return TRUE;
	}

	else if(comparecase(ext,"wma")==0){
		return TRUE;
	}
	else if(comparecase(ext,"WMA")==0){
		return TRUE;
	}
	else if(comparecase(ext,"wax")==0){
		return TRUE;
	}
	else if(comparecase(ext,"WAX")==0){
		return TRUE;
	}

	else if(comparecase(ext,"mod")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MOD")==0){
		return TRUE;
	}

	else if(comparecase(ext,"mid")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MID")==0){
		return TRUE;
	}
	else if(comparecase(ext,"midi")==0){
		return TRUE;
	}
	else if(comparecase(ext,"MIDI")==0){
		return TRUE;
	}

	else {
		//FXMessageBox::error(this,MBOX_OK,"Error on loading Movie","Unsupported Type: %s",ext.text());
		return FALSE;
	}
	return FALSE;
}


// Load imagemovie in the play-TABLE list for preview
FXbool FXMovieManager::previewListMovie(const FXString& filename, FXint i){
    FXString imagename=filename;
    //printf("the current directory is: %s\n", filename.text());

#ifdef HAVE_LIBFOX_1_6
    FXIcon *tempicon=NULL;
    tempicon=new FXIcon(getApp(),NULL,IMAGE_OPAQUE|IMAGE_KEEP);
#endif
#ifdef HAVE_LIBFOX_1_7
    FXJPGIcon *tempicon=NULL;
    tempicon=new FXJPGIcon(getApp(),NULL,IMAGE_OPAQUE|IMAGE_KEEP);
#endif

    if (tempicon){

   		FXFileStream stream;
   		if (stream.open(imagename,FXStreamLoad)){
			getApp()->beginWaitCursor();
			tempicon->loadPixels(stream);
			stream.close();

			tempicon->scale(80,64,75); 
			tempicon->create();

			table->setRowIcon(i,tempicon);

			getApp()->endWaitCursor();

			return TRUE;
		}
    }
    return FALSE;
}


// open filedialog and select multiple files
long FXMovieManager::onCmdFile(FXObject *,FXSelector,void *) {
	resetTable();
	appendFiledialogSelection();
	selectAll();
	return 1;
}


// INSERT selected movies from the Filedialog into the TABLE
void FXMovieManager::insertFiledialogSelection(){
	FXint r,c;
	FXString rootname, imagename;
	currentdir_lp = dirlist_lp->getDirectory();

	FXString* files=FXFileDialog::getOpenFilenames(this,"Select the Movies to Append",currentdir_lp,moviefiles,1);
	if(files){
		for(int i=0; !files[i].empty(); i++){
			filename = files[i];
			FXString ext=FXPath::extension(filename);
			
			if(loadfilevideo(filename)==TRUE){
				numtableitems=numtableitems+1;

				// Return current row index, if any
				r = table->getCurrentRow();
				if(r >= 0){
					table->insertRows(r);
  					table->setRowHeight(r,68);
					table->setItemText(r,0,filename);
					for (c=1; c<9; c++){
	    					table->setItemText(r,c,"00");
					}
  					table->setItemText(r,5,"10");
  					table->removeRows(100);
				}
				else if(r<0){
					FXMessageBox::error(this,MBOX_OK,"Error inserting Movie","The playlist is empty\nuse append item instead");
					continue;
				}

  				rootname="."+FXPath::name(FXPath::stripExtension(filename));
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        		imagename=FXPath::enquote(FXPath::directory(filename)+"/"+rootname+"_thumbnail.jpg");
#else
        		imagename=FXPath::enquote(FXPath::directory(filename)+"/"+rootname+"00000001/00000001.jpg");
#endif
				/* 
    				printf("the rootname is: %s\n", rootname.text());
    				printf("the imagename is: %s\n", imagename.text());
        			printf("the itemnumber is: %d\n", numtableitems-1);
				*/
  				if(!previewListMovie(imagename, r)){
					FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
  	    			if(answer==MBOX_CLICKED_CANCEL){
	      				return;
	    			}
				}
			}
			else if(loadfilevideo(filename)==FALSE){
				FXMessageBox::error(this,MBOX_OK,"Error on loading Movie","Unsupported Type: %s",ext.text());
			}
		}
		// clean up files
		delete [] files;
	}
	return;
}


// APPEND selected movies from the Filedialog to the TABLE
void FXMovieManager::appendFiledialogSelection(){
	FXString rootname, imagename;
	currentdir_lp = dirlist_lp->getDirectory();

	FXString* files=FXFileDialog::getOpenFilenames(this,"Select the Movies to Append",currentdir_lp,moviefiles,1);
	if(files){
		for(int i=0; !files[i].empty(); i++){
			filename = files[i];
			FXString ext=FXPath::extension(filename);

			if(loadfilevideo(filename)==TRUE){
				numtableitems=numtableitems+1;
				table->setItemText(numtableitems-1,0,filename);

  				rootname="."+FXPath::name(FXPath::stripExtension(filename));
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        			imagename=FXPath::enquote(FXPath::directory(filename)+"/"+rootname+"_thumbnail.jpg");
#else
        			imagename=FXPath::enquote(FXPath::directory(filename)+"/"+rootname+"00000001/00000001.jpg");
#endif
				/* 
    				printf("the rootname is: %s\n", rootname.text());
    				printf("the imagename is: %s\n", imagename.text());
        			printf("the itemnumber is: %d\n", numtableitems-1);
				*/
  				if(!previewListMovie(imagename, numtableitems-1)){
					FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
  	    			if(answer==MBOX_CLICKED_CANCEL){
	      				return;
	    			}
				}
			}
			else if(loadfilevideo(filename)==FALSE){
				FXMessageBox::error(this,MBOX_OK,"Error on loading Movie","Unsupported Type: %s",ext.text());
			}
		}
		// clean up files
		delete [] files;
	}
	return;
}


// get selected movies from filelist_lp
void FXMovieManager::getFilelistSelection_lp() {
  // load one file at a time
  FXString ffile,rootname,imagename;
  for(FXint i=0; i<filelist_lp->getNumItems(); i++){
    	if(filelist_lp->isItemSelected(i)){
        	ffile=filelist_lp->getItemFilename(i);
        	if(ffile=="..") continue;

  			rootname="."+FXPath::stripExtension(ffile);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        	imagename=FXPath::enquote(filelist_lp->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        	imagename=FXPath::enquote(filelist_lp->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
    		//printf("the imagename is: %s\n", imagename.text());
			if(!previewListMovie(imagename, numtableitems-1)){
					FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
  	    			if(answer==MBOX_CLICKED_CANCEL){
	      				return;
	    			}
			}
		//printf("the numitems is: %d\n", numtableitems-1);
	}
  }
  return;	
}

// get selected movies from filelist_rp
void FXMovieManager::getFilelistSelection_rp() {
  // load one file at a time
  FXString ffile,rootname,imagename;
  for(FXint i=0; i<filelist_rp->getNumItems(); i++){
    	if(filelist_rp->isItemSelected(i)){
        	ffile=filelist_rp->getItemFilename(i);
        	if(ffile=="..") continue;

  			rootname="."+FXPath::stripExtension(ffile);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        	imagename=FXPath::enquote(filelist_rp->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        	imagename=FXPath::enquote(filelist_rp->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
    		//printf("the imagename is: %s\n", imagename.text());
			if(!previewListMovie(imagename, numtableitems-1)){
					FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
  	    			if(answer==MBOX_CLICKED_CANCEL){
	      				return;
	    			}
			}
			//printf("the numitems is: %d\n", numtableitems-1);
		}
  }
  return;	
}

// reset Table to default values
void FXMovieManager::resetTable() {
	table->killSelection();
	numtableitems=0;
	FXint c, r;
	for (c=1; c<9; c++){
	  for (r=0; r<100; r++){
	    table->setRowIcon(r,NULL);
	    table->setItemText(r,0,"");
	    table->setItemText(r,c,"00");
	  }
	}
	for (r=0; r<100; r++){
	  table->setItemText(r,5,"10");
	}
	return;
}



// load theme 
FXbool FXMovieManager::themeSelection() {
  theme=getApp()->reg().readStringEntry("SETTINGS","theme");

	if(! (theme.length()>0)) { theme="dark";}
  	if(! (theme_fxmmfore=getApp()->reg().readColorEntry("SETTINGS","fxmmforecolor"))) { theme_fxmmfore   = FXRGB(180,180,180);}
  	if(! (theme_fxmmback=getApp()->reg().readColorEntry("SETTINGS","fxmmbackcolor"))) { theme_fxmmback   = FXRGB( 20, 20, 20);}
  	if(! (theme_fxmmbase=getApp()->reg().readColorEntry("SETTINGS","fxmmbasecolor"))) { theme_fxmmbase   = FXRGB( 60, 60, 60);}
  	if(! (theme_fxmmborder=getApp()->reg().readColorEntry("SETTINGS","fxmmbordercolor"))) { theme_fxmmborder = FXRGB(  0,  0,  0);}
  	if(! (theme_fxmmmenufore=getApp()->reg().readColorEntry("SETTINGS","fxmmmenuforecolor"))) { theme_fxmmmenufore  = FXRGB(240,240,240);}
  	if(! (theme_fxmmmenuback=getApp()->reg().readColorEntry("SETTINGS","fxmmmenubackcolor"))) { theme_fxmmmenuback  = FXRGB( 10, 10, 10);}
  	if(! (theme_fxmmselfore=getApp()->reg().readColorEntry("SETTINGS","fxmmselforecolor"))) { theme_fxmmselfore   = FXRGB(  0,  0,  0);}
  	if(! (theme_fxmmselback=getApp()->reg().readColorEntry("SETTINGS","fxmmselbackcolor"))) { theme_fxmmselback   = FXRGB(140,140,140);}
  	if(! (theme_fxmmtipfore=getApp()->reg().readColorEntry("SETTINGS","fxmmtipforecolor"))) { theme_fxmmtipfore  = FXRGB(220,220,220);}
  	if(! (theme_fxmmtipback=getApp()->reg().readColorEntry("SETTINGS","fxmmtipbackcolor"))) { theme_fxmmtipback  = FXRGB( 20, 20, 20);}
  	if(! (theme_fxmmhilite=getApp()->reg().readColorEntry("SETTINGS","fxmmhilitecolor"))) { theme_fxmmhilite  = FXRGB(100,100,100);}
  	if(! (theme_fxmmshadow=getApp()->reg().readColorEntry("SETTINGS","fxmmshadowcolor"))) { theme_fxmmshadow  = FXRGB( 10, 10, 10);}


  	getApp()->setForeColor(theme_fxmmfore);
  	getApp()->setBaseColor(theme_fxmmbase);
  	getApp()->setBackColor(theme_fxmmback);
  	getApp()->setBorderColor(theme_fxmmborder);
  	getApp()->setSelMenuTextColor(theme_fxmmmenufore);
  	getApp()->setSelMenuBackColor(theme_fxmmmenuback);
  	getApp()->setSelforeColor(theme_fxmmselfore);
  	getApp()->setSelbackColor(theme_fxmmselback);
  	getApp()->setTipforeColor(theme_fxmmtipfore);
  	getApp()->setTipbackColor(theme_fxmmtipback);
	getApp()->setHiliteColor(theme_fxmmhilite);
	getApp()->setShadowColor(theme_fxmmshadow);


  if(theme=="dark"||theme=="gentle"){
	themeDG();
  	return TRUE;
  }else if(theme=="shine"||theme=="fog"){
	themeSF();
  	return TRUE;
  }else if(theme=="classic"||theme=="moon"){
	themeCM();
  	return TRUE;
  }else if(theme=="pink"||theme=="blood"){
	themePB();
  	return TRUE;
  }else{
	theme="dark";
	themeDG();
  	return TRUE;
  } 
  return FALSE;
}


void FXMovieManager::themeDG() {
	audioicon=new FXGIFIcon(getApp(),audio_32x32);
	createicon=new FXGIFIcon(getApp(),create_32x32); 
	deleteicon=new FXGIFIcon(getApp(),delete_32x32); 
	deletehiddenicon=new FXGIFIcon(getApp(),deletehidden_32x32); 
	downicon=new FXGIFIcon(getApp(),down_32x32);
	dropicon=new FXGIFIcon(getApp(),drop_32x32); 
	hiddenicon=new FXGIFIcon(getApp(),hidden_32x32);
	hideicon=new FXGIFIcon(getApp(),hide_32x32); 
	loopicon=new FXGIFIcon(getApp(),loop_32x32);
	openicon=new FXGIFIcon(getApp(),open_32x32); 
	pauseicon=new FXGIFIcon(getApp(),pause_32x32);
	playicon=new FXGIFIcon(getApp(),play_32x32);
	quiticon=new FXGIFIcon(getApp(),quit_32x32);
	recreateicon=new FXGIFIcon(getApp(),recreate_32x32); 
	showicon=new FXGIFIcon(getApp(),show_32x32); 
	stopicon=new FXGIFIcon(getApp(),stop_32x32);
	upicon=new FXGIFIcon(getApp(),up_32x32);
	videoicon=new FXGIFIcon(getApp(),video_32x32);
	videoouticon=new FXGIFIcon(getApp(),videoout_32x32);
	videosizeicon=new FXGIFIcon(getApp(),videosize_32x32);
	videozoomicon=new FXGIFIcon(getApp(),videozoom_32x32);
	imgpreview1icon=new FXGIFIcon(getApp(),imgtrue_32x32);
	imgpreview2icon=new FXGIFIcon(getApp(),imgfalse_32x32);
	decreaseicon=new FXGIFIcon(getApp(),decrease_16x32); 
	increaseicon=new FXGIFIcon(getApp(),increase_16x32); 
	icnlisticon=new FXGIFIcon(getApp(),icnlist_16x32);
	detlisticon=new FXGIFIcon(getApp(),detlist_16x32);
	return;
}

void FXMovieManager::themeSF() {
	audioicon=new FXGIFIcon(getApp(),raudio_32x32);
	createicon=new FXGIFIcon(getApp(),rcreate_32x32); 
	deleteicon=new FXGIFIcon(getApp(),rdelete_32x32); 
	deletehiddenicon=new FXGIFIcon(getApp(),rdeletehidden_32x32); 
	downicon=new FXGIFIcon(getApp(),rdown_32x32);
	dropicon=new FXGIFIcon(getApp(),rdrop_32x32); 
	hiddenicon=new FXGIFIcon(getApp(),rhidden_32x32);
	hideicon=new FXGIFIcon(getApp(),rhide_32x32); 
	loopicon=new FXGIFIcon(getApp(),rloop_32x32);
	openicon=new FXGIFIcon(getApp(),ropen_32x32); 
	pauseicon=new FXGIFIcon(getApp(),rpause_32x32);
	playicon=new FXGIFIcon(getApp(),rplay_32x32);
	quiticon=new FXGIFIcon(getApp(),rquit_32x32);
	recreateicon=new FXGIFIcon(getApp(),rrecreate_32x32); 
	showicon=new FXGIFIcon(getApp(),rshow_32x32); 
	stopicon=new FXGIFIcon(getApp(),rstop_32x32);
	upicon=new FXGIFIcon(getApp(),rup_32x32);
	videoicon=new FXGIFIcon(getApp(),rvideo_32x32);
	videoouticon=new FXGIFIcon(getApp(),rvideoout_32x32);
	videosizeicon=new FXGIFIcon(getApp(),rvideosize_32x32);
	videozoomicon=new FXGIFIcon(getApp(),rvideozoom_32x32);
	imgpreview1icon=new FXGIFIcon(getApp(),rimgtrue_32x32);
	imgpreview2icon=new FXGIFIcon(getApp(),rimgfalse_32x32);
	decreaseicon=new FXGIFIcon(getApp(),rdecrease_16x32); 
	increaseicon=new FXGIFIcon(getApp(),rincrease_16x32); 
	icnlisticon=new FXGIFIcon(getApp(),ricnlist_16x32);
	detlisticon=new FXGIFIcon(getApp(),rdetlist_16x32);
	return;
}

void FXMovieManager::themePB() {
	audioicon=new FXGIFIcon(getApp(),naudio_32x32);
	createicon=new FXGIFIcon(getApp(),ncreate_32x32); 
	deleteicon=new FXGIFIcon(getApp(),ndelete_32x32); 
	deletehiddenicon=new FXGIFIcon(getApp(),ndeletehidden_32x32); 
	downicon=new FXGIFIcon(getApp(),ndown_32x32);
	dropicon=new FXGIFIcon(getApp(),ndrop_32x32); 
	hiddenicon=new FXGIFIcon(getApp(),nhidden_32x32);
	hideicon=new FXGIFIcon(getApp(),nhide_32x32); 
	loopicon=new FXGIFIcon(getApp(),nloop_32x32);
	openicon=new FXGIFIcon(getApp(),nopen_32x32); 
	pauseicon=new FXGIFIcon(getApp(),npause_32x32);
	playicon=new FXGIFIcon(getApp(),nplay_32x32);
	quiticon=new FXGIFIcon(getApp(),nquit_32x32);
	recreateicon=new FXGIFIcon(getApp(),nrecreate_32x32); 
	showicon=new FXGIFIcon(getApp(),nshow_32x32); 
	stopicon=new FXGIFIcon(getApp(),nstop_32x32);
	upicon=new FXGIFIcon(getApp(),nup_32x32);
	videoicon=new FXGIFIcon(getApp(),nvideo_32x32);
	videoouticon=new FXGIFIcon(getApp(),nvideoout_32x32);
	videosizeicon=new FXGIFIcon(getApp(),nvideosize_32x32);
	videozoomicon=new FXGIFIcon(getApp(),nvideozoom_32x32);
	imgpreview1icon=new FXGIFIcon(getApp(),nimgtrue_32x32);
	imgpreview2icon=new FXGIFIcon(getApp(),nimgfalse_32x32);
	decreaseicon=new FXGIFIcon(getApp(),ndecrease_16x32); 
	increaseicon=new FXGIFIcon(getApp(),nincrease_16x32); 
	icnlisticon=new FXGIFIcon(getApp(),nicnlist_16x32);
	detlisticon=new FXGIFIcon(getApp(),ndetlist_16x32);
	return;
}

void FXMovieManager::themeCM() {
	audioicon=new FXGIFIcon(getApp(),b_audio);
	createicon=new FXGIFIcon(getApp(),b_previd);
	decreaseicon=new FXGIFIcon(getApp(),b_decvid);
	deleteicon=new FXGIFIcon(getApp(),b_trash);
	deletehiddenicon=new FXGIFIcon(getApp(),b_delvid);
	downicon=new FXGIFIcon(getApp(),b_down);
	dropicon=new FXGIFIcon(getApp(),b_conf);
	hiddenicon=new FXGIFIcon(getApp(),b_hidden);
	hideicon=new FXGIFIcon(getApp(),b_hide);
	increaseicon=new FXGIFIcon(getApp(),b_incvid);
	loopicon=new FXGIFIcon(getApp(),b_conf);
	openicon=new FXGIFIcon(getApp(),b_file);
	pauseicon=new FXGIFIcon(getApp(),b_pause);
	playicon=new FXGIFIcon(getApp(),b_play);
	quiticon=new FXGIFIcon(getApp(),b_quit);
	recreateicon=new FXGIFIcon(getApp(),b_frevid);
	showicon=new FXGIFIcon(getApp(),b_all);
	stopicon=new FXGIFIcon(getApp(),b_stop);
	upicon=new FXGIFIcon(getApp(),b_up);
	videoicon=new FXGIFIcon(getApp(),b_video);
	videoouticon=new FXGIFIcon(getApp(),b_videoout);
	videosizeicon=new FXGIFIcon(getApp(),b_videosize);
	videozoomicon=new FXGIFIcon(getApp(),b_videozoom);
	imgpreview1icon=new FXGIFIcon(getApp(),b_imgtrue);
	imgpreview2icon=new FXGIFIcon(getApp(),b_imgfalse);
	decreaseicon=new FXGIFIcon(getApp(),b_decvid); 
	increaseicon=new FXGIFIcon(getApp(),b_incvid); 
	icnlisticon=new FXGIFIcon(getApp(),b_icnlist);
	detlisticon=new FXGIFIcon(getApp(),b_detlist);
	return;
}


// load localization 
FXbool FXMovieManager::languageSelection() {
  language=getApp()->reg().readStringEntry("SETTINGS","language");

  if(language=="en"){
	languageEN();
  	return TRUE;
  }else if(language=="it"){
	languageIT();
  	return TRUE;
  }else if(language=="fr"){
	languageFR();
  	return TRUE;
  }else if(language=="de"){
	languageDE();
  	return TRUE;
  }else if(language=="es"){
	languageES();
  	return TRUE;
  }else if(language=="ru"){
	languageRU();
  	return TRUE;
  }else if(language=="cn"){
	languageCN();
  	return TRUE;
  }else if(language=="jp"){
	languageJP();
  	return TRUE;
  }else{
	language="en";
	languageEN();
  	return TRUE;
  } 
  return FALSE;
}


void FXMovieManager::languageEN() {
	quitbuttontext = "\tQuit\tQuit the application";
	openbuttontext = "\tAdd movies\tAdd movies to the playlist";
	audiobuttontext1 = "\tMute\tMPlayer must be running";
	audiobuttontext2 = "\tAudio\tMPlayer must be running";
	dropbuttontext1 = "\tDrop frames\tMPlayer must be running";
	dropbuttontext2 = "\tAll frames\tMPlayer must be running";
	stopbuttontext = "\tStop\tStop Player";
	pausebuttontext = "\tPause\tMPlayer must be running";
	playbuttontext = "\tPlay\tPlay the selected movies";
	loopbuttontext1 = "\tLoop\tPlay in loop";
	loopbuttontext2 = "\tOnce\tPlay once";
	zoombuttontext1 = "\tZoom\tZoom video size";
	zoombuttontext2 = "\t1:1\tReal video size";
	hidebuttontext1 = "\tHide\tHide playlist";
	hidebuttontext2 = "\tShow\tShow playlist";
	sizebuttontext = "\tSize\tVideo size";
	outputbuttontext = "\tOutput\tVideo output";
	confbuttontext = "\tConfiguration\tFxMovieManager Configuration";
	upbuttontext = "\tMove one step up\tMove one step up";
	downbuttontext = "\tMove one step down\tMove one step down";
	slistbuttontext = "|_|\tSave playlist\tSave list of movies";
	olistbuttontext = "||>\tLoad playlist\tLoad list of movies";
  	allbuttontext1 = "\tSelect all movies\tSelect all movies";
	allbuttontext2 = "\tDeselect all movies\tDeselect all movies";
	createbuttontext = "\tCreate thumbnails\tCreate thumbnails";
	recreatebuttontext = "\tRecreate thumbnails\tRecreate thumbnails";
  	widebuttontext1 = "\tWide Screen\tShow Wide Screen Thumbnails";
	widebuttontext2 = "\tNormal Screen\tShow Normal Screen Thumbnails";
	deletebuttontext = "\tDelete thumbnails\tDelete thumbnails";
	clearbuttontext = "\tClear the playlist\tClear the playlist";
	increasebuttontext = "\tIncrease thumbnails size\tIncrease thumbnails size";
	decreasebuttontext = "\tDecrease thumbnails size\tDecrease thumbnails size";
	detlisttext = "\tdetailed list view\tdetailed list view";
	icnlisttext = "\tIcon list view\tIcon list view";
	imgbuttontext1 = "\tDon't show thumbnail images\tDon't show thumbnail images";
	imgbuttontext2 = "\tShow thumbnail images\tShow thumbnail images";
	return;
}

void FXMovieManager::languageIT() {
	quitbuttontext = "\tEsci\tEsci dal programma";
	openbuttontext = "\tAggiungi\tAggiungi films alla lista";
	audiobuttontext1 = "\tMuto\tMPlayer deve essere in esecuzione";
	audiobuttontext2 = "\tAudio\tMPlayer deve essere in esecuzione";
	dropbuttontext1 = "\tSfoltisci i frames\tMPlayer deve essere in esecuzione";
	dropbuttontext2 = "\tTutti i frames\tMPlayer deve essere in esecuzione";
	stopbuttontext = "\tInterrompi\tInterrompi";
	pausebuttontext = "\tPausa\tMPlayer deve essere in esecuzione";
	playbuttontext = "\tGuarda\tGuarda i films selezionati";
	loopbuttontext1 = "\tContinuamente\tGuarda continuamente";
	loopbuttontext2 = "\tUna Volta\tGuarda un volta";
	zoombuttontext1 = "\tZoom\tZoom del video";
	zoombuttontext2 = "\t1:1\tDimensioni del video reali";
	hidebuttontext1 = "\tNascondi\tNascondi la lista";
	hidebuttontext2 = "\tMostra\tMostra la lista";
	sizebuttontext = "\tDimensioni\tDimensioni video";
	outputbuttontext = "\tUscita\tUscita video";
	confbuttontext = "\tConfigurazione\tConfigurazione di FxMovieManager";
	upbuttontext = "\tSposta in alto\tSposta in alto di un gradino";
	downbuttontext = "\tSposta in basso\tSposta in basso di un gradino";
	slistbuttontext = "|_|\tSalva la lista\tSalva la lista dei film";
	olistbuttontext = "||>\tCarica la lista\tCarica la lista dei film";
  	allbuttontext1 = "\tSeleziona tutti\tSeleziona tutti i filmati";
	allbuttontext2 = "\tDeseleziona tutti\tDeseleziona tutti i filmati";
	createbuttontext = "\tCrea l'anteprima\tCrea l'anteprima";
	recreatebuttontext = "\tRicrea l'anteprima\tRicrea l'anteprima";
  	widebuttontext1 = "\tWide Screen\tVisualizza Anteprima in modalità Wide Screen";
	widebuttontext2 = "\tSchermo Normale\tVisualizza Anteprima in modalità Normale";
	deletebuttontext = "\tElimina l'anteprima\tElimina l'anteprima";
	clearbuttontext = "\tCancella la lista\tCancella la lista";
	increasebuttontext = "\tIngrandisci l'anteprima\tIngrandisci l'anteprima";
	decreasebuttontext = "\tRiduci l'anteprima\tRiduci l'anteprima";
	detlisttext = "\telenco dettagliato\tvista in modalità elenco dettagliato";
	icnlisttext = "\telenco icona\tvista in modalità elenco icona";
	imgbuttontext1 = "\tNon mostrare le immagini\tNon mostrare l'anteprima delle immagini";
	imgbuttontext2 = "\tMostra le immagini\tMostra l'anteprima delle immagini";
	return;
}

void FXMovieManager::languageFR() {
	quitbuttontext = "\tQuittez l'application\tQuittez l'application";
	openbuttontext = "\tAjouter à la liste de films\tAjouter à la liste de films";
	audiobuttontext1 = "\tMute\tMPlayer doit être en cours d'exécution";
	audiobuttontext2 = "\tAudio\tMPlayer doit être en cours d'exécution";
	dropbuttontext1 = "\tRéduire les images\tMPlayer doit être en cours d'exécution";
	dropbuttontext2 = "\tToutes les images\tMPlayer doit être en cours d'exécution";
	stopbuttontext = "\tArrêter\tMPlayer doit être en cours d'exécution";
	pausebuttontext = "\tPause\tMPlayer doit être en cours d'exécution";
	playbuttontext = "\tRegarder\tRegarder les films sélectionnés";
	loopbuttontext1 = "\tBoucle\tRegarder en boucle";
	loopbuttontext2 = "\tOne fois\tRegarder une fois";
	zoombuttontext1 = "\tZoom\tZoom taille vidéo";
	zoombuttontext2 = "\t1:1\t1:1 taille vidéo";
	hidebuttontext1 = "\tMasquer\tMasquer la liste de film";
	hidebuttontext2 = "\tVoir\tVoir la liste de film";
	sizebuttontext = "\tTaille\ttaille de la vidéo";
	outputbuttontext = "\tSortie\tSortie vidéo";
	confbuttontext = "\tConfiguration\tConfigurer FxMovieManager";
	upbuttontext = "\tDéplacer vers le haut\tDéplacer vers le haut";
	downbuttontext = "\tDéplacez vers le bas\tDéplacez vers le bas";
	slistbuttontext = "|_|\tMettre la liste\tMettre la liste de film";
	olistbuttontext = "||>\tCharge de la liste\tCharge de la liste de film";
  	allbuttontext1 = "\tSélectionnez tous les films\tSélectionnez tous les films";
	allbuttontext2 = "\tDésactivez tous les films\tDésactivez tous les films";
	createbuttontext = "\tCrée une prévisualisation\tCrée une prévisualisation";
	recreatebuttontext = "\tRecrée une prévisualisation\tRecrée une prévisualisation";
  	widebuttontext1 = "\tWide Screen\tVoir les prévisualisation dans le mode Wide Screen";
	widebuttontext2 = "\tEcran Normal\tVoir les prévisualisation dans le mode Normal";
	deletebuttontext = "\tSupprimer une prévisualisation\tSupprimer une prévisualisation";
	clearbuttontext = "\tEffacer la liste de film\tEffacer la liste de film";
	increasebuttontext = "\tAugmenter la taille de prévisualisation\tAugmenter la taille de prévisualisation";
	decreasebuttontext = "\tDiminution de prévisualisation\tDiminution de prévisualisation";
	detlisttext = "\tliste détaillée\tvue en mode liste détaillée";
	icnlisttext = "\tliste d'icônes\tvue en liste d'icônes mode";
	imgbuttontext1 = "\tNe pas afficher les images miniatures\tNe pas afficher les images miniatures";
	imgbuttontext2 = "\tAfficher des miniatures\tAfficher des miniatures";
	return;
}

void FXMovieManager::languageDE() {
	quitbuttontext = "\tSchließen Sie das Programm\tSchließen Sie das Programm";
	openbuttontext = "\tHinzufügen zur Liste Filme\tHinzufügen zur Liste Filme";
	audiobuttontext1 = "\tMute\tMplayer muss ausgeführt werden";
	audiobuttontext2 = "\tAudio\tMplayer muss ausgeführt werden";
	dropbuttontext1 = "\tVerringerung der Bilder\tMplayer muss ausgeführt werden";
	dropbuttontext2 = "\tAlle bilder\tMplayer muss ausgeführt werden";
	stopbuttontext = "\tStoppen\tMplayer muss ausgeführt werden";
	pausebuttontext = "\tPause\tMplayer muss ausgeführt werden";
	playbuttontext = "\tSehen Sie den Film\tSehen Sie sich die ausgewählten Filme";
	loopbuttontext1 = "\tKontinuierlich\tSehen Sie den Film kontinuierlich";
	loopbuttontext2 = "\tEinmal\tSchauen Sie einmal";
	zoombuttontext1 = "\tZoom\tZoom Video-Größe";
	zoombuttontext2 = "\t1:1\tReal Video-Größe";
	hidebuttontext1 = "\tAusblenden\tAusblenden der Liste der Filme";
	hidebuttontext2 = "\tZeigt\tZeigt die Liste der Filme";
	sizebuttontext = "\tGröße\tVideo-Größe";
	outputbuttontext = "\tAusgang\tVideo-Ausgang";
	confbuttontext= "\tKonfiguration\tKonfigurieren FxMovieManager";
	upbuttontext= "\tEinen Schritt bis\tBewegen Sie ein Schritt bis";
	downbuttontext= "\tEinen Schritt nach unten\tBewegen Sie ein Schritt nach unten";
	slistbuttontext = "|_|\tListe speichern\tListe der Filme speichern";
	olistbuttontext = "||>\tLast Liste\tLast Liste der Filme";
  	allbuttontext1= "\tWählen Sie alle Filme\tWählen Sie alle Filme";
	allbuttontext2= "\tDeaktivieren Sie alle Filme\tDeaktivieren Sie alle Filme";
	createbuttontext= "\tSchafft Vorschau\tSchafft Vorschau";
	recreatebuttontext= "\tStellt Vorschau\tStellt Vorschau";
  	widebuttontext1 = "\tWide Screen\tVorschau-Modus Wide Screen";
	widebuttontext2 = "\tNormal\tVorschau-Modus Normal";
	deletebuttontext= "\tEliminiert Vorschau\tEliminiert Vorschau";
	clearbuttontext= "\tLöscht die Liste der Filme\tLöscht die Liste der Filme";
	increasebuttontext= "\tVergrössern Vorschau\tVergrössern Vorschau";
	decreasebuttontext= "\tNiederösterreich Vorschau\tNiederösterreich Vorschau";
	detlisttext = "\tdetaillierte Liste\tBlick in detaillierte Liste Modus";
	icnlisttext = "\tListe Symbol\tAnsicht in Liste Symbol-Modus";
	imgbuttontext1 = "\tzeigen keine Miniaturbilder\tzeigen keine Miniaturbilder";
	imgbuttontext2 = "\tBilder anzeigen\tBilder anzeigen";
	return;
}

void FXMovieManager::languageES() {
	quitbuttontext = "\tSalir del programa\tSalir del programa";
	openbuttontext = "\tAñadir a la lista de películas\tAñadir a la lista de películas";
	audiobuttontext1 = "\tMudo\tMPlayer debería estar ejecutándose";
	audiobuttontext2 = "\tAudio\tMPlayer debería estar ejecutándose";
	dropbuttontext1 = "\tReducir las imágenes\tMPlayer debería estar ejecutándose";
	dropbuttontext2 = "\tTodas las imágenes\tMPlayer debería estar ejecutándose";
	stopbuttontext = "\tParar\tMPlayer debería estar ejecutándose";
	pausebuttontext = "\tRomper\tMPlayer debería estar ejecutándose";
	playbuttontext = "\tMira\tMira las películas seleccionadas";
	loopbuttontext1 = "\tContinua\tMira continua";
	loopbuttontext2 = "\tUna vez\tMira una vez";
	zoombuttontext1 = "\tZoom\tVideo Zoom";
	zoombuttontext2 = "\t1:1\tTamaño real del video";
	hidebuttontext1 = "\tOcultar\tOcultar la lista de películas";
	hidebuttontext2 = "\tMuestra\tMuestra la lista de películas";
	sizebuttontext = "\tTamaño\tel tamaño video";
	outputbuttontext = "\tSalida\tsalida de vídeo";
	confbuttontext = "\tConfiguración\tConfiguración de FXMovieManager";
	upbuttontext = "\tSales de un paso\tSales de un paso";
	downbuttontext = "\tBajar un paso\tBajar un paso";
	slistbuttontext = "|_|\tGuardar la lista\tGuardar la lista de películas";
	olistbuttontext = "||>\tCarga de la lista\tCarga de la lista de películas";
	allbuttontext1 = "\tSeleccione todas las películas\tSeleccione todas las películas";
	allbuttontext2 = "\tDesmarque todas las películas\tDesmarque todas las películas";
	createbuttontext = "\tCrear una vista previa\tCrear una vista previa";
	recreatebuttontext = "\tRecrear la vista previa\tRecrear la vista previa";
  	widebuttontext1 = "\tWide Screen\tMostrar vista previa en modo de pantalla ancha";
	widebuttontext2 = "\tNormal Screen\tMostrar vista previa en modo de pantalla normal";
	deletebuttontext = "\tEliminar la vista previa\tEliminar la vista previa";
	clearbuttontext = "\tBorrar la lista de películas\tBorrar la lista de películas";
	increasebuttontext = "\tAmpliar la vista previa\tAmpliar la vista previa";
	decreasebuttontext = "\tMenos de vista previa\tMenos de vista previa";
	detlisttext = "\tlista detallada\tver lista detallada en el modo";
	icnlisttext = "\ticono de lista\ticono de vista en modo lista";
	imgbuttontext1 = "\tno mostrar las imágenes en miniatura\tno mostrar las imágenes en miniatura";
	imgbuttontext2 = "\tMostrar imágenes en miniatura\tMostrar imágenes en miniatura";
	return;
}

void FXMovieManager::languageRU() {
	quitbuttontext = "\tВыход\tВыход из программы";
	openbuttontext = "\tДобавить\tДобавить в список фильмов";
	audiobuttontext1 = "\tТупой\tMPlayer должна быть запущена";
	audiobuttontext2 = "\tАудио\tMPlayer должна быть запущена";
	dropbuttontext1 = "\tСнижение изображений\tMPlayer должна быть запущена";
	dropbuttontext2 = "\tВсе изображения\tMPlayer должна быть запущена";
	stopbuttontext = "\tОстановка\tОстановка";
	pausebuttontext = "\tПерерыв\tMPlayer должна быть запущена";
	playbuttontext = "\tПосмотреть\tСмотреть фильмы выбранные";
	loopbuttontext1 = "\tНепрерывно\tПосмотреть постоянно";
	loopbuttontext2 = "\tОднажды\tОдин раз увидеть";
	zoombuttontext1 = "\tУвеличить\tУвеличить видео";
	zoombuttontext2 = "\t1:1\tРазмер реального видео";
	hidebuttontext1 = "\tПрятать\tСкрыть список";
	hidebuttontext2 = "\tПоказывать\tПосмотреть список";
	sizebuttontext = "\tРазмер\tРазмер видео";
	outputbuttontext = "\tвидеовыходу\tвидеовыходу";
	confbuttontext = "\tКонфигурация\tНастройка FxMovieManager";
	upbuttontext = "\tВверх\tВверх шаг";
	downbuttontext = "\tВнизo\tВниз шаг";
	slistbuttontext = "|_|\tСохранить список\tСохранить список фильмов";
	olistbuttontext = "||>\tзагрузить список\tзагрузить список фильмов";
  	allbuttontext1 = "\tВыбрать все\tВыбрать все фильмы";
	allbuttontext2 = "\tСнять все\tСнимите все фильмы";
	createbuttontext = "\tСоздать прослушать\tСоздать прослушать";
	recreatebuttontext = "\tЗаново прослушать\tЗаново прослушать";
  	widebuttontext1 = "\tШирокий экран\tШирокий экран";
	widebuttontext2 = "\tОбычный экран\tОбычный экран";
	deletebuttontext = "\tУдалить прослушать\tУдалить прослушать";
	clearbuttontext = "\tУдалить список\tУдалить список";
	increasebuttontext = "\tУвеличить прослушать\tУвеличить прослушать";
	decreasebuttontext = "\tМинус прослушать\tМинус прослушать";
	detlisttext = "\tПодробный список\tПосмотреть в подробный перечень режиме";
	icnlisttext = "\tСписок иконы\tПросмотреть в режиме иконки списка";
	imgbuttontext1 = "\tНе показывать миниатюры\tНе показывать миниатюры";
	imgbuttontext2 = "\tПоказать изображения миниатюры\tПоказать изображения миниатюры";
	return;
}

void FXMovieManager::languageCN() {
	quitbuttontext = "\t退出\t退出应用";
	openbuttontext = "\t添加电影\t电影添加到播放列表";
	audiobuttontext1 = "\t静音\tMPlayer的必须运行";
	audiobuttontext2 = "\t音频\tMPlayer的必须运行";
	dropbuttontext1 = "\t丢弃帧\tMPlayer的必须运行";
	dropbuttontext2 = "\t所有框架\tMPlayer的必须运行";
	stopbuttontext = "\t停止\t停止播放";
	pausebuttontext = "\t暂停\tMPlayer的必须运行";
	playbuttontext = "\t玩\t播放所选的电影";
	loopbuttontext1 = "\t环\t在循环播放";
	loopbuttontext2 = "\t一次\t玩一次";
	zoombuttontext1 = "\t变焦\t变焦视频尺寸";
	zoombuttontext2 = "\t1:1\t真实视频尺寸";
	hidebuttontext1 = "\t隐藏\t隐藏播放列表";
	hidebuttontext2 = "\t显示\t显示播放列表";
	sizebuttontext = "\t大小\t改变视频尺寸";
	outputbuttontext = "\t输出\t更改视频输出";
	confbuttontext = "\t配置\tFxMovieManager配置";
	upbuttontext = "\t移动一个加强\t移动一个加强";
	downbuttontext = "\t移动一个下台\t移动一个下台";
	slistbuttontext = "|_|\t保存电影列表\t保存电影列表";
	olistbuttontext = "||>\t负载电影列表\t负载电影列表";
  	allbuttontext1 = "\t选择所有电影\t选择所有电影";
	allbuttontext2 = "\t取消所有电影\t取消所有电影";
	createbuttontext = "\t创建缩略图\t创建缩略图";
	recreatebuttontext = "\t重新创建缩略图\t重新创建缩略图";
  	widebuttontext1 = "\t宽屏幕\t宽屏幕显示缩略图";
	widebuttontext2 = "\t普通屏幕\t普通屏幕显示缩略图";
	deletebuttontext = "\t删除缩略图\t删除缩略图";
	clearbuttontext = "\t清除播放列表\t清除播放列表";
	increasebuttontext = "\t增加缩略图的尺寸\t增加缩略图的尺寸";
	decreasebuttontext = "\t缩略图的尺寸减少\t缩略图的尺寸减少";
	detlisttext = "\t详细列表视图\t详细列表视图";
	icnlisttext = "\t图标列表视图\t图标列表视图";
	imgbuttontext1 = "\t不显示缩略图\t不显示缩略图";
	imgbuttontext2 = "\t显示缩略图\t显示缩略图";
	return;
}

void FXMovieManager::languageJP() {
	quitbuttontext = "\t終了する\tプログラムの終了";
	openbuttontext = "\t作品を追加\tプレイリストに作品を追加";
	audiobuttontext1 = "\tミュート\tMPlayerに実行する必要があります";
	audiobuttontext2 = "\tオーディオ\tMPlayerに実行する必要があります";
	dropbuttontext1 = "\tドロップフレーム\tMPlayerに実行する必要があります";
	dropbuttontext2 = "\tすべてのフレームを\tMPlayerに実行する必要があります";
	stopbuttontext = "\t停止する\t停止プレーヤー";
	pausebuttontext = "\t一時停止\tMPlayerに実行する必要があります";
	playbuttontext = "\t遊ぶ\t再生する映画を選択";
	loopbuttontext1 = "\tループ\tループ内で再生";
	loopbuttontext2 = "\t一度\t再生1回";
	zoombuttontext1 = "\tズーム\tズームビデオサイズ";
	zoombuttontext2 = "\t1:1\tリアルビデオのサイズ";
	hidebuttontext1 = "\t隠す\tプレイリストを隠す";
	hidebuttontext2 = "\t見せる\t表示されるプレイリスト";
	sizebuttontext = "\t寸法\tビデオサイズ";
	outputbuttontext = "\t出力\tビデオ出力";
	confbuttontext = "\t構成\tFxMovieManager構成";
	upbuttontext = "\t一歩上に移動\t一歩上に移動";
	downbuttontext = "\t移動する1つのステップダウン\t移動する1つのステップダウン";
	slistbuttontext = "|_|\t保存プレイリスト\t保存プレイリスト";
	olistbuttontext = "||>\t負荷プレイリスト\t負荷プレイリスト";
  	allbuttontext1 = "\t選択したすべて\t選択して、すべてのムービー";
	allbuttontext2 = "\tすべて選択解除\t選択をすべて解除映画";
	createbuttontext = "\tサムネールを作成します。\tサムネールを作成します。";
	recreatebuttontext = "\tを再現サムネイル\tを再現サムネイル";
  	widebuttontext1 = "\tワイドスクリーン\t地図をワイドスクリーンのサムネイル";
	widebuttontext2 = "\tノーマル画面\t地図を標準の画面のサムネイル";
	deletebuttontext = "\t削除サムネイル\t削除サムネイル";
	clearbuttontext = "\tクリアプレイリスト\tクリアプレイリスト";
	increasebuttontext = "\tサムネイルのサイズを増やします。\tサムネイルのサイズを増やします。";
	decreasebuttontext = "\t減少サイズのサムネイル\t減少サイズのサムネイル";
	detlisttext = "\t詳細なリストを表示\t詳細なリストを表示";
	icnlisttext = "\tアイコンの一覧を表示\tアイコンの一覧を表示";
	imgbuttontext1 = "\tサムネイル画像を表示しない\tサムネイル画像を表示しない";
	imgbuttontext2 = "\t表示するサムネイル画像\t表示するサムネイル画像";
	return;
}


// load configuration
FXbool FXMovieManager::readConfig() {
  if(!getApp()->reg().read()){
	return FALSE;
  }
  // Get size and position
#ifdef HAVE_LIBFOX_1_7
  FXuint width=getApp()->reg().readUIntEntry("WINDOW","width",800);
  FXuint height=getApp()->reg().readUIntEntry("WINDOW","height",42);
#endif
#ifdef HAVE_LIBFOX_1_6
  FXuint width=getApp()->reg().readUnsignedEntry("WINDOW","width",800);
  FXuint height=getApp()->reg().readUnsignedEntry("WINDOW","height",42);
#endif 

  FXint xpos=getApp()->reg().readIntEntry("WINDOW","xpos",40);
  FXint ypos=getApp()->reg().readIntEntry("WINDOW","ypos",5);

  // Reposition window
  position(xpos,ypos,width,height);

  // Read menu options
  s_full=getApp()->reg().readStringEntry("MENU","s_full");
  s_loop=getApp()->reg().readStringEntry("MENU","s_loop");
  s_frame=getApp()->reg().readStringEntry("MENU","s_frame");
  s_audio=getApp()->reg().readStringEntry("MENU","s_audio");

  i_video=getApp()->reg().readIntEntry("MENU","s_video",i_video);
  i_size=getApp()->reg().readIntEntry("MENU","s_size",i_size);

  s_listshow=getApp()->reg().readStringEntry("MENU","s_listshow");

  currentdir_lp=getApp()->reg().readStringEntry("FileList","currentdir_lp");
  currentdir_rp=getApp()->reg().readStringEntry("FileList","currentdir_rp");

  splitter2->setSplit(0,getApp()->reg().readIntEntry("FileList","dirwidth",260));
  splitter1->setSplit(1,getApp()->reg().readIntEntry("FileList","dirheigth",380));

  fxmmwheellines=getApp()->reg().readIntEntry("SETTINGS","fxmmwheellines",4);

  terminal=getApp()->reg().readStringEntry("SETTINGS","terminal");

  fontspec=getApp()->reg().readStringEntry("SETTINGS","bodyfont","");
  if(!fontspec.empty()){
    bodyfont=new FXFont(getApp(),fontspec);
    filelist_lp->setFont(bodyfont);
    dirlist_lp->setFont(bodyfont);
    filelist_rp->setFont(bodyfont);
    dirlist_rp->setFont(bodyfont);
    table->setFont(bodyfont);
    tooltip->setFont(bodyfont);
	statusline->setFont(bodyfont);
    }

#ifdef HAVE_LIBFFMPEGTHUMBNAILER
  // get thumbnailer options
  thumbnailSize = getApp()->reg().readIntEntry("THUMBNAILER","thumbnailSize",128);
  imageQuality = getApp()->reg().readIntEntry("THUMBNAILER","imageQuality",8);
  filmStripOverlay = getApp()->reg().readBoolEntry("THUMBNAILER","filmStripOverlay",FALSE);
  workaroundIssues = getApp()->reg().readBoolEntry("THUMBNAILER","workaroundIssues",FALSE);
  maintainAspectRatio = getApp()->reg().readBoolEntry("THUMBNAILER","maintainAspectRatio",TRUE);
  smartFrameSelection = getApp()->reg().readBoolEntry("THUMBNAILER","smartFrameSelection",FALSE);
#endif
  autoThumbnailShow = getApp()->reg().readBoolEntry("THUMBNAILER","autoThumbnailShow",TRUE);
  autoThumbnailCreate = getApp()->reg().readBoolEntry("THUMBNAILER","autoThumbnailCreate",FALSE);
  autoThumbnailUpdate = getApp()->reg().readBoolEntry("THUMBNAILER","autoThumbnailUpdate",FALSE);

  firstcolwidth=getApp()->reg().readIntEntry("SETTINGS","firstcolwidth",480);

  // Read registry settings
  if(getApp()->reg().read()){
    return TRUE;
  }
}

// create application
void FXMovieManager::create() {

  if(!readConfig()){
	FXMessageBox::error(this,MBOX_OK,"CONFIGURATION ERROR","Configuration not loaded!");
  }
  getApp()->setWheelLines(fxmmwheellines);
  table->setColumnWidth(0,firstcolwidth);
  dirlist_lp->setDirectory(currentdir_lp);
  filelist_lp->setDirectory(currentdir_lp);
  dirlist_rp->setDirectory(currentdir_rp);
  filelist_rp->setDirectory(currentdir_rp);

  if(s_listshow=="false"){ hidebutton->setState(TRUE); listshow=false; } 
  else if(s_listshow=="true"){ hidebutton->setState(FALSE); listshow=true;}

  if(s_full==""){ zoombutton->setState(FALSE); } 
  else if(s_full=="-fs "){ zoombutton->setState(TRUE); }

  if(s_loop==""){ loopbutton->setState(FALSE); } 
  else if(s_loop=="-loop 0 "){ loopbutton->setState(TRUE); }

  if(s_audio==""){ audiobutton->setState(FALSE); } 
  else if(s_audio=="-nosound "){ audiobutton->setState(TRUE); }

  if(s_frame==""){ dropbutton->setState(FALSE); } 
  else if(s_frame=="-framedrop "){ dropbutton->setState(TRUE); }

  switch(i_video){
	case 0:
		s_video="-vo x11 -fixed-vo ";
		break;
	case 1:
		s_video="-vo xv -fixed-vo ";
		break;
	case 2:
		s_video="-vo gl -fixed-vo ";
		break;
	case 3:
		s_video="-vo gl2 -fixed-vo ";
		break;
	case 4:
		s_video="-vo sdl -fixed-vo ";
		break;
	case 5:
		s_video="-vo dga -fixed-vo ";
		break;
	case 6:
		s_video="-vo xvidix -fixed-vo ";
		break;
	case 7:
		s_video="-vo jpeg ";
		break;
	case 8:
		s_video="-vo png ";
		break;
	case 9:
		s_video="-vo null ";
		break;
	}

  switch(i_size){
	case 0:
		s_size="-zoom -x 1008 -y 752 ";
		break;
	case 1:
		s_size="-zoom -x 1008 -y 576 ";
		break;
	case 2:
		s_size="-zoom -x 720 -y 576 ";
		break;
	case 3:
		s_size="-zoom -x 640 -y 480 ";
		break;
	case 4:
		s_size="-zoom -x 560 -y 320 ";
		break;
	case 5:
		s_size="-zoom -x 500 -y 284 ";
		break;
	case 6:
		s_size="-zoom -x 480 -y 576 ";
		break;
	case 7:
		s_size="-zoom -x 480 -y 480 ";
		break;
	case 8:
		s_size="-zoom -x 352 -y 288 ";
		break;
	case 9:
		s_size="-zoom -x 352 -y 240 ";
		break;
	case 10:
		s_size="-zoom -x 320 -y 240 ";
		break;
	case 11:
		s_size="";
		break;
	}
  getApp()->addTimeout(this,ID_CLOCKTIME,CLOCKTIMER);
  FXMainWindow::create();
  show(PLACEMENT_SCREEN);
  showFilelistThumbnails_lp();
  showFilelistThumbnails_rp();
  return;
}


// quit application
long FXMovieManager::onCmdQuit(FXObject *,FXSelector,void *) {
  quitAppFxmm();
  return 1;
}


void FXMovieManager::quitAppFxmm(){
 
  FXTRACE((100,"Quit\n"));
  // Write new window size and position back to registry
#ifdef HAVE_LIBFOX_1_7
  getApp()->reg().writeUIntEntry("WINDOW","width",(FXuint)getWidth());
  getApp()->reg().writeUIntEntry("WINDOW","height",(FXuint)getHeight());
#endif
#ifdef HAVE_LIBFOX_1_6
  getApp()->reg().writeUnsignedEntry("WINDOW","width",(FXuint)getWidth());
  getApp()->reg().writeUnsignedEntry("WINDOW","height",(FXuint)getHeight());
#endif 
  getApp()->reg().writeIntEntry("WINDOW","xpos",getX());
  getApp()->reg().writeIntEntry("WINDOW","ypos",getY());

  // Write mencoder options
  getApp()->reg().writeIntEntry("MENU","s_video",i_video);
  getApp()->reg().writeIntEntry("MENU","s_size",i_size);

  getApp()->reg().writeStringEntry("MENU","s_full",s_full.text());
  getApp()->reg().writeStringEntry("MENU","s_loop",s_loop.text());
  getApp()->reg().writeStringEntry("MENU","s_frame",s_frame.text());
  getApp()->reg().writeStringEntry("MENU","s_audio",s_audio.text());

  if (listshow==true) {s_listshow="true";}
  else if (listshow==false) {s_listshow="false";}
  getApp()->reg().writeStringEntry("MENU","s_listshow",s_listshow.text());

  currentdir_lp = dirlist_lp->getDirectory();
  getApp()->reg().writeStringEntry("FileList","currentdir_lp",currentdir_lp.text());
  currentdir_rp = dirlist_rp->getDirectory();
  getApp()->reg().writeStringEntry("FileList","currentdir_rp",currentdir_rp.text());

  getApp()->reg().writeIntEntry("FileList","dirwidth",splitter2->getSplit(0));
  getApp()->reg().writeIntEntry("FileList","dirheigth",splitter1->getSplit(1));

  getApp()->reg().writeIntEntry("SETTINGS","firstcolwidth",table->getColumnWidth(0));

  // Write registry settings
  getApp()->reg().write();

  system (removeTempFifo.text());

  // Quit
  getApp()->exit(0);
  return;
}


/**************************PLAYER LIST BUTTONS****************************/


// Change video output
long FXMovieManager::onCmdVideoSwitch(FXObject* sender,FXSelector sel,void* ptr){
  FXint n_video=FXSELID(sel)-ID_VIDEO_1+1;
  if(0<=n_video && n_video<10){
    switch(FXSELID(sel)){
	case ID_VIDEO_0:
		i_video=0;
		s_video="-vo x11 -fixed-vo ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Video Output [x11] Enabled\nVideo on Top Enabled");
		break;
	case ID_VIDEO_1:
		i_video=1;
		s_video="-vo xv -fixed-vo ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Video Output [xv] Enabled\nVideo on Top Disabled");
		break;
	case ID_VIDEO_2:
		i_video=2;
		s_video="-vo gl -fixed-vo ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Video Output [gl] Enabled\nVideo on Top Disabled");
		break;
	case ID_VIDEO_3:
		i_video=3;
		s_video="-vo gl2 -fixed-vo ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Video Output [gl2] Enabled\nVideo on Top Disabled");
		break;
	case ID_VIDEO_4:
		i_video=4;
		s_video="-vo sdl -fixed-vo ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Video Output [sdl] Enabled\nVideo on Top Disabled");
		break;
	case ID_VIDEO_5:
		i_video=5;
		s_video="-vo dga -fixed-vo ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Video Output [dga] Enabled");
		break;
	case ID_VIDEO_6:
		i_video=6;
		s_video="-vo xvidix -fixed-vo ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Video Output [xvidix] Enabled");
		break;
	case ID_VIDEO_7:
		i_video=7;
		s_video="-vo jpeg ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Output >>> jpeg files enabled");
		break;
	case ID_VIDEO_8:
		i_video=8;
		s_video="-vo png ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Output >>> png files enabled");
		break;
	case ID_VIDEO_9:
		i_video=9;
		s_video="-vo null ";
		FXMessageBox::information(this,MBOX_OK,"Option selected","Video Output Disabled");
		break;
      }
    }
  return 1;
  }


// Update video output
long FXMovieManager::onUpdVideoSwitch(FXObject* sender,FXSelector sel,void* ptr){
  FXint n_video=FXSELID(sel)-ID_VIDEO_1+1;
  sender->handle(this,i_video==n_video?FXSEL(SEL_COMMAND,ID_CHECK):FXSEL(SEL_COMMAND,ID_UNCHECK),NULL);
  return 1;
  }


// Change video size
long FXMovieManager::onCmdSizeSwitch(FXObject* sender,FXSelector sel,void* ptr){
  FXint n_size=FXSELID(sel)-ID_SIZE_1+1;
  if(0<=n_size && n_size<12){
    switch(FXSELID(sel)){
	case ID_SIZE_0:
		i_size=0;
		s_size="-zoom -x 1008 -y 752 ";
		break;
	case ID_SIZE_1:
		i_size=1;
		s_size="-zoom -x 1008 -y 576 ";
		break;
	case ID_SIZE_2:
		i_size=2;
		s_size="-zoom -x 720 -y 576 ";
		break;
	case ID_SIZE_3:
		i_size=3;
		s_size="-zoom -x 640 -y 480 ";
		break;
	case ID_SIZE_4:
		i_size=4;
		s_size="-zoom -x 560 -y 320 ";
		break;
	case ID_SIZE_5:
		i_size=5;
		s_size="-zoom -x 500 -y 284 ";
		break;
	case ID_SIZE_6:
		i_size=6;
		s_size="-zoom -x 480 -y 576 ";
		break;
	case ID_SIZE_7:
		i_size=7;
		s_size="-zoom -x 480 -y 480 ";
		break;
	case ID_SIZE_8:
		i_size=8;
		s_size="-zoom -x 352 -y 288 ";
		break;
	case ID_SIZE_9:
		i_size=9;
		s_size="-zoom -x 352 -y 240 ";
		break;
	case ID_SIZE_10:
		i_size=10;
		s_size="-zoom -x 320 -y 240 ";
		break;
	case ID_SIZE_11:
		i_size=11;
		s_size="";
		break;
      }
    }
  return 1;
  }


// Update video size
long FXMovieManager::onUpdSizeSwitch(FXObject* sender,FXSelector sel,void* ptr){
  FXint n_size=FXSELID(sel)-ID_SIZE_1+1;
  sender->handle(this,i_size==n_size?FXSEL(SEL_COMMAND,ID_CHECK):FXSEL(SEL_COMMAND,ID_UNCHECK),NULL);
  return 1;
  }


// save playlist
FXString FXMovieManager::returnPlaylist() {

	FXString s_initial, s_duration;

	if(numtableitems >= 0){
		// Return the text of all items
		for(int r=0; r<numtableitems; ++r){
			if(table->isItemSelected(r,0)){
			//get the position where to start playing of the movie	
			s_initial = "-ss " + table->getItemText(r,1) + ":" + table->getItemText(r,2) + ":" + table->getItemText(r,3)+ "." + table->getItemText(r,4) + " ";
			s_duration = "-endpos " + table->getItemText(r,5) + ":" + table->getItemText(r,6) + ":" + table->getItemText(r,7) + "." + table->getItemText(r,8) + " ";
			s_list += table->getItemText(r,0) + " " + s_initial + s_duration;
			}
		}
	}
	return s_list;
}


// playlist play command
long FXMovieManager::onCmdPlay(FXObject *,FXSelector,void *) {

	// reset the fifo every time we play
	system (removeTempFifo.text());
	system (createTempFifo.text());

	//get the speed
	FXfloat i_speed = speedslider->getValue();
	FXfloat f_speed = (i_speed/10);
	FXString s_speed = "-speed " + (FXString::value(f_speed,2)) + " ";

	//get the volume
	FXint i_volume = volumeslider->getValue();
	//FXString s_volume = "-volume " + (FXString::value(i_volume)) + " 5 ";
	FXString s_volume = "-volume " + (FXString::value(i_volume)) + " ";

	returnPlaylist();
	if(s_list.empty()) return 1;

	// correct command list sequence....
	FXString s_play = s_command + s_loop + s_video + s_size + s_full + s_speed + s_audio + s_volume + s_frame + s_list + " &";
	printf("The playstring was: %s\n", s_play.text());

       while (isrunning == TRUE){
       int ret=system("killall mplayer");
             if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)){
                 break;
	     }
	     isrunning = FALSE;
	}

       while (isrunning == FALSE){
       int ret=system(s_play.text());
             if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)){
                 break;
	     }
	     isrunning = TRUE;
	}

    	// clear liststr
    	s_play.clear();
    	s_list.clear();

	return 1;
}


// playlist pause command
long FXMovieManager::onCmdPause(FXObject *,FXSelector,void *) {
 	if (isrunning==TRUE){
		FILE *f;
		f=fopen(currentFifo.text(), "w");
		fprintf(f, "%s", "pause\n");
		fclose(f);
	} else if (isrunning==FALSE){
		FXMessageBox::information(this,MBOX_OK,"Command Failure","MPlayer is NOT running");
	}
	return 1;
}


// playlist stop command
long FXMovieManager::onCmdStop(FXObject *,FXSelector,void *) {
	if (isrunning==TRUE){
		FILE *f;
		f=fopen(currentFifo.text(), "w");
		fprintf(f, "%s", "quit 1\n");
		fclose(f);
		isrunning=FALSE;
	} else if (isrunning==FALSE){
		FXMessageBox::information(this,MBOX_OK,"Command Failure","MPlayer is NOT running");
	}
	return 1;
}

// playlist audio command
long FXMovieManager::onCmdAudio(FXObject *,FXSelector,void *) {
   	if(isrunning==TRUE){
		if(s_audio==""){
			FILE *f;
			f=fopen(currentFifo.text(), "w");
			fprintf(f, "%s", "mute 1\n");
			fclose(f);
			s_audio="-nosound ";
		} else if(s_audio=="-nosound "){
			FILE *f;
			f=fopen(currentFifo.text(), "w");
			fprintf(f, "%s", "mute 0\n");
			fclose(f);
			s_audio= "";
		}
   	} else if (isrunning==FALSE){
		FXMessageBox::information(this,MBOX_OK,"Command Failure","MPlayer is NOT running");
		if(s_audio==""){
			s_audio="-nosound ";
		} else if(s_audio=="-nosound "){
			s_audio= "";
		}
   	}
   	return 1;
}

// playlist frame drop command
long FXMovieManager::onCmdDrop(FXObject *,FXSelector,void *) {
   	if(isrunning==TRUE){
		if(s_frame==""){
			FILE *f;
			f=fopen(currentFifo.text(), "w");
			fprintf(f, "%s", "frame_drop 1\n");
			fclose(f);
			s_frame = "-framedrop ";
		} else if(s_frame=="-framedrop "){
			FILE *f;
			f=fopen(currentFifo.text(), "w");
			fprintf(f, "%s", "frame_drop 0\n");
			fclose(f);
			s_frame = "";
		}
    } else if (isrunning==FALSE){
		FXMessageBox::information(this,MBOX_OK,"Command Failure","MPlayer is NOT running");
		if(s_frame==""){
			s_frame = "-framedrop ";
		} else if(s_frame=="-framedrop "){
			s_frame = "";
		}
    }
    return 1;
}

// playlist video zoom command
long FXMovieManager::onCmdFull(FXObject *,FXSelector,void *) {
	if(s_full==""){
		s_full = "-fs ";
		zoombutton->setState(TRUE);
	} else if(s_full=="-fs "){
		s_full = "";
		zoombutton->setState(FALSE);
	}
	return 1;
}

// playlist loop command
long FXMovieManager::onCmdLoop(FXObject *,FXSelector,void *) {
	if(s_loop==""){
		s_loop = "-loop 0 ";
		loopbutton->setState(TRUE);
	} else if(s_loop=="-loop 0 "){
		s_loop = "";
		loopbutton->setState(FALSE);
	}
	return 1;
}

// playlist-filemanager hide command
long FXMovieManager::onCmdPutBack(FXObject *,FXSelector,void *) {
	if (listshow==false) {
		FXMainWindow::setHeight(600);
		FXMainWindow::repaint();
       		FXMainWindow::show(PLACEMENT_SCREEN);
		listshow=true;
		hidebutton->setState(FALSE);
	} else if (listshow==true) {
		FXMainWindow::setHeight(42);
		FXMainWindow::repaint();
        	FXMainWindow::show(PLACEMENT_SCREEN);
		listshow=false;
		hidebutton->setState(TRUE);
	}
	return 1;
}



/**************************TABLE COMMANDS****************************/

// load playlist
FXbool FXMovieManager::loadPlaylistTable(const FXString& filename) {
	// the file stream
	FXFileStream stream;	
	FXString tabpic;

	// open the file stream for reading
	if(stream.open(filename, FXStreamLoad)){ 
	//FXbool endianness;
	//stream >> endianness;
	//stream.setBigEndian(endianness);
	FXString totrootname,rootname,imagename,fullpath;
	FXint c,r;
	while(!stream.eof()){
		// Write the text of all items
	  	for (r=0; r<100; r++){
			for (c=0; c<9; c++){
				stream >> tabpic;
				FXStreamStatus sta=stream.status();
				if(sta!=FXStreamEnd){
		    			table->setItemText(r,c,tabpic);
					if(c==0){
						numtableitems++;
						fullpath=tabpic.rbefore('/',1);
    						//printf("fullpath is: %s\n", fullpath.text());
						totrootname=tabpic.rbefore('.',1);
    						//printf("totrootname is: %s\n", totrootname.text());
  						rootname=totrootname.rafter('/',1);
    						//printf("rootname is: %s\n", rootname.text());
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
						imagename=fullpath+"/."+rootname+"_thumbnail.jpg";
#else
						imagename=fullpath+"/."+rootname+"00000001/00000001.jpg";
#endif
    					//printf("imagename is: %s\n", imagename.text());
  						if(!previewListMovie(imagename, numtableitems-1)){
							FXMessageBox::error(this,MBOX_OK,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
						}
        				//printf("itemnumbers is: %d\n", numtableitems-1);
					}
	  			}
			}
		}
	}
	// close the file stream
	stream.close();
	return TRUE;
	}
return FALSE;
}

FXbool FXMovieManager::savePlaylistTable(const FXString& filename) {

	// the file stream
	FXFileStream stream;	
	// open the file stream for writing
	if(stream.open(filename, FXStreamSave)){
		//FXbool endianness=FXStream::isBigEndian();
		//stream << endianness;

		if(numtableitems >= 0){
			// Return the text of all items
			for(int r=0; r<numtableitems; ++r){
				if(table->isItemSelected(r,0)){
				stream << table->getItemText(r,0); 
				stream << table->getItemText(r,1);
				stream << table->getItemText(r,2); 
				stream << table->getItemText(r,3);
				stream << table->getItemText(r,4); 
				stream << table->getItemText(r,5);
				stream << table->getItemText(r,6); 
				stream << table->getItemText(r,7);
				stream << table->getItemText(r,8); 
				}
			}
		}
		// close the file stream
		stream.close();
	return TRUE;
	}
return FALSE;
}



// Move Up Command
long FXMovieManager::onCmdMoveUp(FXObject*,FXSelector,void*){
    FXString rowtext;
    FXIcon *tempicon=NULL;
    tempicon=new FXIcon(getApp(),NULL,IMAGE_OPAQUE|IMAGE_KEEP);
	// Return current row index, if any
	FXint r = table->getCurrentRow();
	if(numtableitems > r){
		if(table->isItemSelected(r,0)){
			if(r >= 0){
				FXint nr = r-1;
				if(nr>=0){
					// Move Up the current item
					table->extractText(rowtext,r,r,0,8);
    				if (tempicon){
						tempicon=table->getRowIcon(r);				
					}
					table->removeRows(r);
					table->insertRows(nr);
	  				table->setRowHeight(nr,68);
					table->overlayText(nr,nr,0,8,rowtext);
					table->setRowIcon(nr,tempicon);
					table->selectRange(nr,nr,0,0);
					table->setCurrentItem(nr,0);
				}
			}
		}
	}
	return 1;
}


// Move Down Command
long FXMovieManager::onCmdMoveDown(FXObject*,FXSelector,void*){
    FXString rowtext;
    FXIcon *tempicon=NULL;
    tempicon=new FXIcon(getApp(),NULL,IMAGE_OPAQUE|IMAGE_KEEP);
	// Return current row index, if any
	FXint r = table->getCurrentRow();
	if(numtableitems > r){
		if(table->isItemSelected(r,0)){
			if(r >= 0){
			FXint nr = r+1;
			if(nr<numtableitems){
				// Move Down the current item
				table->extractText(rowtext,r,r,0,8);
    				if (tempicon){
						tempicon=table->getRowIcon(r);				
					}
					table->removeRows(r);
					table->insertRows(nr);
	  				table->setRowHeight(nr,68);
					table->overlayText(nr,nr,0,8,rowtext);
					table->setRowIcon(nr,tempicon);
					table->selectRange(nr,nr,0,0);
					table->setCurrentItem(nr,0);
				}
			}
		}
	}
	return 1;
}


// Append command
long FXMovieManager::onCmdAppend(FXObject *,FXSelector,void *) {
	appendFiledialogSelection();
	return 1;
}


// Insert Command
long FXMovieManager::onCmdInsert(FXObject*,FXSelector,void*){
	insertFiledialogSelection();
	return 1;
}


// Select All Command
long FXMovieManager::onCmdSelectAll(FXObject*,FXSelector,void*){
	selectAll();
	return 1;
}


// Delete Command
long FXMovieManager::onCmdDelete(FXObject*,FXSelector,void*){

	if(numtableitems >= 0){
		// Remove all selected items
		for(int r=numtableitems; r>=0; --r){
			if(table->isItemSelected(r,0)){
				//printf("The item index selected was %d\n", i);
				table->removeRows(r);
				numtableitems=numtableitems-1;
 				table->insertRows(99);
				table->setRowHeight(99,68);
				for (int c=1; c<9; c++){
	    			table->setItemText(99,c,"00");
	  			}
	  			table->setItemText(99,5,"10");
			}
		}
	}
	return 1;
}


// Save Playlist Command
long FXMovieManager::onCmdSavePlaylist(FXObject*,FXSelector,void*){
	FXString file = "playlist_.pls";
	FXFileDialog savedialog(this,"Save Playlist");
	savedialog.setFilename(file);
	if(savedialog.execute()){
		selectAll();
		file=savedialog.getFilename();
		if(FXStat::exists(file)){
      		if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"Overwrite Playlist","Overwrite existing Playlist: %s?",file.text()))
			return 1;
		}
		if(savePlaylistTable(file)){
			FXMessageBox::information(this,MBOX_OK,"Saving Playlist","The playlist have been saved");	
		}
	}
	return 1;
}


// Open Playlist Command
long FXMovieManager::onCmdOpenPlaylist(FXObject*,FXSelector,void*){
	FXString file;
	FXFileDialog opendialog(this,"Load Playlist");
	if(opendialog.execute()){
		resetTable();
		file=opendialog.getFilename();
		if(loadPlaylistTable(file)){
			FXMessageBox::information(this,MBOX_OK,"Loading Playlist","The playlist have been loaded");	
		}
	}
	return 1;
}



// Clear Command
long FXMovieManager::onCmdClear(FXObject*,FXSelector,void*){
	resetTable();
	return 1;
}


// Released right button
long FXMovieManager::onListRightMouse(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  if(!event->moved){
    popupmenu->popup(NULL,event->root_x,event->root_y);
    getApp()->runModalWhileShown(popupmenu);
    }
  return 1;
}


// Select All Command
void FXMovieManager::selectAll(){
	FXint r;
	if(numtableitems >= 0){
		table->selectRange(0,numtableitems-1,0,0);
	}
  }


// Enable sender when items have been selected 
long FXMovieManager::onUpdSelectedTable(FXObject* sender,FXSelector,void*){
  if(numtableitems<0){
	return 1;
  }	
  sender->handle(this,getTableNums()?FXSEL(SEL_COMMAND,ID_ENABLE):FXSEL(SEL_COMMAND,ID_DISABLE),NULL);
  return 1;
  }


// Get number of selected filenames, in the play table 
FXint FXMovieManager::getTableNums() const {
  register FXint tablenum=0;
  register FXint e;
  if(numtableitems>=0){
    for(e=0; e<numtableitems; e++){
      if(table->isItemSelected(e,0)) tablenum++;
      }
    }
  return tablenum;
  }




/****************************DIRECTORY LIST**********************************/

// Active panel switched
long FXMovieManager::onCmdPanel(FXObject*,FXSelector,void* ptr){
  FXTRACE((1,"Panel = %d\n",(FXint)(FXival)ptr));
  return 1;
  }


// Create new directory
long FXMovieManager::onCmdNewDir_lp(FXObject*,FXSelector,void*){
  FXDialogBox dialog(this,"Create New Directory",DECOR_TITLE|DECOR_BORDER);
  const FXchar suggestedname[]="DirectoryName";
  FXVerticalFrame* content=new FXVerticalFrame(&dialog,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,10,10,10,10);
  new FXLabel(content,"Create new directory in: "+filelist_lp->getDirectory(),NULL,LAYOUT_FILL_X|JUSTIFY_LEFT);
  FXTextField *text=new FXTextField(content,40,&dialog,FXDialogBox::ID_ACCEPT,TEXTFIELD_ENTER_ONLY|FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X);
  new FXHorizontalSeparator(content,SEPARATOR_GROOVE|LAYOUT_FILL_X);
  FXHorizontalFrame* buttons=new FXHorizontalFrame(content,LAYOUT_FILL_X|PACK_UNIFORM_WIDTH,0,0,0,0,0,0,0,0);
  new FXButton(buttons,"&YES",m_accepticon,&dialog,FXDialogBox::ID_ACCEPT,BUTTON_INITIAL|ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT);
  new FXButton(buttons,"&Cancel",m_cancelicon,&dialog,FXDialogBox::ID_CANCEL,ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,0,0,0,0,20,20);
  dialog.create();

  text->setText(suggestedname);
  text->setFocus();
  text->setSelection(0,sizeof(suggestedname));

  if(dialog.execute()){
    FXString dirname=FXPath::absolute(filelist_lp->getDirectory(),text->getText());
    if(FXStat::exists(dirname)){
      FXMessageBox::error(this,MBOX_OK,"Already Exists","File or directory %s already exists.\n",dirname.text());
      }
    else if(!FXDir::create(dirname)){
      FXMessageBox::error(this,MBOX_OK,"Cannot Create","Cannot create directory %s.\n",dirname.text());
      }
    else {
      currentdir_lp=dirname;
      dirlist_lp->setDirectory(currentdir_lp);
      filelist_lp->setDirectory(currentdir_lp);
    }
  }
  return 1;
}

// Create new directory
long FXMovieManager::onCmdNewDir_rp(FXObject*,FXSelector,void*){
  FXDialogBox dialog(this,"Create New Directory",DECOR_TITLE|DECOR_BORDER);
  const FXchar suggestedname[]="DirectoryName";
  FXVerticalFrame* content=new FXVerticalFrame(&dialog,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,10,10,10,10);
  new FXLabel(content,"Create new directory in: "+filelist_rp->getDirectory(),NULL,LAYOUT_FILL_X|JUSTIFY_LEFT);
  FXTextField *text=new FXTextField(content,40,&dialog,FXDialogBox::ID_ACCEPT,TEXTFIELD_ENTER_ONLY|FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X);
  new FXHorizontalSeparator(content,SEPARATOR_GROOVE|LAYOUT_FILL_X);
  FXHorizontalFrame* buttons=new FXHorizontalFrame(content,LAYOUT_FILL_X|PACK_UNIFORM_WIDTH,0,0,0,0,0,0,0,0);
  new FXButton(buttons,"&YES",m_accepticon,&dialog,FXDialogBox::ID_ACCEPT,BUTTON_INITIAL|ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT);
  new FXButton(buttons,"&Cancel",m_cancelicon,&dialog,FXDialogBox::ID_CANCEL,ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,0,0,0,0,20,20);
  dialog.create();

  text->setText(suggestedname);
  text->setFocus();
  text->setSelection(0,sizeof(suggestedname));

  if(dialog.execute()){
    FXString dirname=FXPath::absolute(filelist_rp->getDirectory(),text->getText());
    if(FXStat::exists(dirname)){
      FXMessageBox::error(this,MBOX_OK,"Already Exists","File or directory %s already exists.\n",dirname.text());
      }
    else if(!FXDir::create(dirname)){
      FXMessageBox::error(this,MBOX_OK,"Cannot Create","Cannot create directory %s.\n",dirname.text());
      }
    else {
      currentdir_lp=dirname;
      dirlist_rp->setDirectory(currentdir_lp);
      filelist_rp->setDirectory(currentdir_lp);
    }
  }
  return 1;
}




// Delete directory the directory list
long FXMovieManager::onCmdDirDelete_lp(FXObject*,FXSelector,void*){
  FXuint answer=FXMessageBox::warning(this,MBOX_YES_NO,"Deleting Directory if empty","Are you sure you want to delete this directory?");
  if(answer==MBOX_CLICKED_YES){

    FXString dropdir=dirlist_lp->getDirectory();
    FXTRACE((100,"dropdir=%s\n",dropdir.text()));
        if(!FXDir::remove(dropdir)){
          if(MBOX_CLICKED_NO==FXMessageBox::error(this,MBOX_YES_NO,"Error Deleting Directory","Unable to delete directory: %s\nContinue with operation?",dropdir.text())){
            return 1;
            }
        } else {
	  		currentdir_lp=FXPath::absolute(dirlist_lp->getDirectory());
  	  		dirlist_lp->setDirectory(currentdir_lp);
  	  		filelist_lp->setDirectory(currentdir_lp);
		}
    }
  return 1;
  }

// Delete directory the directory list
long FXMovieManager::onCmdDirDelete_rp(FXObject*,FXSelector,void*){
  FXuint answer=FXMessageBox::warning(this,MBOX_YES_NO,"Deleting Directory if empty","Are you sure you want to delete this directory?");
  if(answer==MBOX_CLICKED_YES){

    FXString dropdir=dirlist_rp->getDirectory();
    FXTRACE((100,"dropdir=%s\n",dropdir.text()));
        if(!FXDir::remove(dropdir)){
          if(MBOX_CLICKED_NO==FXMessageBox::error(this,MBOX_YES_NO,"Error Deleting Directory","Unable to delete directory: %s\nContinue with operation?",dropdir.text())){
            return 1;
            }
        } else {
	  		currentdir_lp=FXPath::absolute(dirlist_rp->getDirectory());
  	  		dirlist_rp->setDirectory(currentdir_lp);
  	  		filelist_rp->setDirectory(currentdir_lp);
		}
    }
  return 1;
  }


// Rename directory in the directory list
long FXMovieManager::onCmdDirRename_lp(FXObject*,FXSelector,void*){

  FXString dirname=dirlist_lp->getDirectory();

  FXDialogBox dialog(this,"Rename Directory",DECOR_TITLE|DECOR_BORDER);
  FXVerticalFrame* content=new FXVerticalFrame(&dialog,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,10,10,10,10);
  new FXLabel(content,"Old Directory Name: "+dirname,NULL,LAYOUT_FILL_X|JUSTIFY_LEFT);
  FXTextField *text=new FXTextField(content,40,&dialog,FXDialogBox::ID_ACCEPT,TEXTFIELD_ENTER_ONLY|FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X);
  new FXHorizontalSeparator(content,SEPARATOR_GROOVE|LAYOUT_FILL_X);
  FXHorizontalFrame* buttons=new FXHorizontalFrame(content,LAYOUT_FILL_X|PACK_UNIFORM_WIDTH,0,0,0,0,0,0,0,0);
  new FXButton(buttons,"&YES",m_accepticon,&dialog,FXDialogBox::ID_ACCEPT,BUTTON_INITIAL|ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT);
  new FXButton(buttons,"&Cancel",m_cancelicon,&dialog,FXDialogBox::ID_CANCEL,ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,0,0,0,0,20,20);
  dialog.create();

  text->setText(dirname);
  text->setFocus();

  if(dialog.execute()){

      //FXString suggestedname=FXPath::absolute(dirlist_lp->getDirectory(),text->getText());
      FXString suggestedname=FXPath::absolute(text->getText());

      if(!FXDir::rename(dirname,suggestedname)){
        FXMessageBox::error(this,MBOX_OK,"Cannot Rename","Cannot Rename directory %s.\n",dirname.text());
      }
      else{
		//currentdir_lp=FXPath::absolute(dirlist_lp->getDirectory(),text->getText());
		currentdir_lp=FXPath::absolute(text->getText());
  		dirlist_lp->setDirectory(currentdir_lp);
  		filelist_lp->setDirectory(currentdir_lp);
      }	
   } 
   return 1;
  }

// Rename directory in the directory list
long FXMovieManager::onCmdDirRename_rp(FXObject*,FXSelector,void*){

  FXString dirname=dirlist_rp->getDirectory();

  FXDialogBox dialog(this,"Rename Directory",DECOR_TITLE|DECOR_BORDER);
  FXVerticalFrame* content=new FXVerticalFrame(&dialog,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,10,10,10,10);
  new FXLabel(content,"Old Directory Name: "+dirname,NULL,LAYOUT_FILL_X|JUSTIFY_LEFT);
  FXTextField *text=new FXTextField(content,40,&dialog,FXDialogBox::ID_ACCEPT,TEXTFIELD_ENTER_ONLY|FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X);
  new FXHorizontalSeparator(content,SEPARATOR_GROOVE|LAYOUT_FILL_X);
  FXHorizontalFrame* buttons=new FXHorizontalFrame(content,LAYOUT_FILL_X|PACK_UNIFORM_WIDTH,0,0,0,0,0,0,0,0);
  new FXButton(buttons,"&YES",m_accepticon,&dialog,FXDialogBox::ID_ACCEPT,BUTTON_INITIAL|ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT);
  new FXButton(buttons,"&Cancel",m_cancelicon,&dialog,FXDialogBox::ID_CANCEL,ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,0,0,0,0,20,20);
  dialog.create();

  text->setText(dirname);
  text->setFocus();

  if(dialog.execute()){

      //FXString suggestedname=FXPath::absolute(dirlist_rp->getDirectory(),text->getText());
      FXString suggestedname=FXPath::absolute(text->getText());

      if(!FXDir::rename(dirname,suggestedname)){
        FXMessageBox::error(this,MBOX_OK,"Cannot Rename","Cannot Rename directory %s.\n",dirname.text());
      }
      else{
		//currentdir_lp=FXPath::absolute(dirlist_lp->getDirectory(),text->getText());
		currentdir_lp=FXPath::absolute(text->getText());
  		dirlist_rp->setDirectory(currentdir_lp);
  		filelist_rp->setDirectory(currentdir_lp);
      }	
   } 
   return 1;
  }


// Update create new directory
long FXMovieManager::onUpdNewDir_lp(FXObject* sender,FXSelector,void*){
  FXString path=filelist_lp->getDirectory();
  if(FXStat::isWritable(path))
    sender->handle(this,FXSEL(SEL_COMMAND,ID_ENABLE),NULL);
  else
    sender->handle(this,FXSEL(SEL_COMMAND,ID_DISABLE),NULL);
  return 1;
  }

// Update create new directory
long FXMovieManager::onUpdNewDir_rp(FXObject* sender,FXSelector,void*){
  FXString path=filelist_rp->getDirectory();
  if(FXStat::isWritable(path))
    sender->handle(this,FXSEL(SEL_COMMAND,ID_ENABLE),NULL);
  else
    sender->handle(this,FXSEL(SEL_COMMAND,ID_DISABLE),NULL);
  return 1;
  }


// Change the directory from the FXDirList
long FXMovieManager::onCmdDirectory_lp(FXObject*,FXSelector,void* ptr){
  FXTreeItem *item=(FXTreeItem*)ptr;
  FXString path=dirlist_lp->getItemPathname(item);
  filelist_lp->setDirectory(path);
  showFilelistThumbnails_lp();
  return 1;
  }

// Change the directory from the FXDirList
long FXMovieManager::onCmdDirectory_rp(FXObject*,FXSelector,void* ptr){
  FXTreeItem *item=(FXTreeItem*)ptr;
  FXString path=dirlist_rp->getItemPathname(item);
  filelist_rp->setDirectory(path);
  showFilelistThumbnails_rp();
  return 1;
  }


// Move up one directory
long FXMovieManager::onCmdUpDirectory_lp(FXObject*,FXSelector,void*){
  FXString path=FXPath::upLevel(filelist_lp->getDirectory());
  filelist_lp->setDirectory(path);
  dirlist_lp->setDirectory(path);
  showFilelistThumbnails_lp();
  return 1;
  }

// Move up one directory
long FXMovieManager::onCmdUpDirectory_rp(FXObject*,FXSelector,void*){
  FXString path=FXPath::upLevel(filelist_rp->getDirectory());
  filelist_rp->setDirectory(path);
  dirlist_rp->setDirectory(path);
  showFilelistThumbnails_rp();
  return 1;
  }


// Popup menu for item in dir list
long FXMovieManager::onDirListPopup_lp(FXObject*,FXSelector,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  if(event->moved) return 1;

  FXMenuPane pane(dirlist_lp);

    FXLabel *dirlist_lplabel = new FXLabel(&pane,"Directory Item");
    dirlist_lplabel->setTextColor(FXRGB(120,0,0));

    new FXMenuCommand(&pane,"Map Host",m_mapicon,this,ID_MAPHOST_LP);
    new FXMenuCommand(&pane,"Unmap Host",m_unmapicon,this,ID_UNMAPHOST_LP);
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"New Dir...",m_appendicon,this,ID_NEWDIR_LP);
    new FXMenuCommand(&pane,"Rename Dir",m_renamediricon,this,ID_DIRRENAME_LP);
    new FXMenuCommand(&pane,"Show/Hide",m_showhiddenicon,dirlist_lp,FXDirList::ID_TOGGLE_HIDDEN);
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"Delete Dir",m_deleteicon,this,ID_DIRDELETE_LP);



  pane.create();
  pane.popup(NULL,event->root_x,event->root_y);
  getApp()->runModalWhileShown(&pane);
  return 1;
  }

// Popup menu for item in dir list
long FXMovieManager::onDirListPopup_rp(FXObject*,FXSelector,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  if(event->moved) return 1;

  FXMenuPane pane(dirlist_rp);

    FXLabel *dirlist_lplabel = new FXLabel(&pane,"Directory Item");
    dirlist_lplabel->setTextColor(FXRGB(120,0,0));

    new FXMenuCommand(&pane,"Map Host",m_mapicon,this,ID_MAPHOST_RP);
    new FXMenuCommand(&pane,"Unmap Host",m_unmapicon,this,ID_UNMAPHOST_RP);
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"New Dir...",m_appendicon,this,ID_NEWDIR_RP);
    new FXMenuCommand(&pane,"Rename Dir",m_renamediricon,this,ID_DIRRENAME_RP);
    new FXMenuCommand(&pane,"Show/Hide",m_showhiddenicon,dirlist_lp,FXDirList::ID_TOGGLE_HIDDEN);
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"Delete Dir",m_deleteicon,this,ID_DIRDELETE_RP);

  pane.create();
  pane.popup(NULL,event->root_x,event->root_y);
  getApp()->runModalWhileShown(&pane);
  return 1;
  }


/****************************FILE LIST**********************************/

// Popup menu for item in file list
long FXMovieManager::onFileListPopup_lp(FXObject*,FXSelector,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  if(event->moved) return 1;
  FXint index=filelist_lp->getItemAt(event->win_x,event->win_y);
  FXMenuPane pane(filelist_lp);


  // We clicked in the background
  if(index<0){
    FXLabel *filebacklabel = new FXLabel(&pane,"Background");
    filebacklabel->setTextColor(FXRGB(120,0,0));

    new FXMenuCommand(&pane,"Terminal",m_terminalicon,this,ID_TERMINAL_LP);
    new FXMenuCommand(&pane,"Run here",m_consoleicon,this,ID_COMMAND_LP);
#ifdef HAVE_LIBCLAMAV
    new FXMenuCommand(&pane,"Antivirus",m_bugicon,this,ID_SCANVIRUS);
#endif
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"New Dir...",m_appendicon,this,ID_NEWDIR_LP);
    new FXMenuCommand(&pane,"Show/Hide",m_showhiddenicon,filelist_lp,FXFileList::ID_TOGGLE_HIDDEN);
    new FXMenuCommand(&pane,"Refresh",m_refreshicon,filelist_lp,FXFileList::ID_REFRESH);
    }

  // We clicked on an item file or directory
  else{
    FXLabel *filelist_lplabel = new FXLabel(&pane,"File Item");
    filelist_lplabel->setTextColor(FXRGB(120,0,0));

    new FXMenuCommand(&pane,"Open...",m_appendicon,this,ID_OPEN_LP);
    new FXMenuCommand(&pane,"Open With",m_appendicon,this,ID_OPENWITH_LP);
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"Properties",m_propicon,this,ID_PROPERTIES_LP);
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"Rename",m_renameicon,this,ID_FILERENAME_LP);
    new FXMenuCommand(&pane,"Move--->",m_renameicon,this,ID_FILEMOVE_LP);
    new FXMenuCommand(&pane,"Symlink--->",m_moveicon,this,ID_FILESYMLINK_LP);
    new FXMenuCommand(&pane,"Copy--->",m_copyicon,this,ID_FILECOPY_LP);
    new FXMenuSeparator(&pane);
    // Archive menu
    archivemenu=new FXMenuPane(&pane);
    new FXMenuCommand(archivemenu," Extract",m_archiveicon,this,ID_EXTRACT_LP);
#ifndef DISABLE_TAR
    new FXMenuCommand(archivemenu," Tar",m_archiveicon,this,ID_COMPRESS_TAR_LP);
#endif
    // Compress menu
    compressmenu=new FXMenuPane(archivemenu);
#ifndef DISABLE_GZIP
    new FXMenuCommand(compressmenu," GZIP",m_archiveicon,this,ID_COMPRESS_GZ_LP);
#endif
#ifndef DISABLE_BZIP2
    new FXMenuCommand(compressmenu," BZIP2",m_archiveicon,this,ID_COMPRESS_BZ2_LP);
#endif
#ifndef DISABLE_XZIP
    new FXMenuCommand(compressmenu," XZIP",m_archiveicon,this,ID_COMPRESS_XZ_LP);
#endif
#ifndef DISABLE_SEVENZIP
    new FXMenuCommand(compressmenu," 7ZIP",m_archiveicon,this,ID_COMPRESS_7ZIP_LP);
#endif
#ifndef DISABLE_ZIP
    new FXMenuCommand(compressmenu," ZIP",m_archiveicon,this,ID_COMPRESS_ZIP_LP);
#endif


    new FXMenuCascade(archivemenu,"Compress",m_archiveicon,compressmenu);

    new FXMenuCascade(&pane,"Archive",m_archiveicon,archivemenu);

    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"DELETE...",m_deleteicon,this,ID_FILEDELETE_LP);
    }
  pane.create();
  pane.popup(NULL,event->root_x,event->root_y);
  getApp()->runModalWhileShown(&pane);
  return 1;
  }

// Popup menu for item in file list
long FXMovieManager::onFileListPopup_rp(FXObject*,FXSelector,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  if(event->moved) return 1;
  FXint index=filelist_rp->getItemAt(event->win_x,event->win_y);
  FXMenuPane pane(filelist_rp);


  // We clicked in the background
  if(index<0){
    FXLabel *filebacklabel = new FXLabel(&pane,"Background");
    filebacklabel->setTextColor(FXRGB(120,0,0));

    new FXMenuCommand(&pane,"Terminal",m_terminalicon,this,ID_TERMINAL_RP);
    new FXMenuCommand(&pane,"Run here",m_consoleicon,this,ID_COMMAND_RP);
#ifdef HAVE_LIBCLAMAV
    new FXMenuCommand(&pane,"Antivirus",m_bugicon,this,ID_SCANVIRUS);
#endif
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"New Dir...",m_appendicon,this,ID_NEWDIR_RP);
    new FXMenuCommand(&pane,"Show/Hide",m_showhiddenicon,filelist_rp,FXFileList::ID_TOGGLE_HIDDEN);
    new FXMenuCommand(&pane,"Refresh",m_refreshicon,filelist_rp,FXFileList::ID_REFRESH);
    }

  // We clicked on an item file or directory
  else{
    FXLabel *filelist_lplabel = new FXLabel(&pane,"File Item");
    filelist_lplabel->setTextColor(FXRGB(120,0,0));

    new FXMenuCommand(&pane,"Open...",m_appendicon,this,ID_OPEN_RP);
    new FXMenuCommand(&pane,"Open With",m_appendicon,this,ID_OPENWITH_RP);
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"Properties",m_propicon,this,ID_PROPERTIES_RP);
    new FXMenuSeparator(&pane);
    new FXMenuCommand(&pane,"Rename",m_renameicon,this,ID_FILERENAME_RP);
    new FXMenuCommand(&pane,"<---Move",m_renameicon,this,ID_FILEMOVE_RP);
    new FXMenuCommand(&pane,"<---Symlink",m_moveicon,this,ID_FILESYMLINK_RP);
    new FXMenuCommand(&pane,"<---Copy",m_copyicon,this,ID_FILECOPY_RP);
    new FXMenuSeparator(&pane);

    // Archive menu
    archivemenu=new FXMenuPane(&pane);
    new FXMenuCommand(archivemenu," Extract",m_archiveicon,this,ID_EXTRACT_RP);
#ifndef DISABLE_TAR
    new FXMenuCommand(archivemenu," Tar",m_archiveicon,this,ID_COMPRESS_TAR_RP);
#endif
    // Compress menu
    compressmenu=new FXMenuPane(archivemenu);
#ifndef DISABLE_GZIP
    new FXMenuCommand(compressmenu," GZIP",m_archiveicon,this,ID_COMPRESS_GZ_RP);
#endif
#ifndef DISABLE_BZIP2
    new FXMenuCommand(compressmenu," BZIP2",m_archiveicon,this,ID_COMPRESS_BZ2_RP);
#endif
#ifndef DISABLE_XZIP
    new FXMenuCommand(compressmenu," XZIP",m_archiveicon,this,ID_COMPRESS_XZ_RP);
#endif
#ifndef DISABLE_SEVENZIP
    new FXMenuCommand(compressmenu," 7ZIP",m_archiveicon,this,ID_COMPRESS_7ZIP_RP);
#endif
#ifndef DISABLE_ZIP
    new FXMenuCommand(compressmenu," ZIP",m_archiveicon,this,ID_COMPRESS_ZIP_RP);
#endif


    new FXMenuCascade(archivemenu,"Compress",m_archiveicon,compressmenu);

    new FXMenuCascade(&pane,"Archive",m_archiveicon,archivemenu);

    new FXMenuSeparator(&pane);

    new FXMenuCommand(&pane,"DELETE...",m_deleteicon,this,ID_FILEDELETE_RP);
    }
  pane.create();
  pane.popup(NULL,event->root_x,event->root_y);
  getApp()->runModalWhileShown(&pane);
  return 1;
  }




// Command from the tree list
long FXMovieManager::onCmdFileListDC_lp(FXObject*,FXSelector,void* ptr){
  FXint index=(FXint)(FXival)ptr;

  if(0<=index){
    if(filelist_lp->isItemDirectory(index)){
      currentdir_lp = filelist_lp->getItemPathname(index);
      filelist_lp->setDirectory(currentdir_lp);
      dirlist_lp->setDirectory(currentdir_lp);
      showFilelistThumbnails_lp();
      }
    else if(filelist_lp->isItemFile(index)){
      filename=filelist_lp->getItemPathname(index);
      FXString ext=FXPath::extension(filename);
	if(loadfilevideo(filename)==TRUE){
	  table->setItemText(numtableitems,0,filename);
	  numtableitems=numtableitems+1;
          getFilelistSelection_lp();
	}
	else if(loadfilevideo(filename)==FALSE){
	  FXMessageBox::error(this,MBOX_OK,"Error on loading Movie","Unsupported Type: %s",ext.text());
	}
      }
    }
  return 1;
}

// Command from the tree list
long FXMovieManager::onCmdFileListDC_rp(FXObject*,FXSelector,void* ptr){
  FXint index=(FXint)(FXival)ptr;

  if(0<=index){
    if(filelist_rp->isItemDirectory(index)){
      currentdir_rp = filelist_rp->getItemPathname(index);
      filelist_rp->setDirectory(currentdir_rp);
      dirlist_rp->setDirectory(currentdir_rp);
      showFilelistThumbnails_rp();
      }
    else if(filelist_rp->isItemFile(index)){
      filename=filelist_rp->getItemPathname(index);
      FXString ext=FXPath::extension(filename);
	if(loadfilevideo(filename)==TRUE){
	  table->setItemText(numtableitems,0,filename);
	  numtableitems=numtableitems+1;
          getFilelistSelection_rp();
	}
	else if(loadfilevideo(filename)==FALSE){
	  FXMessageBox::error(this,MBOX_OK,"Error on loading Movie","Unsupported Type: %s",ext.text());
	}
      }
    }
  return 1;
}



// Symlink file or directory
long FXMovieManager::onCmdFileSymlink_lp(FXObject*,FXSelector,void*){
	cmdFileSymlink(filelist_lp, filelist_rp);
  return 1;
  }

long FXMovieManager::onCmdFileSymlink_rp(FXObject*,FXSelector,void*){
	cmdFileSymlink(filelist_rp, filelist_lp);
  return 1;
  }

// Symlink file or directory
void FXMovieManager::cmdFileSymlink(FXFileList* panel, FXFileList* out_panel){
	deselectMainDir(panel);
	FXString rootname,imagename,newname;
  	FXString *filenamelist=getFilenames(panel);

  	if(filenamelist){
    	OperationsDialog operationsdialog(this,"FileOperation: Symlink");
    	for(FXint i=0; !filenamelist[i].empty(); i++){
		if(loadfilevideo(filenamelist[i])==TRUE) {
			rootname="."+FXPath::stripExtension(FXPath::name(filenamelist[i]));
			// printf("rootname: %s\n",rootname.text());
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        	imagename=FXPath::enquote(panel->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        	imagename=FXPath::enquote(panel->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
			// printf("imagename: %s\n",imagename.text());

		} else if (loadfilevideo(filenamelist[i])==FALSE && loadfileaudio(filenamelist[i])==TRUE){
			imagename="/usr/local/share/pixmaps/audio_file.jpg";

		} else if (loadfilevideo(filenamelist[i])==FALSE && loadfileaudio(filenamelist[i])==FALSE){

  			if(FXStat::isDirectory(filenamelist[i])){
				imagename="/usr/local/share/pixmaps/unknow_dir.jpg";
    			}
  			else if(FXStat::isExecutable(filenamelist[i])){
				imagename="/usr/local/share/pixmaps/unknow_exe.jpg";
    			}
			else{
				FXString ext=FXPath::extension(FXPath::name(filenamelist[i]));
			    // printf("extension: %s\n",ext.text());
				if(comparecase(ext,"gz")==0||comparecase(ext,"xz")==0||comparecase(ext,"lzma")==0||comparecase(ext,"bz2")==0||comparecase(ext,"zip")==0||comparecase(ext,"tgz")==0||comparecase(ext,"7z")==0||comparecase(ext,"tar")==0){
					imagename="/usr/local/share/pixmaps/unknow_arch.jpg";
				}
				else if(comparecase(ext,"gif")==0||comparecase(ext,"tiff")==0||comparecase(ext,"png")==0||comparecase(ext,"jpg")==0||comparecase(ext,"xpm")==0||comparecase(ext,"bmp")==0){
					imagename="/usr/local/share/pixmaps/unknow_img.jpg";
				}
  				else{
					imagename="/usr/local/share/pixmaps/unknow_file.jpg";
    			}
			}
		}
			// printf("imagename: %s\n",imagename.text());
			// upgrade image if renamed more than one movie file
  			FXImage *img=NULL;
  			FXImage *old=NULL;
#ifdef HAVE_LIBFOX_1_6
  			img=new FXImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
#endif
#ifdef HAVE_LIBFOX_1_7
			img=new FXJPGImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
#endif
			// Perhaps failed
  			if(img){
    			FXFileStream stream;
    			// Load image
    			if(stream.open(imagename,FXStreamLoad)){
    				getApp()->beginWaitCursor();
    				img->loadPixels(stream);
    				stream.close();
   					// Create image
   					img->create();
   					img->scale(400,320,FXCLAMP(0,1,1));
   					// Set new image, deleting old
   					old=operationsdialog.getNewImage();
   					operationsdialog.setNewImage(img);	
   					delete old;
   					getApp()->endWaitCursor();
				}
    		}

      		operationsdialog.setOldName(FXPath::absolute(FXPath::directory(filenamelist[i]),FXPath::name(filenamelist[i])));
      		//	operationsdialog.setNewName(FXPath::absolute(FXPath::directory(filenamelist[i]),FXPath::name(filenamelist[i])));
	  		FXString destinationdir=out_panel->getDirectory();
      		operationsdialog.setNewName(FXPath::absolute(destinationdir)+"/"+FXPath::name(filenamelist[i]));
      		if(!operationsdialog.execute()) break;
      		newname=operationsdialog.getNewName();

    		if(FXStat::exists(newname)){
		    	if(FXMessageBox::error(this,MBOX_OK,"Error this Link already Exists","Unable to Link:\n\n%s  to:  %s",filenamelist[i].text(),newname.text())==MBOX_CLICKED_OK) break;
      		}
		if(!FXFile::symlink(filenamelist[i],FXPath::absolute(FXPath::directory(filenamelist[i]),newname))){
		    if(FXMessageBox::error(this,MBOX_YES_NO,"Error Linking","Unable to Link:\n\n%s  to:  %s\n\nContinue with operation?",filenamelist[i].text(),newname.text())==MBOX_CLICKED_NO) break;
		}else{
			FXFile::symlink(filenamelist[i],FXPath::absolute(FXPath::directory(filenamelist[i]),newname));
       		// recreate imageicon
			if(loadfilevideo(newname)==TRUE){
        		rootname="."+FXPath::name(FXPath::stripExtension(newname));
				if(!createImageMovie(newname,rootname)){
					FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Creation Failure","Error on creating thumbnail: %s",imagename.text());
    				if(answer==MBOX_CLICKED_CANCEL){
      					return;
    				}
				}
				if(autoThumbnailUpdate==TRUE){
					out_panel->update();
					out_panel->repaint();
					showThumbnails(out_panel);
					//printf("updating thumbnails\n");
				}
			}
		}
   	  }
    	    delete [] filenamelist;
  	}
  return;
  }



// Copy file or directory
long FXMovieManager::onCmdFileCopy_lp(FXObject*,FXSelector,void*){
	cmdFileCopy(filelist_lp, filelist_rp);
  return 1;
  }

long FXMovieManager::onCmdFileCopy_rp(FXObject*,FXSelector,void*){
	cmdFileCopy(filelist_rp, filelist_lp);
  return 1;
  }

void FXMovieManager::cmdFileCopy(FXFileList* panel, FXFileList* out_panel){
	deselectMainDir(panel);
	FXString rootname,imagename,newname;
  	FXString *filenamelist=getFilenames(panel);
	FXbool overwrite_file=TRUE;

  	if(filenamelist){
    	OperationsDialog operationsdialog(this,"FileOperation: Copy");
    	for(FXint i=0; !filenamelist[i].empty(); i++){
		if(loadfilevideo(filenamelist[i])==TRUE) {
			rootname="."+FXPath::stripExtension(FXPath::name(filenamelist[i]));
			// printf("rootname: %s\n",rootname.text());
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        	imagename=FXPath::enquote(panel->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        	imagename=FXPath::enquote(panel->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
			// printf("imagename: %s\n",imagename.text());

		} else if (loadfilevideo(filenamelist[i])==FALSE && loadfileaudio(filenamelist[i])==TRUE){
			imagename="/usr/local/share/pixmaps/audio_file.jpg";

		} else if (loadfilevideo(filenamelist[i])==FALSE && loadfileaudio(filenamelist[i])==FALSE){

  			if(FXStat::isDirectory(filenamelist[i])){
				imagename="/usr/local/share/pixmaps/unknow_dir.jpg";
    			}
  			else if(FXStat::isExecutable(filenamelist[i])){
				imagename="/usr/local/share/pixmaps/unknow_exe.jpg";
    			}
			else{
				FXString ext=FXPath::extension(FXPath::name(filenamelist[i]));
			    // printf("extension: %s\n",ext.text());
				if(comparecase(ext,"gz")==0||comparecase(ext,"xz")==0||comparecase(ext,"lzma")==0||comparecase(ext,"bz2")==0||comparecase(ext,"zip")==0||comparecase(ext,"tgz")==0||comparecase(ext,"7z")==0||comparecase(ext,"tar")==0){
					imagename="/usr/local/share/pixmaps/unknow_arch.jpg";
				}
				else if(comparecase(ext,"gif")==0||comparecase(ext,"tiff")==0||comparecase(ext,"png")==0||comparecase(ext,"jpg")==0||comparecase(ext,"xpm")==0||comparecase(ext,"bmp")==0){
					imagename="/usr/local/share/pixmaps/unknow_img.jpg";
				}
  				else{
					imagename="/usr/local/share/pixmaps/unknow_file.jpg";
    			}
			}
		}
			// printf("imagename: %s\n",imagename.text());
			// upgrade image if renamed more than one movie file
  			FXImage *img=NULL;
  			FXImage *old=NULL;
#ifdef HAVE_LIBFOX_1_6
  			img=new FXImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
#endif
#ifdef HAVE_LIBFOX_1_7
			img=new FXJPGImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
#endif
			// Perhaps failed
  			if(img){
    			FXFileStream stream;
    			// Load image
    			if(stream.open(imagename,FXStreamLoad)){
    				getApp()->beginWaitCursor();
    				img->loadPixels(stream);
    				stream.close();
   					// Create image
   					img->create();
   					img->scale(400,320,FXCLAMP(0,1,1));
   					// Set new image, deleting old
   					old=operationsdialog.getNewImage();
   					operationsdialog.setNewImage(img);	
   					delete old;
   					getApp()->endWaitCursor();
				}
    		}

      		operationsdialog.setOldName(FXPath::absolute(FXPath::directory(filenamelist[i]),FXPath::name(filenamelist[i])));
      		//	operationsdialog.setNewName(FXPath::absolute(FXPath::directory(filenamelist[i]),"_"+FXPath::name(filenamelist[i])));
	  		FXString destinationdir=out_panel->getDirectory();
      		operationsdialog.setNewName(FXPath::absolute(destinationdir)+"/"+FXPath::name(filenamelist[i]));
      		if(!operationsdialog.execute()) break;
			newname=operationsdialog.getNewName();

    		if(FXStat::exists(newname)){
      		    if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"This Copy Already Exists","Overwrite Existing Copy: %s?",filenamelist[i].text())) {
		    		overwrite_file=FALSE;
		    		break;
		    	}
      		}
			if(!FXFile::copyFiles(filenamelist[i],FXPath::absolute(FXPath::directory(filenamelist[i]),newname),overwrite_file)){
		    	if(FXMessageBox::error(this,MBOX_YES_NO,"Error Copying","Unable to Copy:\n\n%s  to:  %s\n\nContinue with operation?",filenamelist[i].text(),newname.text())==MBOX_CLICKED_NO) break;
			}else{
				FXFile::copyFiles(filenamelist[i],FXPath::absolute(FXPath::directory(filenamelist[i]),newname),overwrite_file);
       			// recreate imageicon
				if(loadfilevideo(newname)==TRUE){
					//printf("newname=%s\n",newname.text());
        			rootname="."+FXPath::name(FXPath::stripExtension(newname));
					//printf("rootname=%s\n",rootname.text());
					if(!createImageMovie(newname,rootname)){
						FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Creation Failure","Error on creating thumbnail: %s",imagename.text());
    					if(answer==MBOX_CLICKED_CANCEL){
      						return;
    					}
					}
					if(autoThumbnailUpdate==TRUE){
						out_panel->update();
						out_panel->repaint();
						showThumbnails(out_panel);
						//printf("updating thumbnails\n");
					}
				}
			}
   	    }
    	delete [] filenamelist;
  	}
  return;
  }


// Move movie file or directory
long FXMovieManager::onCmdFileMove_lp(FXObject*,FXSelector,void*){
	cmdFileRename(filelist_lp, filelist_rp);
  return 1;
  }

long FXMovieManager::onCmdFileMove_rp(FXObject*,FXSelector,void*){
	cmdFileRename(filelist_rp, filelist_lp);
  return 1;
  }

// Rename movie file or directory
long FXMovieManager::onCmdFileRename_lp(FXObject*,FXSelector,void*){
	cmdFileRename(filelist_lp, filelist_lp);
  return 1;
  }

long FXMovieManager::onCmdFileRename_rp(FXObject*,FXSelector,void*){
	cmdFileRename(filelist_rp, filelist_rp);
  return 1;
  }

void FXMovieManager::cmdFileRename(FXFileList* panel, FXFileList* out_panel){
	deselectMainDir(panel);
	FXString rootname,imagename,newname;
  	FXString *filenamelist=getFilenames(panel);
	FXbool overwrite_file=TRUE;

  	if(filenamelist){
    	OperationsDialog operationsdialog(this,"FileOperation: Rename-Move");
    	for(FXint i=0; !filenamelist[i].empty(); i++){
			if(loadfilevideo(filenamelist[i])==TRUE) {
				rootname="."+FXPath::stripExtension(FXPath::name(filenamelist[i]));
				// printf("rootname: %s\n",rootname.text());
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        		imagename=FXPath::enquote(panel->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        		imagename=FXPath::enquote(panel->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
				// printf("imagename: %s\n",imagename.text());

		} else if (loadfilevideo(filenamelist[i])==FALSE && loadfileaudio(filenamelist[i])==TRUE){
			imagename="/usr/local/share/pixmaps/audio_file.jpg";

		} else if (loadfilevideo(filenamelist[i])==FALSE && loadfileaudio(filenamelist[i])==FALSE){

  				if(FXStat::isDirectory(filenamelist[i])){
					imagename="/usr/local/share/pixmaps/unknow_dir.jpg";
  				}else if(FXStat::isExecutable(filenamelist[i])){
					imagename="/usr/local/share/pixmaps/unknow_exe.jpg";
    			}else{
					FXString ext=FXPath::extension(FXPath::name(filenamelist[i]));
			   	 	// printf("extension: %s\n",ext.text());
					if(comparecase(ext,"gz")==0||comparecase(ext,"xz")==0||comparecase(ext,"lzma")==0||comparecase(ext,"bz2")==0||comparecase(ext,"zip")==0||comparecase(ext,"tgz")==0||comparecase(ext,"7z")==0||comparecase(ext,"tar")==0){
						imagename="/usr/local/share/pixmaps/unknow_arch.jpg";
					}else if(comparecase(ext,"gif")==0||comparecase(ext,"tiff")==0||comparecase(ext,"png")==0||comparecase(ext,"jpg")==0||comparecase(ext,"xpm")==0||comparecase(ext,"bmp")==0){
						imagename="/usr/local/share/pixmaps/unknow_img.jpg";
					}else{
						imagename="/usr/local/share/pixmaps/unknow_file.jpg";
    				}
				}
			}

			// upgrade image if renamed more than one movie file
  			FXImage *img=NULL;
  			FXImage *old=NULL;
#ifdef HAVE_LIBFOX_1_6
  		 	img=new FXImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
#endif
#ifdef HAVE_LIBFOX_1_7
		  	img=new FXJPGImage(getApp(),NULL,IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
#endif
			// Perhaps failed
  			if(img){
    			FXFileStream stream;
    			// Load image
    			if(stream.open(imagename,FXStreamLoad)){
    				getApp()->beginWaitCursor();
    				img->loadPixels(stream);
    				stream.close();
   					// Create image
   					img->create();
   					img->scale(400,320,FXCLAMP(0,1,1));
   					// Set new image, deleting old
   					old=operationsdialog.getNewImage();
   					operationsdialog.setNewImage(img);	
   					delete old;
   					getApp()->endWaitCursor();
				}
    		}

      		operationsdialog.setOldName(FXPath::absolute(FXPath::directory(filenamelist[i]),FXPath::name(filenamelist[i])));
	  		FXString destinationdir=out_panel->getDirectory();
      		operationsdialog.setNewName(FXPath::absolute(destinationdir)+"/"+FXPath::name(filenamelist[i]));
       		if(!operationsdialog.execute()) break;
			newname=operationsdialog.getNewName();

    		if(FXStat::exists(newname)){
      	  	  if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"Already Exists","Overwrite Existing: %s?",filenamelist[i].text())) {
		   			overwrite_file=FALSE;
		   			break;
		  	 	}
      		}

			if(!FXFile::moveFiles(filenamelist[i],FXPath::absolute(FXPath::directory(filenamelist[i]),newname),overwrite_file)){
		   		if(FXMessageBox::error(this,MBOX_YES_NO,"Error Renaming/Moving","Unable to Rename/Move:\n\n%s  to:  %s\n\nContinue with operation?",filenamelist[i].text(),newname.text())==MBOX_CLICKED_NO) break;
			}else{
				FXFile::moveFiles(filenamelist[i],FXPath::absolute(FXPath::directory(filenamelist[i]),newname),overwrite_file);
       			// recreate imageicon
				if(loadfilevideo(newname)==TRUE){
					//printf("newname=%s\n",newname.text());
        			rootname="."+FXPath::name(FXPath::stripExtension(newname));
					//printf("rootname=%s\n",rootname.text());
					if(!createImageMovie(newname,rootname)){
						FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Creation Failure","Error on creating thumbnail: %s",imagename.text());
    					if(answer==MBOX_CLICKED_CANCEL){
      						return;
    					}
					}
					if(autoThumbnailUpdate==TRUE){
						out_panel->update();
						out_panel->repaint();
						showThumbnails(out_panel);
						//printf("updating thumbnails\n");
					}
				}
			}
   		}
    	delete [] filenamelist;
  	}
  return;
  }


// Delete file or directory in the file list 
long FXMovieManager::onCmdFileDelete_lp(FXObject*,FXSelector,void*){
	cmdFileDelete(filelist_lp);
  return 1;
  }

long FXMovieManager::onCmdFileDelete_rp(FXObject*,FXSelector,void*){
	cmdFileDelete(filelist_rp);
  return 1;
  }

void FXMovieManager::cmdFileDelete(FXFileList* panel){
  FXint count=0;
  FXString logfiles,filetoremove,file;

  for(FXint i=0; i<panel->getNumItems(); i++){
    if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;
	if(count>20) {logfiles+=".........................\n\nDelete more than 20 files?\n"; break;}
	logfiles+=file+"\n";
        count=count+1;
    }
  }
	
  FXuint answer=FXMessageBox::warning(this,MBOX_YES_NO,"Deleting Movie Files",logfiles.text());
  if(answer==MBOX_CLICKED_YES){

    for(FXint i=0; i<panel->getNumItems(); i++){
      if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;
        filetoremove=FXPath::absolute(panel->getDirectory(),file);
        FXTRACE((100,"filetoremove=%s\n",filetoremove.text()));
        if(!FXFile::removeFiles(filetoremove,TRUE)){
          if(MBOX_CLICKED_NO==FXMessageBox::error(this,MBOX_YES_NO,"Error Deleting Movie","Unable to delete movie: %s\nContinue with operation?",filetoremove.text())){
            break;
            }
          }
        }
      }
    }
  return;
  }


// Enable sender when Movie Files have been selected
long FXMovieManager::onUpdSelectedFile_lp(FXObject* sender,FXSelector,void*){
    FXString file;

    for(FXint i=0; i<filelist_lp->getNumItems(); i++){
    	if(filelist_lp->isItemSelected(i)){
       		file=filelist_lp->getItemFilename(i);
        	if(file=="..") continue;
			if(filelist_lp->isItemDirectory(i)) return 1;
    		if(loadfilevideo(file)==FALSE && loadfileaudio(file)==FALSE)return 1;
		}
    }

    sender->handle(this,getNumFilenames(filelist_lp)?FXSEL(SEL_COMMAND,ID_ENABLE):FXSEL(SEL_COMMAND,ID_DISABLE),NULL);
    return 1;
}

long FXMovieManager::onUpdSelectedFile_rp(FXObject* sender,FXSelector,void*){
    FXString file;

    for(FXint i=0; i<filelist_rp->getNumItems(); i++){
    	if(filelist_rp->isItemSelected(i)){
       		file=filelist_rp->getItemFilename(i);
        	if(file=="..") continue;
			if(filelist_rp->isItemDirectory(i)) return 1;
    		if(loadfilevideo(file)==FALSE && loadfileaudio(file)==FALSE)return 1;
		}
    }

    sender->handle(this,getNumFilenames(filelist_rp)?FXSEL(SEL_COMMAND,ID_ENABLE):FXSEL(SEL_COMMAND,ID_DISABLE),NULL);
    return 1;
}



/*********************************ROUTINES**********************************/


// Get number of selected filenames, not including "." and ".."
FXint FXMovieManager::getNumFilenames(FXFileList* panel) const {
  register FXint num=0;
  register FXint i;
  if(panel->getNumItems()){
    for(i=0; i<panel->getNumItems(); i++){
#ifdef HAVE_LIBFOX_1_6
      if(panel->isItemSelected(i) && panel->getItemFilename(i)!=".." && panel->getItemFilename(i)!=".") num++;
#endif
#ifdef HAVE_LIBFOX_1_7
      if(panel->isItemSelected(i) && !panel->isItemNavigational(i)) num++;
#endif
      }
    }
  return num;
  }


// Return selected filenames, not including "." and ".."
FXString* FXMovieManager::getFilenames(FXFileList* panel) const {
  register FXint num=getNumFilenames(panel);
  register FXString *files=NULL;
  register FXint i,n;
  if(0<num){
    files=new FXString [num+1];
    for(i=n=0; i<panel->getNumItems(); i++){
#ifdef HAVE_LIBFOX_1_6
      if(panel->isItemSelected(i) && panel->getItemFilename(i)!=".." && panel->getItemFilename(i)!="."){
#endif
#ifdef HAVE_LIBFOX_1_7
      if(panel->isItemSelected(i) && !panel->isItemNavigational(i)){
#endif
	files[n++]=panel->getItemPathname(i);
	}
      }
    files[n]=FXString::null;
    }
  return files;
  }



// deselect (..) directory
void FXMovieManager::deselectMainDir(FXFileList* panel){

  	for(FXint i=0; i<panel->getNumItems(); i++){
    	if(panel->isItemSelected(i)){
        	FXString file=panel->getItemFilename(i);
		// prevent crash when renaming directory (..)
        	if(file=="..") {
				panel->deselectItem(i);
				continue;
			}
		}
	}
	return;
}


/*****************************************************************/


// Pop up terminal
long FXMovieManager::onCmdTerminal_lp(FXObject*,FXSelector,void*){
  FXString executable="cd "+FXPath::enquote(filelist_lp->getDirectory())+"; "+terminal+" &";
  system(executable.text());
  return 1;
  }

long FXMovieManager::onCmdTerminal_rp(FXObject*,FXSelector,void*){
  FXString executable="cd "+FXPath::enquote(filelist_rp->getDirectory())+"; "+terminal+" &";
  system(executable.text());
  return 1;
  }


// call command
long FXMovieManager::onCmdCommand_lp(FXObject*,FXSelector,void*){
  cmdRunCommand(filelist_lp);
  return 1;
  }

long FXMovieManager::onCmdCommand_rp(FXObject*,FXSelector,void*){
  cmdRunCommand(filelist_rp);
  return 1;
  }

void FXMovieManager::cmdRunCommand(FXFileList* filepanel){
  FXString runcommand=getApp()->reg().readStringEntry("SETTINGS","runcommand","chmod 666 *");

  if(FXInputDialog::getString(runcommand,this,"Run Command","Command:")){
    getApp()->reg().writeStringEntry("SETTINGS","runcommand",runcommand.text());
    runcommand="cd "+FXPath::enquote(filepanel->getDirectory())+"; "+runcommand+" &";
    // printf("runcommand: %s\n", runcommand.text()); 
    system(runcommand.text());
    }
  return;
}


/****************************MIYME TYPES*********************************/

// Pop up mimetype dialog
long FXMovieManager::onCmdMimeType(FXObject*,FXSelector,void*){
    mimetype->show(PLACEMENT_OWNER);
  return 1;
  }


// associated Command
long FXMovieManager::onCmdOpen_lp(FXObject*,FXSelector,void*){
  cmdOpen(filelist_lp, dirlist_lp);
  return 1;
  }

long FXMovieManager::onCmdOpen_rp(FXObject*,FXSelector,void*){
  cmdOpen(filelist_rp, dirlist_rp);
  return 1;
  }

// associated Command
void FXMovieManager::cmdOpen(FXFileList* filepanel, FXDirList* dirpanel){
  FXint index=filepanel->getCurrentItem();
  if(index>=0){

    // If directory, open the directory
    if(filepanel->isItemDirectory(index)){
      FXString directory=filepanel->getItemPathname(index);
      FXTRACE((100,"directory=%s\n",directory.text()));
      filepanel->setDirectory(directory);
      dirpanel->setDirectory(directory);
      }

    // If regular file return as the selected file
    else if(filepanel->isItemFile(index)){

      FXFileAssoc *fileassoc=filepanel->getItemAssoc(index);
      if(fileassoc){
        if(fileassoc->command.text()){
          FXString command=FXString::value(fileassoc->command.text());
	  FXString runcommand = "cd "+FXPath::enquote(filepanel->getDirectory())+"; "+command+" "+filepanel->getItemFilename(index)+" &";
          FXTRACE((100,"system(%s)\n",command.text()));
	      // printf("runcommand: %s\n", runcommand.text()); 
          system(runcommand.text());
          }
        else{
          FXMessageBox::information(this,MBOX_OK,"Unknown Command","No command has been defined for this file: %s",filepanel->getItemFilename(index).text());
          }
        }
      else{
        FXMessageBox::information(this,MBOX_OK,"Unknown Mime Type","No Association has been set for this file: %s",filepanel->getItemFilename(index).text());
        }
      }
  }
  return;
}



// Open with program
long FXMovieManager::onCmdOpenWith_lp(FXObject*,FXSelector,void*){
  cmdOpenWith(filelist_lp);
  return 1;
  }

long FXMovieManager::onCmdOpenWith_rp(FXObject*,FXSelector,void*){
  cmdOpenWith(filelist_rp);
  return 1;
  }

void FXMovieManager::cmdOpenWith(FXFileList* filepanel){
  FXString command=getApp()->reg().readStringEntry("SETTINGS","command","mplayer");
  FXString filename=filepanel->getCurrentFile();

  if(FXInputDialog::getString(command,this,"Open Movie File","Open " + FXPath::name(filename) + " with:")){
    getApp()->reg().writeStringEntry("SETTINGS","command",command.text());
    command=command+" "+filename+" &";
    // printf("command: %s\n", command.text()); 
    system(command.text());
    }
  return;
}


long FXMovieManager::onUpdItemFound_lp(FXObject* sender,FXSelector,void*){
  sender->handle(this,filelist_lp->getCurrentFile().empty()?FXSEL(SEL_COMMAND,ID_DISABLE):FXSEL(SEL_COMMAND,ID_ENABLE),NULL);
  return 1;
}

long FXMovieManager::onUpdItemFound_rp(FXObject* sender,FXSelector,void*){
  sender->handle(this,filelist_rp->getCurrentFile().empty()?FXSEL(SEL_COMMAND,ID_DISABLE):FXSEL(SEL_COMMAND,ID_ENABLE),NULL);
  return 1;
}




/****************************PROPERTIES**********************************/

// Pop up properties dialog
long FXMovieManager::onCmdProperties_lp(FXObject*,FXSelector,void*){
  cmdProperties(filelist_lp, properties_lp);
  return 1;
  }

long FXMovieManager::onCmdProperties_rp(FXObject*,FXSelector,void*){
  cmdProperties(filelist_rp, properties_rp);
  return 1;
  }

void FXMovieManager::cmdProperties(FXFileList* filepanel, FXDialogBox* properties){

  FXString removeInfo="test -e "+currentInfo+" && rm "+currentInfo;
  system (removeInfo.text());

  FXString filename=filepanel->getCurrentFile();
  if(!createInfoMovie(filename)){
	FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Information Creation Failure","Error on creating information for: %s",filename.text());
  	if(answer==MBOX_CLICKED_CANCEL){
		return;
	}
  }	
  system("sleep 0.5s");

  FXchar *text;
  FXint size;

  FXFile fvideo(currentInfo,FXFile::Reading);
  size=fvideo.size();

#ifdef HAVE_LIBFOX_1_6
  if(!FXMALLOC(&text,FXchar,size)){
#endif
#ifdef HAVE_LIBFOX_1_7
  if(allocElms(text,size)){
#endif

    fvideo.readBlock(text,size);
    FXString block = text;

    FXRex rexvideo1("VIDEO:",0);
    FXint begv1,endv1;
    FXRex rexvideo2("kbyte",0);
    FXint begv2,endv2;
    if(rexvideo1.match(text,&begv1,&endv1)){
	if(rexvideo2.match(text,&begv2,&endv2)){
	FXint lenv=endv2-endv1+3;
    	FXString seltextv=block.mid(endv1,lenv);
    	grepvideo=seltextv;
        }
    }
    else {grepvideo="NOT DETECTED by mplayer";}

    FXRex rexaudio1("AUDIO:",0);
    FXint bega1,enda1;
    FXRex rexaudio2("->",0);
    FXint bega2,enda2;
    if(rexaudio1.match(text,&bega1,&enda1)){
	if(rexaudio2.match(text,&bega2,&enda2)){
	FXint lena=enda2-enda1+7;
    	FXString seltexta=block.mid(enda1,lena);
    	grepaudio=seltexta;
        }
    }
    else {grepaudio="NOT DETECTED by mplayer";}

  }
  fvideo.close();

#ifdef HAVE_LIBFOX_1_6
  FXFREE(&text);
#endif
#ifdef HAVE_LIBFOX_1_7
  freeElms(text);
#endif

  properties->show(PLACEMENT_OWNER);

  return;
  }


// Update create time
long FXMovieManager::onUpdCreatedTime_lp(FXObject* sender,FXSelector,void*){
#ifdef HAVE_LIBFOX_1_6
  FXString time=FXSystem::time(FXStat::created(filelist_lp->getCurrentFile()));
#endif
#ifdef HAVE_LIBFOX_1_7
  FXString time=FXSystem::localTime(FXStat::created(filelist_lp->getCurrentFile()));
#endif
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&time);
  return 1;
  }

long FXMovieManager::onUpdCreatedTime_rp(FXObject* sender,FXSelector,void*){
#ifdef HAVE_LIBFOX_1_6
  FXString time=FXSystem::time(FXStat::created(filelist_rp->getCurrentFile()));
#endif
#ifdef HAVE_LIBFOX_1_7
  FXString time=FXSystem::localTime(FXStat::created(filelist_rp->getCurrentFile()));
#endif
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&time);
  return 1;
  }


// Update modified time
long FXMovieManager::onUpdModifiedTime_lp(FXObject* sender,FXSelector,void*){
#ifdef HAVE_LIBFOX_1_6
  FXString time=FXSystem::time(FXStat::modified(filelist_lp->getCurrentFile()));
#endif
#ifdef HAVE_LIBFOX_1_7
  FXString time=FXSystem::localTime(FXStat::modified(filelist_lp->getCurrentFile()));
#endif
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&time);
  return 1;
  }

long FXMovieManager::onUpdModifiedTime_rp(FXObject* sender,FXSelector,void*){
#ifdef HAVE_LIBFOX_1_6
  FXString time=FXSystem::time(FXStat::modified(filelist_rp->getCurrentFile()));
#endif
#ifdef HAVE_LIBFOX_1_7
  FXString time=FXSystem::localTime(FXStat::modified(filelist_rp->getCurrentFile()));
#endif
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&time);
  return 1;
  }


// Update file location
long FXMovieManager::onUpdFileLocation_lp(FXObject* sender,FXSelector,void*){
  FXString string=filelist_lp->getCurrentFile();
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&string);
  return 1;
  }

long FXMovieManager::onUpdFileLocation_rp(FXObject* sender,FXSelector,void*){
  FXString string=filelist_rp->getCurrentFile();
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&string);
  return 1;
  }


// Update file size
long FXMovieManager::onUpdFileSize_lp(FXObject* sender,FXSelector,void*){
  FXString size;
#ifdef HAVE_LIBFOX_1_6
  size=FXStringVal(FXStat::size(filelist_lp->getCurrentFile()));
#endif
#ifdef HAVE_LIBFOX_1_7
  size.fromInt(FXStat::size(filelist_lp->getCurrentFile()));
#endif
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&size);
  return 1;
  }

long FXMovieManager::onUpdFileSize_rp(FXObject* sender,FXSelector,void*){
  FXString size;
#ifdef HAVE_LIBFOX_1_6
  size=FXStringVal(FXStat::size(filelist_rp->getCurrentFile()));
#endif
#ifdef HAVE_LIBFOX_1_7
  size.fromInt(FXStat::size(filelist_rp->getCurrentFile()));
#endif
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&size);
  return 1;
  }


// Update file type
long FXMovieManager::onUpdFileType_lp(FXObject* sender,FXSelector,void*){

  FXString filename=filelist_lp->getCurrentFile();
  FXFileAssoc *fileassoc=NULL;
  FXString type;

  if(FXStat::isDirectory(filename)){
    fileassoc=associations->findDirBinding(filename.text());
    type="Directory";
    }
  else if(FXStat::isExecutable(filename)){
    fileassoc=associations->findExecBinding(filename.text());
    type="Executable";
    }
  else{
    fileassoc=associations->findFileBinding(filename.text());
	FXString ext=FXPath::extension(FXPath::name(filename));
	    // printf("extension: %s\n",ext.text());
		if(comparecase(ext,"gz")==0||comparecase(ext,"bz2")==0||comparecase(ext,"zip")==0||comparecase(ext,"tar")==0||comparecase(ext,"tgz")==0||comparecase(ext,"7z")==0){
			type="Archive";
		}
		else if(comparecase(ext,"gif")==0||comparecase(ext,"tiff")==0||comparecase(ext,"png")==0||comparecase(ext,"jpg")==0||comparecase(ext,"xpm")==0||comparecase(ext,"bmp")==0){
			type="Image";
		}
		else{
			type="File";
		}
    }      
    sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&type);
	
  return 1;
  }

long FXMovieManager::onUpdFileType_rp(FXObject* sender,FXSelector,void*){

  FXString filename=filelist_rp->getCurrentFile();
  FXFileAssoc *fileassoc=NULL;
  FXString type;

  if(FXStat::isDirectory(filename)){
    fileassoc=associations->findDirBinding(filename.text());
    type="Directory";
    }
  else if(FXStat::isExecutable(filename)){
    fileassoc=associations->findExecBinding(filename.text());
    type="Executable";
    }
  else{
    fileassoc=associations->findFileBinding(filename.text());
	FXString ext=FXPath::extension(FXPath::name(filename));
	    // printf("extension: %s\n",ext.text());
		if(comparecase(ext,"gz")==0||comparecase(ext,"bz2")==0||comparecase(ext,"zip")==0||comparecase(ext,"tar")==0||comparecase(ext,"tgz")==0||comparecase(ext,"7z")==0){
			type="Archive";
		}
		else if(comparecase(ext,"gif")==0||comparecase(ext,"tiff")==0||comparecase(ext,"png")==0||comparecase(ext,"jpg")==0||comparecase(ext,"xpm")==0||comparecase(ext,"bmp")==0){
			type="Image";
		}
		else{
			type="File";
		}
    }
      
    sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&type);
	
  return 1;
  }  


// Update file mimetype
long FXMovieManager::onUpdFileMimeType_lp(FXObject* sender,FXSelector,void*){

  FXString filename=filelist_lp->getCurrentFile();
  FXFileAssoc *fileassoc=NULL;
  FXString mimetype;

  if(FXStat::isFile(filename)){
    fileassoc=associations->findFileBinding(filename.text());
    }
  if(fileassoc){    
  	if(fileassoc->mimetype.text()){
          mimetype=FXString::value(fileassoc->mimetype.text());
  	}
  }
    sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&mimetype);
	
  return 1;
  }

long FXMovieManager::onUpdFileMimeType_rp(FXObject* sender,FXSelector,void*){

  FXString filename=filelist_rp->getCurrentFile();
  FXFileAssoc *fileassoc=NULL;
  FXString mimetype;

  if(FXStat::isFile(filename)){
    fileassoc=associations->findFileBinding(filename.text());
    }
  if(fileassoc){    
  	if(fileassoc->mimetype.text()){
          mimetype=FXString::value(fileassoc->mimetype.text());
  	}
  }
    sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&mimetype);
	
  return 1;
  }


  // Update file command
long FXMovieManager::onUpdFileCommand_lp(FXObject* sender,FXSelector,void*){

  FXString filename=filelist_lp->getCurrentFile();
  FXFileAssoc *fileassoc=NULL;
  FXString command;

  if(FXStat::isFile(filename)){
    fileassoc=associations->findFileBinding(filename.text());
    }
  if(fileassoc){    
  	if(fileassoc->command.text()){
          command=FXString::value(fileassoc->command.text());
  	}
  }
    sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&command);
	
  return 1;
  }

long FXMovieManager::onUpdFileCommand_rp(FXObject* sender,FXSelector,void*){

  FXString filename=filelist_rp->getCurrentFile();
  FXFileAssoc *fileassoc=NULL;
  FXString command;

  if(FXStat::isFile(filename)){
    fileassoc=associations->findFileBinding(filename.text());
    }
  if(fileassoc){    
  	if(fileassoc->command.text()){
          command=FXString::value(fileassoc->command.text());
  	}
  }
    sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&command);
	
  return 1;
  }

  
// Update file description
long FXMovieManager::onUpdFileDesc_lp(FXObject* sender,FXSelector,void*){

  FXString filename=filelist_lp->getCurrentFile();
  FXLabel *label=(FXLabel*)sender;
  FXFileAssoc *fileassoc=NULL;

  if(FXStat::isDirectory(filename)){
    fileassoc=associations->findDirBinding(filename.text());
    }
  else if(FXStat::isExecutable(filename)){
    fileassoc=associations->findExecBinding(filename.text());
    }
  else{
    fileassoc=associations->findFileBinding(filename.text());
    }

  label->setText(FXPath::name(filename));
  if(fileassoc){
    if(fileassoc->bigicon) fileassoc->bigicon->create();
    label->setIcon(fileassoc->bigicon);
    }
  else{
    label->setIcon(NULL); 
    }
  return 1;
  }

long FXMovieManager::onUpdFileDesc_rp(FXObject* sender,FXSelector,void*){

  FXString filename=filelist_rp->getCurrentFile();
  FXLabel *label=(FXLabel*)sender;
  FXFileAssoc *fileassoc=NULL;

  if(FXStat::isDirectory(filename)){
    fileassoc=associations->findDirBinding(filename.text());
    }
  else if(FXStat::isExecutable(filename)){
    fileassoc=associations->findExecBinding(filename.text());
    }
  else{
    fileassoc=associations->findFileBinding(filename.text());
    }

  label->setText(FXPath::name(filename));
  if(fileassoc){
    if(fileassoc->bigicon) fileassoc->bigicon->create();
    label->setIcon(fileassoc->bigicon);
    }
  else{
    label->setIcon(NULL); 
    }
  return 1;
  }


// find video aspect
long FXMovieManager::onUpdMovieAspect(FXObject* sender,FXSelector,void*){
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&grepvideo);
  return 1;
  }


// find audio format
long FXMovieManager::onUpdAudioFormat(FXObject* sender,FXSelector,void*){
  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&grepaudio);
  return 1;
  }


/****************************PERMISSIONS**********************************/
/***************see FXIO.h and FXStat.h for more information**************/

// filemode modification
long FXMovieManager::onCmdFilemode_lp(FXObject*,FXSelector sel,void*){
  FXString filename=filelist_lp->getCurrentFile();
  FXuint fmode=FXStat::mode(filename);
#ifndef WIN32
  switch(FXSELID(sel)) {
    case ID_READOWNER_LP: 
	fmode^=FXIO::OwnerRead; 
	break;
    case ID_WRITEOWNER_LP: 
	fmode^=FXIO::OwnerWrite; 
	break;
    case ID_EXECUTEOWNER_LP: 
	fmode^=FXIO::OwnerExec; 
	break;
    case ID_READGROUP_LP: 
	fmode^=FXIO::GroupRead; 
	break;
    case ID_WRITEGROUP_LP: 
	fmode^=FXIO::GroupWrite; 
	break;
    case ID_EXECUTEGROUP_LP: 
	fmode^=FXIO::GroupExec; 
	break;
    case ID_READOTHER_LP: 
	fmode^=FXIO::OtherRead; 
	break;
    case ID_WRITEOTHER_LP: 
	fmode^=FXIO::OtherWrite; 
	break;
    case ID_EXECUTEOTHER_LP: 
	fmode^=FXIO::OtherExec; 
	break;
    case ID_SETUID_LP: 
	fmode^=FXIO::SetUser; 
	break;
    case ID_SETGID_LP: 
	fmode^=FXIO::SetGroup; 
	break;
    case ID_STICKY_LP: 
	fmode^=FXIO::Sticky; 
	break;
    }
#endif
  if(!FXStat::mode(filename,fmode)){
    FXMessageBox::error(this,MBOX_OK,"Changing Permissions","Unable to Change Permissions on file: %s",filename.text());
    }
  return 1;
  }

long FXMovieManager::onCmdFilemode_rp(FXObject*,FXSelector sel,void*){
  FXString filename=filelist_rp->getCurrentFile();
  FXuint fmode=FXStat::mode(filename);
#ifndef WIN32
  switch(FXSELID(sel)) {
    case ID_READOWNER_RP: 
	fmode^=FXIO::OwnerRead; 
	break;
    case ID_WRITEOWNER_RP: 
	fmode^=FXIO::OwnerWrite; 
	break;
    case ID_EXECUTEOWNER_RP: 
	fmode^=FXIO::OwnerExec; 
	break;
    case ID_READGROUP_RP: 
	fmode^=FXIO::GroupRead; 
	break;
    case ID_WRITEGROUP_RP: 
	fmode^=FXIO::GroupWrite; 
	break;
    case ID_EXECUTEGROUP_RP: 
	fmode^=FXIO::GroupExec; 
	break;
    case ID_READOTHER_RP: 
	fmode^=FXIO::OtherRead; 
	break;
    case ID_WRITEOTHER_RP: 
	fmode^=FXIO::OtherWrite; 
	break;
    case ID_EXECUTEOTHER_RP: 
	fmode^=FXIO::OtherExec; 
	break;
    case ID_SETUID_RP: 
	fmode^=FXIO::SetUser; 
	break;
    case ID_SETGID_RP: 
	fmode^=FXIO::SetGroup; 
	break;
    case ID_STICKY_RP: 
	fmode^=FXIO::Sticky; 
	break;
    }
#endif
  if(!FXStat::mode(filename,fmode)){
    FXMessageBox::error(this,MBOX_OK,"Changing Permissions","Unable to Change Permissions on file: %s",filename.text());
    }
  return 1;
  }



// update filemode
long FXMovieManager::onUpdFilemode_lp(FXObject* sender,FXSelector sel,void*){
  FXString filename=filelist_lp->getCurrentFile();
  FXuint fmode=FXStat::mode(filename);
  FXuint find=0;
#ifndef WIN32
  switch(FXSELID(sel)) {
    case ID_READOWNER_LP: 
	find=(fmode&FXIO::OwnerRead); 
	break;
    case ID_WRITEOWNER_LP: 
	find=(fmode&FXIO::OwnerWrite); 
	break;
    case ID_EXECUTEOWNER_LP: 
	find=(fmode&FXIO::OwnerExec); 
	break;
    case ID_READGROUP_LP: 
	find=(fmode&FXIO::GroupRead); 
	break;
    case ID_WRITEGROUP_LP: 
	find=(fmode&FXIO::GroupWrite); 
	break;
    case ID_EXECUTEGROUP_LP: 
	find=(fmode&FXIO::GroupExec); 
	break;
    case ID_READOTHER_LP: 
	find=(fmode&FXIO::OtherRead); 
	break;
    case ID_WRITEOTHER_LP: 
	find=(fmode&FXIO::OtherWrite); 
	break;
    case ID_EXECUTEOTHER_LP: 
	find=(fmode&FXIO::OtherExec); 
	break;
    case ID_SETUID_LP: 
	find=(fmode&FXIO::SetUser); 
	break;
    case ID_SETGID_LP: 
	find=(fmode&FXIO::SetGroup); 
	break;
    case ID_STICKY_LP: 
	find=(fmode&FXIO::Sticky); 
	break;
    }
#endif
  sender->handle(this,find ? FXSEL(SEL_COMMAND,ID_CHECK) : FXSEL(SEL_COMMAND,ID_UNCHECK),NULL);
  return 1;
  }

long FXMovieManager::onUpdFilemode_rp(FXObject* sender,FXSelector sel,void*){
  FXString filename=filelist_rp->getCurrentFile();
  FXuint fmode=FXStat::mode(filename);
  FXuint find=0;
#ifndef WIN32
  switch(FXSELID(sel)) {
    case ID_READOWNER_RP: 
	find=(fmode&FXIO::OwnerRead); 
	break;
    case ID_WRITEOWNER_RP: 
	find=(fmode&FXIO::OwnerWrite); 
	break;
    case ID_EXECUTEOWNER_RP: 
	find=(fmode&FXIO::OwnerExec); 
	break;
    case ID_READGROUP_RP: 
	find=(fmode&FXIO::GroupRead); 
	break;
    case ID_WRITEGROUP_RP: 
	find=(fmode&FXIO::GroupWrite); 
	break;
    case ID_EXECUTEGROUP_RP: 
	find=(fmode&FXIO::GroupExec); 
	break;
    case ID_READOTHER_RP: 
	find=(fmode&FXIO::OtherRead); 
	break;
    case ID_WRITEOTHER_RP: 
	find=(fmode&FXIO::OtherWrite); 
	break;
    case ID_EXECUTEOTHER_RP: 
	find=(fmode&FXIO::OtherExec); 
	break;
    case ID_SETUID_RP: 
	find=(fmode&FXIO::SetUser); 
	break;
    case ID_SETGID_RP: 
	find=(fmode&FXIO::SetGroup); 
	break;
    case ID_STICKY_RP: 
	find=(fmode&FXIO::Sticky); 
	break;
    }
#endif
  sender->handle(this,find ? FXSEL(SEL_COMMAND,ID_CHECK) : FXSEL(SEL_COMMAND,ID_UNCHECK),NULL);
  return 1;
  }

// show only functions
// Update owner
long FXMovieManager::onUpdOwner_lp(FXObject* sender,FXSelector,void*){
  FXStat showonly;
  FXStat::statFile(filelist_lp->getCurrentFile(),showonly);
  FXString oname=FXSystem::userName(showonly.user());

  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&oname);
  return 1;
  }

long FXMovieManager::onUpdOwner_rp(FXObject* sender,FXSelector,void*){
  FXStat showonly;
  FXStat::statFile(filelist_rp->getCurrentFile(),showonly);
  FXString oname=FXSystem::userName(showonly.user());

  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&oname);
  return 1;
  }


// Update group
long FXMovieManager::onUpdGroup_lp(FXObject* sender,FXSelector,void*){
  FXStat showonly;
  FXStat::statFile(filelist_lp->getCurrentFile(),showonly);
  FXString gname=FXSystem::groupName(showonly.group());

  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&gname);
  return 1;
  }

long FXMovieManager::onUpdGroup_rp(FXObject* sender,FXSelector,void*){
  FXStat showonly;
  FXStat::statFile(filelist_rp->getCurrentFile(),showonly);
  FXString gname=FXSystem::groupName(showonly.group());

  sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&gname);
  return 1;
  }



/****************************FILE LIST BUTTONS**********************************/

// Play the movies from the file list
long FXMovieManager::onCmdFilePlay_lp(FXObject*,FXSelector,void*){
     cmdFilePlay(filelist_lp);
  return 1;
  }

long FXMovieManager::onCmdFilePlay_rp(FXObject*,FXSelector,void*){
     cmdFilePlay(filelist_rp);
  return 1;
  }

void FXMovieManager::cmdFilePlay(FXFileList* panel){
   // reset the fifo every time we play
   system (removeTempFifo.text());
   system (createTempFifo.text());

   FXString filetoplay, file, playlist;

    for(FXint i=0; i<panel->getNumItems(); i++){
      if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;
		if(panel->isItemDirectory(i)) continue;
		if(panel->isItemFile(i)){
          filetoplay=FXPath::absolute(panel->getDirectory(),file);
		  playlist=playlist+" "+filetoplay;
		}
      }
    }

    //get the speed
    FXfloat i_speed = speedslider->getValue();
    FXfloat f_speed = (i_speed/10);
    FXString s_speed = "-speed " + (FXString::value(f_speed,2)) + " ";

    //get the volume
    FXint i_volume = volumeslider->getValue();
    FXString s_volume = "-volume " + (FXString::value(i_volume)) + " ";

    // correct command list sequence....
    FXString i_command = s_command + s_loop + s_video + s_size + s_full + s_speed + s_volume;
    if(!playlist.empty()){
       s_play = i_command + playlist + " &";

       while (isrunning == TRUE){
       int ret=system("killall mplayer");
             if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)){
                 break;
	     }
	     isrunning = FALSE;
	}

       while (isrunning == FALSE){
       int ret=system(s_play.text());
            if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)){
                 break;
	     	}
	    isrunning = TRUE;
		}
    }
    // clear liststr
    i_command.clear();
    playlist.clear();
    s_play.clear();

  return;
  }

// Stop player
long FXMovieManager::onCmdFileStop_lp(FXObject *,FXSelector,void *) {
     cmdFileStop();	
  return 1;
  }

long FXMovieManager::onCmdFileStop_rp(FXObject *,FXSelector,void *) {
     cmdFileStop();	
  return 1;
  }

void FXMovieManager::cmdFileStop() {
	if (isrunning==TRUE){
		FILE *f;
		f=fopen(currentFifo.text(), "w");
		fprintf(f, "%s", "quit 1\n");
		fclose(f);
		isrunning=FALSE;
	} else if (isrunning==FALSE){
		FXMessageBox::information(this,MBOX_OK,"Command Failure","MPlayer is NOT running");
	}
	return;
}

// Pause playing the movie from the file list
long FXMovieManager::onCmdFilePause_lp(FXObject *,FXSelector,void *) {
     cmdFilePause();	
  return 1;
  }

long FXMovieManager::onCmdFilePause_rp(FXObject *,FXSelector,void *) {
     cmdFilePause();	
  return 1;
  }

void FXMovieManager::cmdFilePause(){
	if (isrunning==TRUE){
		FILE *f;
		f=fopen(currentFifo.text(), "w");
		fprintf(f, "%s", "pause\n");
		fclose(f);
	} else if (isrunning==FALSE){
		FXMessageBox::information(this,MBOX_OK,"Command Failure","MPlayer is NOT running");
	}
	return;
  }




/***************************MOVIE PREVIEW*******************************/

// Thumbnails creation
long FXMovieManager::onCmdFilePreview_lp(FXObject*,FXSelector,void*){
    cmdFilePreview(filelist_lp);
    return 1;
}

long FXMovieManager::onCmdFilePreview_rp(FXObject*,FXSelector,void*){
    cmdFilePreview(filelist_rp);
    return 1;
}

void FXMovieManager::cmdFilePreview(FXFileList* panel){
    FXString filetoplay,file,rootname,imagename,infoname;

    progressbar->setProgress(0);
    progressbar->setTotal(panel->getNumItems());
    progressbar->setBarColor(FXRGB(0,0,150));

    for(FXint i=0; i<panel->getNumItems(); i++){
      progressbar->setProgress(i+1);

      if(panel->isItemSelected(i)){
      	file=panel->getItemFilename(i);
      	if(file=="..") continue;
		if(panel->isItemDirectory(i)) continue;
	      
        filetoplay=FXPath::absolute(panel->getDirectory(),file);
        rootname="."+FXPath::stripExtension(file);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        imagename=FXPath::enquote(panel->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        imagename=FXPath::enquote(panel->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
    	FXFile truefile;
		FXbool istruefile = truefile.open(imagename);
		if(!istruefile){
			if(!createImageMovie(filetoplay,rootname)){
				FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Creation Failure","Error on creating thumbnail: %s",imagename.text());
    			if(answer==MBOX_CLICKED_CANCEL){
      				return;
    			}
			}
		}
		else{
			continue;
		}
      }
    }

#ifndef HAVE_LIBFFMPEGTHUMBNAILER
    system("sleep 2s");
#endif
    progressbar->setProgress(0);
    progressbar->setBarColor(FXRGB(150,0,0));


    for(FXint i=0; i<panel->getNumItems(); i++){
      progressbar->setProgress(i+1);

      if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;
		if(panel->isItemDirectory(i)) continue;

        rootname="."+FXPath::stripExtension(file);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        imagename=FXPath::enquote(panel->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        imagename=FXPath::enquote(panel->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
		if(!previewMovie(panel, imagename, i, imgw, imgh)){
			FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
   			if(answer==MBOX_CLICKED_CANCEL){
   				return;
   			}
		}
      }
    }	
    return;
}


// Force tumbnails recreation 
long FXMovieManager::onCmdForceFilePreview_lp(FXObject*,FXSelector,void*){
    cmdForceFilePreview(filelist_lp);
    return 1;
}

long FXMovieManager::onCmdForceFilePreview_rp(FXObject*,FXSelector,void*){
    cmdForceFilePreview(filelist_rp);
    return 1;
}

void FXMovieManager::cmdForceFilePreview(FXFileList* panel){
    FXString filetoplay,file,rootname,imagename;

    progressbar->setProgress(0);
    progressbar->setTotal(panel->getNumItems());
    progressbar->setBarColor(FXRGB(0,0,150));

    for(FXint i=0; i<panel->getNumItems(); i++){
      progressbar->setProgress(i+1);

      if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;
		if(panel->isItemDirectory(i)) continue;
	      
        filetoplay=FXPath::absolute(panel->getDirectory(),file);
        rootname="."+FXPath::stripExtension(file);

    	FXFile truefile;
		FXbool istruefile = truefile.open(imagename);
		if(!istruefile){
			if(!createImageMovie(filetoplay,rootname)){
				FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Creation Failure","Error on creating thumbnail: %s",imagename.text());
    			if(answer==MBOX_CLICKED_CANCEL){
      				return;
    			}
			}
		}
		else if(istruefile){
			if(!createImageMovie(filetoplay,rootname)){
				FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Creation Failure","Error on creating thumbnail: %s",imagename.text());
  				if(answer==MBOX_CLICKED_CANCEL){
   					return;
   				}
			}
		}
		else{
			continue;
		}
      }
   }

#ifndef HAVE_LIBFFMPEGTHUMBNAILER
    system("sleep 2s");
#endif
    progressbar->setProgress(0);
    progressbar->setBarColor(FXRGB(150,0,0));


    for(FXint i=0; i<panel->getNumItems(); i++){
      progressbar->setProgress(i+1);

      if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;
		if(panel->isItemDirectory(i)) continue;

        rootname="."+FXPath::stripExtension(file);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        imagename=FXPath::enquote(panel->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        imagename=FXPath::enquote(panel->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
		if(!previewMovie(panel, imagename, i, imgw, imgh)){
			FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
   			if(answer==MBOX_CLICKED_CANCEL){
   				return;
   			}
		}
      }
   }
   return;
}


// Resize tumbnails preview 
inline
long FXMovieManager::onCmdWideScreen(FXObject*,FXSelector,void*){

	if(widescreen==TRUE){
		widescreen=FALSE;
		imgw=80; imgh=64;
	} 
	else if(widescreen==FALSE){
		widescreen=TRUE;
		imgw=120; imgh=64;
	}
	else {
		widescreen=FALSE;
		imgw=80; imgh=64;
	}	
   return 1;
}


// Resize tumbnails preview 
inline
long FXMovieManager::onCmdIncFilePreview_lp(FXObject*,FXSelector,void*){
    cmdIncFilePreview(filelist_lp);
    return 1;
}

inline
long FXMovieManager::onCmdIncFilePreview_rp(FXObject*,FXSelector,void*){
    cmdIncFilePreview(filelist_rp);
    return 1;
}
    
inline
void FXMovieManager::cmdIncFilePreview(FXFileList* panel){
    FXString file,rootname,imagename;

	if(widescreen==TRUE){
 		imgw=imgw+60;
    	imgh=imgh+32;
	} 
	else if(widescreen==FALSE){
 		imgw=imgw+40;
    	imgh=imgh+32;
	}
	else {
  		imgw=imgw+40;
    	imgh=imgh+32;
	}	

    for(FXint i=0; i<panel->getNumItems(); i++){

      if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;
		if(panel->isItemDirectory(i)) continue;

        rootname="."+FXPath::stripExtension(file);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        imagename=FXPath::enquote(panel->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        imagename=FXPath::enquote(panel->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
		if(!previewMovie(panel, imagename, i, imgw, imgh)){
			FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
   			if(answer==MBOX_CLICKED_CANCEL){
   				return;
   			}
		}
      }
    }	
   return;
}


// Resize tumbnails preview 
inline
long FXMovieManager::onCmdDecFilePreview_lp(FXObject*,FXSelector,void*){
    cmdDecFilePreview(filelist_lp);
    return 1;
}

inline
long FXMovieManager::onCmdDecFilePreview_rp(FXObject*,FXSelector,void*){
    cmdDecFilePreview(filelist_rp);
    return 1;
}

inline
void FXMovieManager::cmdDecFilePreview(FXFileList* panel){
    FXString file,rootname,imagename;

	if(widescreen==TRUE){
 		imgw=imgw-60;
    	imgh=imgh-32;
	} 
	else if(widescreen==FALSE){
 		imgw=imgw-40;
    	imgh=imgh-32;
	}
	else {
  		imgw=imgw-40;
    	imgh=imgh-32;
	}

    for(FXint i=0; i<panel->getNumItems(); i++){

      if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;
		if(panel->isItemDirectory(i)) continue;

        rootname="."+FXPath::stripExtension(file);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
        imagename=FXPath::enquote(panel->getDirectory()+"/"+rootname+"_thumbnail.jpg");
#else
        imagename=FXPath::enquote(panel->getDirectory())+"/"+rootname+"00000001/00000001.jpg";
#endif
		if(!previewMovie(panel, imagename, i, imgw, imgh)){
			FXuint answer=FXMessageBox::error(this,MBOX_OK_CANCEL,"Thumbnail Loading Failure","Error on loading thumbnail: %s",imagename.text());
   			if(answer==MBOX_CLICKED_CANCEL){
   				return;
   			}
		}
      }
    }	
   return;
}



// clear all preview and all hidden directories
inline
long FXMovieManager::onCmdClearPreview_lp(FXObject*,FXSelector,void*){
    cmdClearPreview(filelist_lp);
    return 1;
}

inline
long FXMovieManager::onCmdClearPreview_rp(FXObject*,FXSelector,void*){
    cmdClearPreview(filelist_rp);
    return 1;
}

inline
void FXMovieManager::cmdClearPreview(FXFileList* panel){
    FXString dirname,executable;
    progressbar->setProgress(0);
    progressbar->setBarColor(FXRGB(0,150,0));
    progressbar->setTotal(panel->getNumItems());

    FXbool state=panel->showHiddenFiles();
    if(state==FALSE){
		panel->showHiddenFiles(TRUE);
	}
    if(panel->showHiddenFiles()==TRUE){
    	for(FXint i=0; i<panel->getNumItems(); i++){
       		progressbar->setProgress(i+1);
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
       		if(panel->isItemFile(i)){
#else
       		if(panel->isItemDirectory(i)){
#endif
       	  		filename=panel->getItemFilename(i);
          		if(filename=="..") continue;
	  		// use of FXRex
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
    	  		FXRex rexfile("_thumbnail.jpg",0);
#else
    	  		FXRex rexfile("00000001",0);
#endif
          		FXint begd,endd;
          		if(rexfile.match(filename,&begd,&endd)){
             			executable="cd "+FXPath::enquote(panel->getDirectory())+"; rm -r "+filename+" &";
             			system(executable.text());
	  		}
		}
     	}
  	}
  panel->showHiddenFiles(state);
  return;
  }


// Create thumbnails of movies
inline
FXbool FXMovieManager::createImageMovie(const FXString& filetoplay,const FXString& rootname){
#ifdef HAVE_LIBFFMPEGTHUMBNAILER
    FXfloat seekvalue=seekspinner->getValue();
    FXString seekstring = "00:00:"+(FXString::value(seekvalue,2));	
    string seekTime = seekstring.text();

    int     seekPercentage = 10;

    string  inputFile = filetoplay.text();
    FXString imagename=FXPath::enquote(FXPath::directory(filetoplay)+"/"+rootname+"_thumbnail.jpg");
    string  outputFile = imagename.text();

    try
    {
        VideoThumbnailer videoThumbnailer(thumbnailSize, workaroundIssues, maintainAspectRatio, imageQuality, smartFrameSelection);
        FilmStripFilter* filmStripFilter = NULL;

        if (filmStripOverlay)
        {
            filmStripFilter = new FilmStripFilter();
            videoThumbnailer.addFilter(filmStripFilter);
        }

        if (!seekTime.empty())
        {
            videoThumbnailer.setSeekTime(seekTime);
        }
        else
        {
            videoThumbnailer.setSeekPercentage(seekPercentage);
        }
        videoThumbnailer.generateThumbnail(inputFile, Jpeg, outputFile);
		// free memory
        delete filmStripFilter;
    }
    catch (exception& e)
    {
        cerr << "Error: " << e.what() << endl;
        return FALSE;
    }
    catch (...)
    {
        return FALSE;
    }
#else
	FXfloat seekvalue=seekspinner->getValue();
	FXString seekstring	= "-ss " + (FXString::value(seekvalue,2)) + " ";
	FXString command="mplayer "+seekstring+"-frames 1 -nocache -ao null -vo jpeg:subdirs="+rootname+" "+filetoplay+" &";
	FXString executable="cd "+FXPath::enquote(FXPath::directory(filetoplay))+"; "+command;

	// printf("executable: %s\n", executable.text());
    system(executable.text());
#endif
    //printf("done creating thumbnail\n");
    return TRUE;
 }


inline
FXbool FXMovieManager::createInfoMovie(const FXString& filename){

	FXString propcommand="mplayer -frames 0 -nocache -ao null -vo null "+filename+" &> "+currentInfo+" &";
	// printf("propcommand: %s\n", propcommand.text());
	system(propcommand.text());

    return TRUE;
 }


// Load imagemovie to icon for preview
inline
FXbool FXMovieManager::previewMovie(FXFileList* panel, const FXString& filename, FXint i, FXint w, FXint h){

    FXString imagename=filename;
#ifdef HAVE_LIBFOX_1_6
    FXIcon *tempicon=NULL;
    tempicon=new FXIcon(getApp(),NULL,IMAGE_OPAQUE|IMAGE_KEEP);
#endif
#ifdef HAVE_LIBFOX_1_7
    FXJPGIcon *tempicon=NULL;
    tempicon=new FXJPGIcon(getApp(),NULL,IMAGE_OPAQUE|IMAGE_KEEP);
#endif
    // printf("the imagename is: %s\n", imagename.text());
   //  printf("the leftpanel is: %i\n", leftpanel);
    if (tempicon){
   		FXFileStream stream;
   		if (stream.open(imagename,FXStreamLoad)){
			getApp()->beginWaitCursor();
			tempicon->loadPixels(stream);
			stream.close();
			tempicon->scale(w,h,75); 
   	 		tempicon->create();
			//freeImageMemory(panel,i);
			panel->setItemBigIcon(i,tempicon);
			getApp()->endWaitCursor();
			return TRUE;
		}
    }
    return FALSE;
}

inline
FXbool FXMovieManager::freeImageMemory(FXFileList* panel, FXint i) {
    FXIcon *oldicon=NULL;
    FXString file;
    file=panel->getItemFilename(i);
    if(file=="..") return FALSE;
  	if(panel->getItemBigIcon(i)){
		oldicon=panel->getItemBigIcon(i);
		panel->setItemBigIcon(i,NULL);
  		FXASSERT(oldicon != NULL);
		delete oldicon;  // delete the current icon
		oldicon=NULL;
   	}
	return TRUE;
}



// Map Host
FXbool FXMovieManager::executeMapping(FXFileList* panel, const FXString& map_cmd, const FXString& map_label){

  FXString location=FXPath::enquote(panel->getDirectory());
  FXString executable=map_cmd+location;

  	  FXDialogBox dialog(this,map_label,DECOR_TITLE|DECOR_BORDER);
	  FXVerticalFrame* content=new FXVerticalFrame(&dialog,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,10,10,10,10);
  	  new FXLabel(content,location,NULL,LAYOUT_FILL_X);
  	  FXHorizontalFrame* buttons=new FXHorizontalFrame(content,LAYOUT_FILL_X|PACK_UNIFORM_WIDTH,0,0,0,0,0,0,0,0);
  	  new FXButton(buttons,"&YES",m_accepticon,&dialog,FXDialogBox::ID_ACCEPT,BUTTON_INITIAL|ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT);
  	  new FXButton(buttons,"&Cancel",m_cancelicon,&dialog,FXDialogBox::ID_CANCEL,ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,0,0,0,0,20,20);
  	  dialog.create();

  	  if(dialog.execute()){

         int ret = system(executable.text());
         if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)){
            return FALSE;
	     }
      }
  return TRUE;
  }

// Map Host
long FXMovieManager::onCmdMapHost_lp(FXObject*,FXSelector,void*){
  FXString map_cmd="mount ";
  FXString map_label="MAPPING DEVICE";
  if(!executeMapping(filelist_lp, map_cmd, map_label)){
  	FXMessageBox::error(this,MBOX_OK,"MAPPING NOT EXECUTED","Operation Failed!");
  }
  return 1;
  }

// Unmap Host
long FXMovieManager::onCmdUnmapHost_lp(FXObject*,FXSelector,void*){
  FXString map_cmd="umount ";
  FXString map_label="UNMAPPING DEVICE";
  if(!executeMapping(filelist_lp, map_cmd, map_label)){
  	FXMessageBox::error(this,MBOX_OK,"UNMAPPING NOT EXECUTED","Operation Failed!");
  }
  return 1;
  }

// Map Host
long FXMovieManager::onCmdMapHost_rp(FXObject*,FXSelector,void*){
  FXString map_cmd="mount ";
  FXString map_label="MAPPING DEVICE";
  if(!executeMapping(filelist_rp, map_cmd, map_label)){
  	FXMessageBox::error(this,MBOX_OK,"MAPPING NOT EXECUTED","Operation Failed!");
  }
  return 1;
  }

// Unmap Host
long FXMovieManager::onCmdUnmapHost_rp(FXObject*,FXSelector,void*){
  FXString map_cmd="umount ";
  FXString map_label="UNMAPPING DEVICE";
  if(!executeMapping(filelist_rp, map_cmd, map_label)){
  	FXMessageBox::error(this,MBOX_OK,"UNMAPPING NOT EXECUTED","Operation Failed!");
  }
  return 1;
  }


// Execute Compress Command
FXbool FXMovieManager::executeCompressCommand(FXFileList* panel, const FXString& dlg_cmd, const FXString& dlg_ext, const FXString& dlg_label, FXint dlg_form){

    FXString filetocomp,file, executable;
    for(FXint i=0; i<panel->getNumItems(); i++){
      if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;
        filetocomp=FXPath::name(file);

		if (dlg_form==1)    { executable="cd "+FXPath::enquote(panel->getDirectory())+"; "+dlg_cmd+filetocomp+dlg_ext+filetocomp+" &"; }
		else if(dlg_form==2){ executable="cd "+FXPath::enquote(panel->getDirectory())+"; "+dlg_cmd+filetocomp+" &"; }

		FXDialogBox dialog(this,dlg_label,DECOR_TITLE|DECOR_BORDER);
		FXVerticalFrame* content=new FXVerticalFrame(&dialog,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,10,10,10,10);
		new FXLabel(content,filetocomp,NULL,LAYOUT_FILL_X);
  	  	FXHorizontalFrame* buttons=new FXHorizontalFrame(content,LAYOUT_FILL_X|PACK_UNIFORM_WIDTH,0,0,0,0,0,0,0,0);
  	  	new FXButton(buttons,"&YES",m_accepticon,&dialog,FXDialogBox::ID_ACCEPT,BUTTON_INITIAL|ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT);
  	  	new FXButton(buttons,"&Cancel",m_cancelicon,&dialog,FXDialogBox::ID_CANCEL,ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,0,0,0,0,20,20);
  	  	dialog.create();

  	  	if(dialog.execute()){
	        int ret = system(executable.text());
	        if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)){
    	        return FALSE;
	     	}
       	}
      }
    }
  return TRUE;
  }

// Compress TAR
void FXMovieManager::CompressTAR(FXFileList* panel){
#ifndef DISABLE_TAR
    dlg_cmd="tar -cf ";
	dlg_ext =".tar ";
	dlg_label="COMPRESSING: tar -cf";
	dlg_form=1;
	if(!executeCompressCommand(panel, dlg_cmd, dlg_ext, dlg_label, dlg_form)){
  	  	FXMessageBox::error(this,MBOX_OK,"COMPRESSION NOT EXECUTED","Operation Failed!");
	}
#endif
  return;
  }

// Compress GZ
void FXMovieManager::CompressGZ(FXFileList* panel){
#ifndef DISABLE_GZIP
    dlg_cmd="gzip ";
	dlg_ext =".gz ";
	dlg_label="COMPRESSING: gzip";
	dlg_form=2;
	if(!executeCompressCommand(panel, dlg_cmd, dlg_ext, dlg_label, dlg_form)){
  	  	FXMessageBox::error(this,MBOX_OK,"COMPRESSION NOT EXECUTED","Operation Failed!");
	}
#endif
  return;
  }

// Compress BZ2
void FXMovieManager::CompressBZ2(FXFileList* panel){
#ifndef DISABLE_BZIP2
    dlg_cmd="bzip2 -z ";
	dlg_ext =".bz2 ";
	dlg_label="COMPRESSING: bzip2 -z";
	dlg_form=2;
	if(!executeCompressCommand(panel, dlg_cmd, dlg_ext, dlg_label, dlg_form)){
  	  	FXMessageBox::error(this,MBOX_OK,"COMPRESSION NOT EXECUTED","Operation Failed!");
	}
#endif
  return;
  }

// Compress XZ
void FXMovieManager::CompressXZ(FXFileList* panel){
#ifndef DISABLE_XZIP
    dlg_cmd="xz -z ";
	dlg_ext =".xz ";
	dlg_label="COMPRESSING: xz -z";
	dlg_form=2;
	if(!executeCompressCommand(panel, dlg_cmd, dlg_ext, dlg_label, dlg_form)){
  	  	FXMessageBox::error(this,MBOX_OK,"COMPRESSION NOT EXECUTED","Operation Failed!");
	}
#endif
  return;
  }

// Compress ZIP
void FXMovieManager::CompressZIP(FXFileList* panel){
#ifndef DISABLE_ZIP
    dlg_cmd="zip ";
	dlg_ext =".zip -xi ";
	dlg_label="COMPRESSING: zip -xi";
	dlg_form=1;
	if(!executeCompressCommand(panel, dlg_cmd, dlg_ext, dlg_label, dlg_form)){
  	  	FXMessageBox::error(this,MBOX_OK,"COMPRESSION NOT EXECUTED","Operation Failed!");
	}
#endif
  return;
  }

// Compress 7Z
void FXMovieManager::Compress7ZIP(FXFileList* panel){
#ifndef DISABLE_SEVENZIP
    dlg_cmd="7z a ";
	dlg_ext =".7z ";
	dlg_label="COMPRESSING: 7z a";
	dlg_form=1;
	if(!executeCompressCommand(panel, dlg_cmd, dlg_ext, dlg_label, dlg_form)){
  	  	FXMessageBox::error(this,MBOX_OK,"COMPRESSION NOT EXECUTED","Operation Failed!");
	}
#endif
  return;
  }



// Left Panel Compression commands
// Compress TAR
long FXMovieManager::onCmdCompressTAR_lp(FXObject*,FXSelector,void*){
    CompressTAR(filelist_lp);
    return 1;
  }

// Compress XZ
long FXMovieManager::onCmdCompressXZ_lp(FXObject*,FXSelector,void*){
	CompressXZ(filelist_lp);
  	return 1;
  }

// Compress ZIP
long FXMovieManager::onCmdCompressZIP_lp(FXObject*,FXSelector,void*){
	CompressZIP(filelist_lp);
  	return 1;
  }

// Compress 7Z
long FXMovieManager::onCmdCompress7ZIP_lp(FXObject*,FXSelector,void*){
	Compress7ZIP(filelist_lp);
    return 1;
  }

// Compress GZ
long FXMovieManager::onCmdCompressGZ_lp(FXObject*,FXSelector,void*){
	CompressGZ(filelist_lp);
    return 1;
  }

// Compress BZ2
long FXMovieManager::onCmdCompressBZ2_lp(FXObject*,FXSelector,void*){
	CompressBZ2(filelist_lp);
    return 1;
  }

// Right Panel Compression commands
// Compress TAR
long FXMovieManager::onCmdCompressTAR_rp(FXObject*,FXSelector,void*){
	CompressTAR(filelist_rp);
    return 1;
  }

// Compress XZ
long FXMovieManager::onCmdCompressXZ_rp(FXObject*,FXSelector,void*){
	CompressXZ(filelist_rp);
    return 1;
  }

// Compress ZIP
long FXMovieManager::onCmdCompressZIP_rp(FXObject*,FXSelector,void*){
	CompressZIP(filelist_rp);
    return 1;
  }

// Compress 7Z
long FXMovieManager::onCmdCompress7ZIP_rp(FXObject*,FXSelector,void*){
	Compress7ZIP(filelist_rp);
    return 1;
  }

// Compress GZ
long FXMovieManager::onCmdCompressGZ_rp(FXObject*,FXSelector,void*){
	CompressGZ(filelist_rp);
    return 1;
  }

// Compress BZ2
long FXMovieManager::onCmdCompressBZ2_rp(FXObject*,FXSelector,void*){
	CompressBZ2(filelist_rp);
    return 1;
  }




// Extract files from archive
long FXMovieManager::onCmdExtract_lp(FXObject*,FXSelector,void*){
	if(!cmdExtractArchive(filelist_lp)){
  	  FXMessageBox::error(this,MBOX_OK,"EXTRACTION NOT EXECUTED","Operation Failed!");
	}
return 1;
}

long FXMovieManager::onCmdExtract_rp(FXObject*,FXSelector,void*){
	if(!cmdExtractArchive(filelist_rp)){
  	  FXMessageBox::error(this,MBOX_OK,"EXTRACTION NOT EXECUTED","Operation Failed!");
	}
return 1;
}

// Extract archive
FXbool FXMovieManager::cmdExtractArchive(FXFileList* panel){
  FXString command,file,filename,rootname;

  for(FXint i=0; i<panel->getNumItems(); i++){
    if(panel->isItemSelected(i)){
        file=panel->getItemFilename(i);
        if(file=="..") continue;

		// determine extension
    	filename=FXPath::absolute(panel->getDirectory(),file);
		FXString ext=FXPath::extension(filename);
		printf("ext: %s\n", ext.text());

		// determine sub-extension
		rootname=FXPath::stripExtension(file);
		FXString root=FXPath::extension(rootname);
		printf("root: %s\n", root.text());

	  	// Determine type of archive
		if(comparecase(ext,"gz")==0){
			if(comparecase(root,"tar")==0){
				command="tar -zxf";
			}	
			else if(comparecase(root,"tar")!=0){
				command="gunzip"; 
			}	
		}
		else if(comparecase(ext,"bz2")==0){
			if(comparecase(root,"tar")==0){
				command="tar -jxf"; 
			}	
			else if(comparecase(root,"tar")!=0){
				command="bunzip2"; 
			}	
		}
		else if(comparecase(ext,"xz")==0){
			command="xz -d";
		}
		else if(comparecase(ext,"lzma")==0){
			command="lzma -d";
		}
		else if(comparecase(ext,"tgz")==0){
			command="tar -zxf";
		}
		else if(comparecase(ext,"tar")==0){
			command="tar -xf";
		}
		else if(comparecase(ext,"zip")==0){
			command="unzip";
		}
		else if(comparecase(ext,"rar")==0){
	  		command="unrar x"; 
		}
		else if(comparecase(ext,"7z")==0){
	  		command="7z e"; 
		}
		else{
	  		FXMessageBox::error(this,MBOX_OK,"COMMAND FAILURE","Not suitable extension found %s\n",root.text());
	  		return FALSE;
		}	
	
		FXString executable="cd "+FXPath::enquote(panel->getDirectory())+"; "+command+" "+file+" &";

  		FXDialogBox dialog(this,"EXTRACTING",DECOR_TITLE|DECOR_BORDER);
		FXVerticalFrame* content=new FXVerticalFrame(&dialog,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,10,10,10,10);
  		new FXLabel(content,file,NULL,LAYOUT_FILL_X);
  		FXHorizontalFrame* buttons=new FXHorizontalFrame(content,LAYOUT_FILL_X|PACK_UNIFORM_WIDTH,0,0,0,0,0,0,0,0);
  		new FXButton(buttons,"&YES",m_accepticon,&dialog,FXDialogBox::ID_ACCEPT,BUTTON_INITIAL|ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT);
  		new FXButton(buttons,"&Cancel",m_cancelicon,&dialog,FXDialogBox::ID_CANCEL,ICON_BEFORE_TEXT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT,0,0,0,0,20,20);
  		dialog.create();

  		if(dialog.execute()){
    		int ret = system(executable.text());
        	if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)){
    	        return FALSE;
	    	}
    	}
  	  }
 	}
  return TRUE;
  }


// Scan for Virus with libclamav
long FXMovieManager::onCmdScanVirus(FXObject*,FXSelector,void*){
    scanvirus->show(PLACEMENT_OWNER);
    return 1;
  }


// EOF
